@@ -18,13 +18,13 @@
 
 #include <config.h>
 
-#include <stdio.h>
+#include "openconnect-internal.h"
+
+#include <libxml/parser.h>
+#include <libxml/tree.h>
+
 #include <unistd.h>
 #include <fcntl.h>
-#include <time.h>
-#include <string.h>
-#include <ctype.h>
-#include <errno.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #ifndef _WIN32
@@ -33,10 +33,11 @@
 #include <grp.h>
 #endif
 
-#include <libxml/parser.h>
-#include <libxml/tree.h>
-
-#include "openconnect-internal.h"
+#include <stdio.h>
+#include <time.h>
+#include <string.h>
+#include <ctype.h>
+#include <errno.h>
 
 static int xmlpost_append_form_opts(struct openconnect_info *vpninfo,
 				    struct oc_auth_form *form, struct oc_text_buf *body);
@@ -89,7 +90,7 @@ static int parse_auth_choice(struct openconnect_info *vpninfo, struct oc_auth_fo
 
 	/* Return early when there is a <select/> tag with no children */
 	if (max_choices == 0) {
-	    return 0;
+		return 0;
 	}
 
 	opt = calloc(1, sizeof(*opt));
@@ -386,6 +387,29 @@ static int xmlnode_get_text(xmlNode *xml_node, const char *name, char **var)
  * 2) The new <form> tag tends to omit the method/action properties.
  */
 
+/* Translate platform names (derived from AnyConnect) into the relevant
+ * CSD tag names
+ */
+static inline const char *csd_tag_name(struct openconnect_info *vpninfo)
+{
+	if (!strcmp(vpninfo->platname, "mac-intel"))
+		return "csdMac";
+	else if (!strcmp(vpninfo->platname, "win"))
+		return "csd";
+	else
+		/* linux, linux-64, android, apple-ios */
+		return "csdLinux";
+}
+
+/* Ignore stubs on mobile platforms */
+static inline int csd_use_stub(struct openconnect_info *vpninfo)
+{
+	if (!strcmp(vpninfo->platname, "android") || !strcmp(vpninfo->platname, "apple-ios"))
+		return 0;
+	else
+		return 1;
+}
+
 static int parse_auth_node(struct openconnect_info *vpninfo, xmlNode *xml_node,
 			   struct oc_auth_form *form)
 {
@@ -434,9 +458,9 @@ static int parse_auth_node(struct openconnect_info *vpninfo, xmlNode *xml_node,
 		   nodes; one with token/ticket and one with the URLs. Process them both
 		   the same and rely on the fact that xmlnode_get_prop() will not *clear*
 		   the variable if no such property is found. */
-		if (!vpninfo->csd_scriptname && xmlnode_is_named(xml_node, vpninfo->csd_xmltag)) {
+		if (!vpninfo->csd_scriptname && xmlnode_is_named(xml_node, csd_tag_name(vpninfo))) {
 			/* ignore the CSD trojan binary on mobile platforms */
-			if (!vpninfo->csd_nostub)
+			if (csd_use_stub(vpninfo))
 				xmlnode_get_prop(xml_node, "stuburl", &vpninfo->csd_stuburl);
 			xmlnode_get_prop(xml_node, "starturl", &vpninfo->csd_starturl);
 			xmlnode_get_prop(xml_node, "waiturl", &vpninfo->csd_waiturl);
@@ -658,8 +682,18 @@ static int handle_auth_form(struct openconnect_info *vpninfo, struct oc_auth_for
 	if (!form->opts) {
 		if (form->message)
 			vpn_progress(vpninfo, PRG_INFO, "%s\n", form->message);
-		if (form->error)
-			vpn_progress(vpninfo, PRG_ERR, "%s\n", form->error);
+		if (form->error) {
+			if (!strcmp(form->error, "Certificate Validation Failure")) {
+				/* XX: Cisco servers send this ambiguous error string when the CLIENT certificate
+				 * is absent or incorrect. We rewrite it to make this clearer, while preserving
+				 * the original error as a substring.
+				 */
+				free(form->error);
+				if (!(form->error = strdup(_("Client certificate missing or incorrect (Certificate Validation Failure)"))))
+					return -ENOMEM;
+			} else
+				vpn_progress(vpninfo, PRG_ERR, "%s\n", form->error);
+		}
 		if (!strcmp(form->auth_id, "openconnect_authentication_complete"))
 			goto justpost;
 		return -EPERM;
@@ -789,23 +823,16 @@ static int xmlpost_initial_req(struct openconnect_info *vpninfo,
 {
 	xmlNodePtr root, node;
 	xmlDocPtr doc = xmlpost_new_query(vpninfo, "init", &root);
-	struct oc_text_buf *url_buf;
+	char *url;
 
 	if (!doc)
 		return -ENOMEM;
 
-	url_buf = buf_alloc();
-	buf_append(url_buf, "https://%s", vpninfo->hostname);
-	if (vpninfo->port != 443)
-		buf_append(url_buf, ":%d", vpninfo->port);
-	/* Do we *need* to omit the trailing / here when no path? */
-	if (vpninfo->urlpath)
-		buf_append(url_buf, "/%s", vpninfo->urlpath);
-
-	if (buf_error(url_buf))
+	url = internal_get_url(vpninfo);
+	if (!url)
 		goto bad;
 
-	node = xmlNewTextChild(root, NULL, XCAST("group-access"), XCAST(url_buf->data));
+	node = xmlNewTextChild(root, NULL, XCAST("group-access"), XCAST(url));
 	if (!node)
 		goto bad;
 	if (cert_fail) {
@@ -818,11 +845,10 @@ static int xmlpost_initial_req(struct openconnect_info *vpninfo,
 		if (!node)
 			goto bad;
 	}
-	buf_free(url_buf);
+	free(url);
 	return xmlpost_complete(doc, request_body);
 
 bad:
-	buf_free(url_buf);
 	xmlpost_complete(doc, NULL);
 	return -ENOMEM;
 }
@@ -910,8 +936,8 @@ static int cstp_can_gen_tokencode(struct openconnect_info *vpninfo,
 #endif
 	/* Otherwise it's an OATH token of some kind. */
 	if (!strcmp(opt->name, "secondary_password") ||
-        (form->auth_id && !strcmp(form->auth_id, "challenge")))
-	  return can_gen_tokencode(vpninfo, form, opt);
+	    (form->auth_id && !strcmp(form->auth_id, "challenge")))
+		return can_gen_tokencode(vpninfo, form, opt);
 
 	return -EINVAL;
 }
@@ -1069,11 +1095,6 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 		return -EPERM;
 	}
 
-#ifndef __linux__
-	vpn_progress(vpninfo, PRG_INFO,
-		     _("Trying to run Linux CSD trojan script.\n"));
-#endif
-
 	fname[0] = 0;
 	if (buflen) {
 		struct oc_vpn_option *opt;
@@ -1123,6 +1144,10 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 		close(fd);
 	}
 
+	vpn_progress(vpninfo, PRG_INFO,
+		     _("Trying to run CSD Trojan script '%s'.\n"),
+		     vpninfo->csd_wrapper ?: fname);
+
 	child = fork();
 	if (child == -1) {
 		goto out;
@@ -1145,6 +1170,10 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Authentication may fail. If your script is not returning zero, fix it.\n"
 					       "Future versions of openconnect will abort on this error.\n"));
+			} else {
+				vpn_progress(vpninfo, PRG_INFO,
+					     _("CSD script '%s' completed successfully.\n"),
+					     vpninfo->csd_wrapper ?: fname);
 			}
 
 			free(vpninfo->urlpath);
@@ -1171,8 +1200,7 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
                 if (set_csd_user(vpninfo) < 0)
                         exit(1);
                 if (getuid() == 0 && !vpninfo->csd_wrapper) {
-                        fprintf(stderr, _("Warning: you are running insecure "
-                                          "CSD code with root privileges\n"
+                        fprintf(stderr, _("Warning: you are running insecure CSD code with root privileges\n"
                                           "\t Use command line option \"--csd-user\"\n"));
                 }
                 /*
@@ -1274,7 +1302,7 @@ int cstp_obtain_cookie(struct openconnect_info *vpninfo)
 	const char *method = "POST";
 	char *orig_host = NULL, *orig_path = NULL, *form_path = NULL;
 	int orig_port = 0;
-	int cert_rq, cert_sent = !vpninfo->cert;
+	int cert_rq, cert_sent = !vpninfo->certinfo[0].cert;
 	int newgroup_attempts = 5;
 
 	if (!vpninfo->xmlpost)
@@ -1334,8 +1362,7 @@ newgroup:
 		}
 
 		request_body_type = vpninfo->xmlpost ? "application/xml; charset=utf-8" : "application/x-www-form-urlencoded";
-		result = do_https_request(vpninfo, method, request_body_type, request_body,
-					  &form_buf, 0);
+		result = do_https_request(vpninfo, method, request_body_type, request_body, &form_buf, NULL, HTTP_NO_FLAGS);
 		if (vpninfo->got_cancel_cmd) {
 			result = 1;
 			goto out;
@@ -1368,10 +1395,10 @@ newgroup:
 			free_auth_form(form);
 			form = NULL;
 
-			if (!cert_sent && vpninfo->cert) {
+			if (!cert_sent && vpninfo->certinfo[0].cert) {
 				/* Try again on a fresh connection. */
 				cert_sent = 1;
-			} else if (cert_sent && vpninfo->cert) {
+			} else if (cert_sent && vpninfo->certinfo[0].cert) {
 				/* Try again with <client-cert-fail/> in the request */
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Server requested SSL client certificate after one was provided\n"));
@@ -1414,7 +1441,7 @@ newgroup:
 			vpninfo->csd_stuburl = NULL;
 			handle_redirect(vpninfo);
 
-			buflen = do_https_request(vpninfo, "GET", NULL, NULL, &form_buf, 0);
+			buflen = do_https_request(vpninfo, "GET", NULL, NULL, &form_buf, NULL, HTTP_NO_FLAGS);
 			if (buflen <= 0) {
 				if (vpninfo->csd_wrapper) {
 					vpn_progress(vpninfo, PRG_ERR,
@@ -1424,7 +1451,10 @@ newgroup:
 					result = -EINVAL;
 					goto out;
 				}
-			}
+			} else
+				vpn_progress(vpninfo, PRG_INFO,
+					     _("Fetched CSD stub for %s platform (size is %d bytes).\n"),
+					     vpninfo->platname, buflen);
 		}
 
 		/* This is the CSD stub script, which we now need to run */
@@ -1434,7 +1464,7 @@ newgroup:
 
 		/* vpninfo->urlpath now points to the wait page */
 		while (1) {
-			result = do_https_request(vpninfo, "GET", NULL, NULL, &form_buf, 0);
+			result = do_https_request(vpninfo, "GET", NULL, NULL, &form_buf, NULL, HTTP_NO_FLAGS);
 			if (result <= 0)
 				break;
 
@@ -1457,7 +1487,7 @@ newgroup:
 
 		result = do_https_request(vpninfo,
 					  vpninfo->xmlpost ? "POST" : "GET",
-					  request_body_type, request_body, &form_buf, 1);
+					  request_body_type, request_body, &form_buf, NULL, HTTP_REDIRECT);
 		if (result < 0)
 			goto out;
 
@@ -1483,8 +1513,7 @@ newgroup:
 			goto newgroup;
 		}
 
-		result = do_https_request(vpninfo, method, request_body_type, request_body,
-					  &form_buf, 1);
+		result = do_https_request(vpninfo, method, request_body_type, request_body, &form_buf, NULL, 1);
 		if (result < 0)
 			goto out;
 
