@@ -17,13 +17,32 @@
 
 #include <config.h>
 
+#include "openconnect-internal.h"
+
 #include <string.h>
 #include <stdarg.h>
 #include <stdlib.h>
 #include <errno.h>
 #include <ctype.h>
+#include <inttypes.h>
 
-#include "openconnect-internal.h"
+#ifdef _WIN32
+#include <sec_api/stdlib_s.h> /* errno_t, size_t */
+#ifndef HAVE_GETENV_S_DECL
+errno_t getenv_s(
+    size_t     *ret_required_buf_size,
+    char       *buf,
+    size_t      buf_size_in_bytes,
+    const char *name
+);
+#endif
+#ifndef HAVE_PUTENV_S_DECL
+errno_t _putenv_s(
+   const char *varname,
+   const char *value_string
+);
+#endif
+#endif
 
 #ifdef HAVE_SUNOS_BROKEN_TIME
 /*
@@ -172,8 +191,8 @@ char *openconnect__strndup(const char *s, size_t n)
 {
 	char *r;
 
-	if (n > strlen(s))
-		n = strlen(s);
+	if ((r = memchr(s, '\0', n)) != NULL)
+		n = r - s;
 
 	r = malloc(n + 1);
 	if (r) {
@@ -184,7 +203,19 @@ char *openconnect__strndup(const char *s, size_t n)
 }
 #endif
 
+#ifndef HAVE_STRCHRNUL
+const char *openconnect__strchrnul(const char *s, int c)
+{
+	while (*s && *s != c) s++;
+	return s;
+}
+#endif
+
 #ifndef HAVE_INET_ATON
+/* XX: unlike "real" inet_aton(), inet_pton() only accepts dotted-decimal notation, not
+ * looser/rarer formats like 32-bit decimal values. For example, inet_aton() accepts both
+ * "127.0.0.1" and "2130706433" as equivalent, but inet_pton() only accepts the former.
+ */
 int openconnect__inet_aton(const char *cp, struct in_addr *addr)
 {
 	return inet_pton(AF_INET, cp, addr);
@@ -192,6 +223,18 @@ int openconnect__inet_aton(const char *cp, struct in_addr *addr)
 #endif
 
 #ifdef _WIN32
+int openconnect__win32_setenv(const char *name, const char *value, int overwrite)
+{
+	/* https://stackoverflow.com/a/23616164 */
+	int errcode = 0;
+	if(!overwrite) {
+		size_t envsize = 0;
+		errcode = getenv_s(&envsize, NULL, 0, name);
+		if(errcode || envsize) return errcode;
+	}
+	return _putenv_s(name, value);
+}
+
 char *openconnect__win32_strerror(DWORD err)
 {
 	wchar_t *msgw;
@@ -226,7 +269,7 @@ char *openconnect__win32_strerror(DWORD err)
 	return msgutf8;
 }
 
-int openconnect__win32_sock_init()
+int openconnect__win32_sock_init(void)
 {
 	WSADATA data;
 	if (WSAStartup (MAKEWORD(1, 1), &data) != 0) {
@@ -259,7 +302,7 @@ int openconnect__win32_inet_pton(int af, const char *src, void *dst)
 	 * inet_aton() (and WSAStringToAddress()) will support, but
 	 * which inet_pton() should not. Not to mention the fact that
 	 * Wine's implementation will even succeed for strings like
-	 * "2001::1" (http://bugs.winehq.org/show_bug.cgi?id=36991) */
+	 * "2001::1" (https://bugs.winehq.org/show_bug.cgi?id=36991) */
 	if (af == AF_INET) {
 		char canon[16];
 		unsigned char *a = (unsigned char *)&sa.s4.sin_addr;
@@ -278,20 +321,20 @@ int openconnect__win32_inet_pton(int af, const char *src, void *dst)
 	}
 }
 
-/* https://github.com/ncm/selectable-socketpair
+/* https://github.com/ncm/selectable-socketpair/blob/master/socketpair.c
 
 Copyright 2007, 2010 by Nathan C. Myers <ncm@cantrip.org>
-Redistribution and use in source and binary forms, with or without modification, 
+Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:
 
     Redistributions of source code must retain the above copyright notice, this
     list of conditions and the following disclaimer.
-    
+
     Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
-    The name of the author must not be used to endorse or promote products derived 
+    The name of the author must not be used to endorse or promote products derived
     from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
@@ -306,13 +349,18 @@ WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  */
 
 /* Changes:
+ * 2014-02-12: merge David Woodhouse, Ger Hobbelt improvements
+ *     git.infradead.org/users/dwmw2/openconnect.git/commitdiff/bdeefa54
+ *     github.com/GerHobbelt/selectable-socketpair
+ *   always init the socks[] to -1/INVALID_SOCKET on error, both on Win32/64
+ *   and UNIX/other platforms
  * 2013-07-18: Change to BSD 3-clause license
  * 2010-03-31:
  *   set addr to 127.0.0.1 because win32 getsockname does not always set it.
  * 2010-02-25:
  *   set SO_REUSEADDR option to avoid leaking some windows resource.
- *   Windows System Error 10049, "Event ID 4226 TCP/IP has reached 
- *   the security limit imposed on the number of concurrent TCP connect 
+ *   Windows System Error 10049, "Event ID 4226 TCP/IP has reached
+ *   the security limit imposed on the number of concurrent TCP connect
  *   attempts."  Bleah.
  * 2007-04-25:
  *   preserve value of WSAGetLastError() on all error returns.
@@ -325,83 +373,166 @@ WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 
 #include <string.h>
 
-# include <winsock2.h>
+# include <ws2tcpip.h>  /* socklen_t, et al (MSVC20xx) */
 # include <windows.h>
 # include <io.h>
 
+#ifdef HAVE_AFUNIX_H
+#include <afunix.h>
+#else
+#define UNIX_PATH_MAX 108
+struct sockaddr_un
+{
+    ADDRESS_FAMILY sun_family;     /* AF_UNIX */
+    char sun_path[UNIX_PATH_MAX];  /* pathname */
+} SOCKADDR_UN, *PSOCKADDR_UN;;
+#endif /* HAS_AFUNIX_H */
+
 /* dumb_socketpair:
  *   If make_overlapped is nonzero, both sockets created will be usable for
  *   "overlapped" operations via WSASend etc.  If make_overlapped is zero,
- *   socks[0] (only) will be usable with regular ReadFile etc., and thus 
+ *   socks[0] (only) will be usable with regular ReadFile etc., and thus
  *   suitable for use as stdin or stdout of a child process.  Note that the
  *   sockets must be closed with closesocket() regardless.
  */
 
-OPENCONNECT_CMD_SOCKET dumb_socketpair(OPENCONNECT_CMD_SOCKET socks[2], int make_overlapped)
+int dumb_socketpair(OPENCONNECT_CMD_SOCKET socks[2], int make_overlapped)
 {
     union {
-       struct sockaddr_in inaddr;
-       struct sockaddr addr;
+        struct sockaddr_un unaddr;
+        struct sockaddr_in inaddr;
+        struct sockaddr addr;
     } a;
     OPENCONNECT_CMD_SOCKET listener;
-    int e;
-    socklen_t addrlen = sizeof(a.inaddr);
+    int e, ii;
+    int domain = AF_UNIX;
+    socklen_t addrlen = sizeof(a.unaddr);
     DWORD flags = (make_overlapped ? WSA_FLAG_OVERLAPPED : 0);
     int reuse = 1;
 
     if (socks == 0) {
-      WSASetLastError(WSAEINVAL);
-      return SOCKET_ERROR;
-    }
-
-    listener = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
-    if (listener == INVALID_SOCKET) 
+        WSASetLastError(WSAEINVAL);
         return SOCKET_ERROR;
-
-    memset(&a, 0, sizeof(a));
-    a.inaddr.sin_family = AF_INET;
-    a.inaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
-    a.inaddr.sin_port = 0; 
-
+    }
     socks[0] = socks[1] = -1;
-    do {
-        if (setsockopt(listener, SOL_SOCKET, SO_REUSEADDR, 
-               (char*) &reuse, (socklen_t) sizeof(reuse)) == -1)
-            break;
-        if  (bind(listener, &a.addr, sizeof(a.inaddr)) == SOCKET_ERROR)
-            break;
+
+    /* AF_UNIX/SOCK_STREAM became available in Windows 10
+     * ( https://devblogs.microsoft.com/commandline/af_unix-comes-to-windows )
+     *
+     * We will attempt to use AF_UNIX, but fallback to using AF_INET if
+     * setting up AF_UNIX socket fails in any other way, which it surely will
+     * on earlier versions of Windows.
+     */
+    for (ii = 0; ii < 2; ii++) {
+        listener = socket(domain, SOCK_STREAM, domain == AF_INET ? IPPROTO_TCP : 0);
+        if (listener == INVALID_SOCKET)
+            goto fallback;
 
         memset(&a, 0, sizeof(a));
-        if  (getsockname(listener, &a.addr, &addrlen) == SOCKET_ERROR)
-            break;
-        // win32 getsockname may only set the port number, p=0.0005.
-        // ( http://msdn.microsoft.com/library/ms738543.aspx ):
-        a.inaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
-        a.inaddr.sin_family = AF_INET;
+        if (domain == AF_UNIX) {
+            /* XX: Abstract sockets (filesystem-independent) don't work, contrary to
+             * the claims of the aforementioned blog post:
+             * https://github.com/microsoft/WSL/issues/4240#issuecomment-549663217
+             *
+             * So we must use a named path, and that comes with all the attendant
+             * problems of permissions and collisions. Trying various temporary
+             * directories and putting high-res time and PID in the filename, that
+             * seems like a less-bad option.
+             */
+            LARGE_INTEGER ticks;
+            DWORD n;
+            int bind_try = 0;
+
+            for (;;) {
+                switch (bind_try++) {
+                case 0:
+                    /* "The returned string ends with a backslash" */
+                    n = GetTempPath(UNIX_PATH_MAX, a.unaddr.sun_path);
+                    break;
+                case 1:
+                    /* Heckuva job with API consistency, Microsoft! Reversed argument order, and
+                     * "This path does not end with a backslash unless the Windows directory is the root directory.."
+                     */
+                    n = GetWindowsDirectory(a.unaddr.sun_path, UNIX_PATH_MAX);
+                    n += snprintf(a.unaddr.sun_path + n, UNIX_PATH_MAX - n, "\\Temp\\");
+                    break;
+                case 2:
+                    n = snprintf(a.unaddr.sun_path, UNIX_PATH_MAX, "C:\\Temp\\");
+                    break;
+                case 3:
+                    n = 0; /* Current directory */
+                    break;
+                case 4:
+                    goto fallback;
+                }
+
+                /* GetTempFileName could be used here.
+                 * (https://docs.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-gettempfilenamea)
+                 * However it only adds 16 bits of time-based random bits,
+                 * fails if there isn't room for a 14-character filename, and
+                 * seems to offers no other apparent advantages. So we will
+                 * use high-res timer ticks and PID for filename.
+                 */
+                QueryPerformanceCounter(&ticks);
+                snprintf(a.unaddr.sun_path + n, UNIX_PATH_MAX - n,
+                         "%"PRIx64"-%"PRId32".$$$", ticks.QuadPart, GetCurrentProcessId());
+                a.unaddr.sun_family = AF_UNIX;
+
+                if (bind(listener, &a.addr, addrlen) != SOCKET_ERROR)
+                    break;
+            }
+        } else {
+            a.inaddr.sin_family = AF_INET;
+            a.inaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
+            a.inaddr.sin_port = 0;
+
+            if (setsockopt(listener, SOL_SOCKET, SO_REUSEADDR,
+                           (char *) &reuse, (socklen_t) sizeof(reuse)) == -1)
+                goto fallback;;
+
+            if (bind(listener, &a.addr, addrlen) == SOCKET_ERROR)
+                goto fallback;
+
+            memset(&a, 0, sizeof(a));
+            if (getsockname(listener, &a.addr, &addrlen) == SOCKET_ERROR)
+                goto fallback;
+
+            // win32 getsockname may only set the port number, p=0.0005.
+            // ( https://docs.microsoft.com/windows/win32/api/winsock/nf-winsock-getsockname ):
+            a.inaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
+            a.inaddr.sin_family = AF_INET;
+        }
 
         if (listen(listener, 1) == SOCKET_ERROR)
-            break;
+            goto fallback;
 
-        socks[0] = WSASocket(AF_INET, SOCK_STREAM, 0, NULL, 0, flags);
+        socks[0] = WSASocket(domain, SOCK_STREAM, 0, NULL, 0, flags);
         if (socks[0] == INVALID_SOCKET)
-            break;
-        if (connect(socks[0], &a.addr, sizeof(a.inaddr)) == SOCKET_ERROR)
-            break;
+            goto fallback;
+        if (connect(socks[0], &a.addr, addrlen) == SOCKET_ERROR)
+            goto fallback;
+        if (domain == AF_UNIX)
+            DeleteFile(a.unaddr.sun_path);  // Socket file no longer needed
 
         socks[1] = accept(listener, NULL, NULL);
         if (socks[1] == INVALID_SOCKET)
-            break;
+            goto fallback;
 
         closesocket(listener);
         return 0;
 
-    } while (0);
+    fallback:
+        domain = AF_INET;
+        addrlen = sizeof(a.inaddr);
 
-    e = WSAGetLastError();
-    closesocket(listener);
-    closesocket(socks[0]);
-    closesocket(socks[1]);
-    WSASetLastError(e);
+        e = WSAGetLastError();
+        closesocket(listener);
+        closesocket(socks[0]);
+        closesocket(socks[1]);
+        WSASetLastError(e);
+    }
+
+    socks[0] = socks[1] = -1;
     return SOCKET_ERROR;
 }
-#endif /* _WIN32 */
+#endif
