@@ -18,16 +18,8 @@
 
 #include <config.h>
 
-#include <unistd.h>
-#include <fcntl.h>
-#include <time.h>
-#include <string.h>
-#include <ctype.h>
-#include <errno.h>
-#include <stdlib.h>
-#include <stdio.h>
-#include <sys/types.h>
-#include <stdarg.h>
+#include "openconnect-internal.h"
+
 #ifdef HAVE_LZ4
 #include <lz4.h>
 #ifndef HAVE_LZ4_COMPRESS_DEFAULT
@@ -35,12 +27,21 @@
 #endif
 #endif
 
+#include <unistd.h>
+#include <fcntl.h>
+#include <sys/types.h>
 #if defined(__linux__)
 /* For TCP_INFO */
 # include <linux/tcp.h>
 #endif
 
-#include "openconnect-internal.h"
+#include <time.h>
+#include <string.h>
+#include <ctype.h>
+#include <errno.h>
+#include <stdlib.h>
+#include <stdio.h>
+#include <stdarg.h>
 
 /*
  * Data packets are encapsulated in the SSL stream as follows:
@@ -96,7 +97,7 @@ static const struct pkt dpd_resp_pkt = {
  * If we don't even have TCP_MAXSEG, then default to sending a legacy MTU of
  * 1406 which is what we always used to do.
  */
-static void calculate_mtu(struct openconnect_info *vpninfo, int *base_mtu, int *mtu)
+static void calculate_dtls_mtu(struct openconnect_info *vpninfo, int *base_mtu, int *mtu)
 {
 	*mtu = vpninfo->reqmtu;
 	*base_mtu = vpninfo->basemtu;
@@ -157,19 +158,16 @@ static void calculate_mtu(struct openconnect_info *vpninfo, int *base_mtu, int *
 static void append_compr_types(struct oc_text_buf *buf, const char *proto, int avail)
 {
 	if (avail) {
-		char sep = ' ';
+		const char sep = ',';
 		buf_append(buf, "X-%s-Accept-Encoding:", proto);
 		if (avail & COMPR_LZ4) {
 			buf_append(buf, "%coc-lz4", sep);
-			sep = ',';
 		}
 		if (avail & COMPR_LZS) {
 			buf_append(buf, "%clzs", sep);
-			sep = ',';
 		}
 		if (avail & COMPR_DEFLATE) {
 			buf_append(buf, "%cdeflate", sep);
-			sep = ',';
 		}
 		buf_append(buf, "\r\n");
 	}
@@ -218,29 +216,13 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	int i;
 	int dtls_secret_set = 0;
 	int retried = 0, sessid_found = 0;
-	struct oc_vpn_option **next_dtls_option = &vpninfo->dtls_options;
-	struct oc_vpn_option **next_cstp_option = &vpninfo->cstp_options;
-	struct oc_vpn_option *old_cstp_opts = vpninfo->cstp_options;
-	struct oc_vpn_option *old_dtls_opts = vpninfo->dtls_options;
 	const char *old_addr = vpninfo->ip_info.addr;
-	const char *old_netmask = vpninfo->ip_info.netmask;
 	const char *old_addr6 = vpninfo->ip_info.addr6;
-	const char *old_netmask6 = vpninfo->ip_info.netmask6;
+	const char *banner = NULL;
 	int base_mtu = 0, mtu = 0;
 
-	/* Clear old options which will be overwritten */
-	vpninfo->ip_info.addr = vpninfo->ip_info.netmask = NULL;
-	vpninfo->ip_info.addr6 = vpninfo->ip_info.netmask6 = NULL;
-	vpninfo->cstp_options = vpninfo->dtls_options = NULL;
-	vpninfo->ip_info.domain = vpninfo->ip_info.proxy_pac = NULL;
-	vpninfo->banner = NULL;
-
-	for (i = 0; i < 3; i++)
-		vpninfo->ip_info.dns[i] = vpninfo->ip_info.nbns[i] = NULL;
-	free_split_routes(vpninfo);
-
  retry:
-	calculate_mtu(vpninfo, &base_mtu, &mtu);
+	calculate_dtls_mtu(vpninfo, &base_mtu, &mtu);
 	vpninfo->cstp_basemtu = base_mtu;
 
 	reqbuf = buf_alloc();
@@ -262,11 +244,21 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 		buf_append(reqbuf, "X-CSTP-MTU: %d\r\n", mtu);
 	buf_append(reqbuf, "X-CSTP-Address-Type: %s\r\n",
 			       vpninfo->disable_ipv6 ? "IPv4" : "IPv6,IPv4");
-	if (!vpninfo->disable_ipv6)
+	/* Explicitly request the same IPv4 and IPv6 addresses on reconnect
+	 *
+	 * XX: It's not clear which Cisco servers attempt to follow specific
+	 * IP address requests from the X-CSTP-Address headers in the CONNECT
+	 * request; most seem to ignore it. */
+	if (old_addr)
+		buf_append(reqbuf, "X-CSTP-Address: %s\r\n", old_addr);
+	if (!vpninfo->disable_ipv6) {
 		buf_append(reqbuf, "X-CSTP-Full-IPv6-Capability: true\r\n");
+		if (old_addr6)
+			buf_append(reqbuf, "X-CSTP-Address: %s\r\n", old_addr6);
+	}
 #ifdef HAVE_DTLS
 	if (vpninfo->dtls_state != DTLS_DISABLED) {
-		/* The X-DTLS-Master-Secret is only used for the legacy protocol negotation
+		/* The X-DTLS-Master-Secret is only used for the legacy protocol negotiation
 		 * which required the client to send explicitly the secret. In the PSK-NEGOTIATE
 		 * method, the master secret is implicitly agreed on */
 		buf_append(reqbuf, "X-DTLS-Master-Secret: ");
@@ -303,6 +295,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			buf_free(dtls12_cl);
 		}
 		append_compr_types(reqbuf, "DTLS", vpninfo->req_compr & ~COMPR_DEFLATE);
+
+		vpninfo->delay_tunnel_reason = "DTLS MTU detection";
 	}
 #endif
 	buf_append(reqbuf, "\r\n");
@@ -345,7 +339,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 		if (!strncmp(buf, "HTTP/1.1 503 ", 13)) {
 			/* "Service Unavailable. Why? */
 			const char *reason = "<unknown>";
-			while ((i = vpninfo->ssl_gets(vpninfo, buf, sizeof(buf)))) {
+			while ((vpninfo->ssl_gets(vpninfo, buf, sizeof(buf)))) {
 				if (!strncmp(buf, "X-Reason: ", 10)) {
 					reason = buf + 10;
 					break;
@@ -370,12 +364,19 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	vpninfo->cstp_compr = vpninfo->dtls_compr = 0;
 	mtu = 0;
 
-	while ((i = vpninfo->ssl_gets(vpninfo, buf, sizeof(buf)))) {
+	struct oc_vpn_option *new_cstp_opts = NULL;
+	struct oc_vpn_option *new_dtls_opts = NULL;
+	struct oc_vpn_option **next_dtls_option = &new_dtls_opts;
+	struct oc_vpn_option **next_cstp_option = &new_cstp_opts;
+	struct oc_ip_info new_ip_info = {};
+	int ret = 0;
+
+	while ((ret = vpninfo->ssl_gets(vpninfo, buf, sizeof(buf)))) {
 		struct oc_vpn_option *new_option;
 		char *colon;
 
-		if (i < 0)
-			return i;
+		if (ret < 0)
+			goto err;
 
 		colon = strchr(buf, ':');
 		if (!colon)
@@ -394,7 +395,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 		new_option = malloc(sizeof(*new_option));
 		if (!new_option) {
 			vpn_progress(vpninfo, PRG_ERR, _("No memory for options\n"));
-			return -ENOMEM;
+			ret = -ENOMEM;
+			goto err;
 		}
 		new_option->option = strdup(buf);
 		new_option->value = strdup(colon);
@@ -405,7 +407,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			free(new_option->option);
 			free(new_option->value);
 			free(new_option);
-			return -ENOMEM;
+			ret = -ENOMEM;
+			goto err;
 		}
 
 		/* This contains the whole document, including the webvpn cookie. */
@@ -430,9 +433,10 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 				if (vsize != 32) {
 					vpn_progress(vpninfo, PRG_ERR,
 					     _("X-DTLS-Session-ID not 64 characters; is: \"%s\"\n"),
-					     	colon);
+					     colon);
 					vpninfo->dtls_attempt_period = 0;
-					return -EINVAL;
+					ret = -EINVAL;
+					goto err;
 				}
 
 				sessid_found = 1;
@@ -447,9 +451,10 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 				if (vsize <= 0) {
 					vpn_progress(vpninfo, PRG_ERR,
 					     _("X-DTLS-Session-ID is invalid; is: \"%s\"\n"),
-					     	colon);
+					     colon);
 					vpninfo->dtls_attempt_period = 0;
-					return -EINVAL;
+					ret = -EINVAL;
+					goto err;
 				}
 
 				vpninfo->dtls_app_id_size = vsize;
@@ -466,13 +471,34 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 					vpn_progress(vpninfo, PRG_ERR,
 						     _("Unknown DTLS-Content-Encoding %s\n"),
 						     colon);
-					return -EINVAL;
+					ret = -EINVAL;
+					goto err;
 				}
 			} else if (!strcmp(buf + i, "CipherSuite")) {
 				/* Remember if it came from a 'X-DTLS12-CipherSuite:' header */
-				vpninfo->cisco_dtls12 = (i == 9);
+				vpninfo->dtls12 = (i == 9);
 				vpninfo->dtls_cipher = strdup(colon);
+			} else if (!strcmp(buf + i, "Port")) {
+				int dtls_port = atol(colon);
+				if (dtls_port)
+					udp_sockaddr(vpninfo, dtls_port);
+			} else if (!strcmp(buf + i, "Keepalive")) {
+				vpninfo->dtls_times.keepalive = atol(colon);
+			} else if (!strcmp(buf + i, "DPD")) {
+				int j = atol(colon);
+				if (j && !vpninfo->dtls_times.dpd)
+					vpninfo->dtls_times.dpd = j;
+			} else if (!strcmp(buf + i, "Rekey-Method")) {
+				if (!strcmp(colon, "new-tunnel"))
+					vpninfo->dtls_times.rekey_method = REKEY_TUNNEL;
+				else if (!strcmp(colon, "ssl"))
+					vpninfo->dtls_times.rekey_method = REKEY_SSL;
+				else
+					vpninfo->dtls_times.rekey_method = REKEY_NONE;
+			} else if (!strcmp(buf + i, "Rekey-Time")) {
+				vpninfo->dtls_times.rekey = atol(colon);
 			}
+
 			continue;
 		}
 		/* CSTP options... */
@@ -482,11 +508,23 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 
 		if (!strcmp(buf + 7, "Keepalive")) {
 			vpninfo->ssl_times.keepalive = atol(colon);
+		} else if (!strcmp(buf + 7, "Lease-Duration") ||
+			   !strcmp(buf + 7, "Session-Timeout") ||
+			   !strcmp(buf + 7, "Session-Timeout-Remaining")) {
+
+			/* XX: Distinction between Lease-Duration and Session-Timeout is rather unclear. Cisco doc:
+			 * https://www.cisco.com/assets/sol/sb/RV345P_Emulators/RV345P_Emulator_v1-0-01-17/help/help/t_SSL_VPN.html
+			 * Empirically, it appears that the best behavior is to accept whichever of these headers has the
+			 * lowest non-zero value.
+			 */
+			long j = atol(colon);
+			if (j && (!vpninfo->auth_expiration || j < vpninfo->auth_expiration))
+				vpninfo->auth_expiration = time(NULL) + j;
 		} else if (!strcmp(buf + 7, "Idle-Timeout")) {
 			vpninfo->idle_timeout = atol(colon);
 		} else if (!strcmp(buf + 7, "DPD")) {
 			int j = atol(colon);
-			if (j && (!vpninfo->ssl_times.dpd || j < vpninfo->ssl_times.dpd))
+			if (j && !vpninfo->ssl_times.dpd)
 				vpninfo->ssl_times.dpd = j;
 		} else if (!strcmp(buf + 7, "Rekey-Time")) {
 			vpninfo->ssl_times.rekey = atol(colon);
@@ -508,7 +546,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Unknown CSTP-Content-Encoding %s\n"),
 					     colon);
-				return -EINVAL;
+				ret= -EINVAL;
+				goto err;
 			}
 		} else if (!strcmp(buf + 7, "Base-MTU")) {
 			vpninfo->cstp_basemtu = atol(colon);
@@ -520,121 +559,89 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			if (!strcmp(colon, "true"))
 				vpninfo->is_dyndns = 1;
 		} else if (!strcmp(buf + 7, "Address-IP6")) {
-			vpninfo->ip_info.netmask6 = new_option->value;
+			new_ip_info.netmask6 = new_option->value;
 		} else if (!strcmp(buf + 7, "Address")) {
 			if (strchr(new_option->value, ':')) {
 				if (!vpninfo->disable_ipv6)
-					vpninfo->ip_info.addr6 = new_option->value;
+					new_ip_info.addr6 = new_option->value;
 			} else
-				vpninfo->ip_info.addr = new_option->value;
+				new_ip_info.addr = new_option->value;
 		} else if (!strcmp(buf + 7, "Netmask")) {
 			if (strchr(new_option->value, ':')) {
 				if (!vpninfo->disable_ipv6)
-					vpninfo->ip_info.netmask6 = new_option->value;
+					new_ip_info.netmask6 = new_option->value;
 			} else
-				vpninfo->ip_info.netmask = new_option->value;
+				new_ip_info.netmask = new_option->value;
 		} else if (!strcmp(buf + 7, "DNS") ||
 			   !strcmp(buf + 7, "DNS-IP6")) {
 			int j;
 			for (j = 0; j < 3; j++) {
-				if (!vpninfo->ip_info.dns[j]) {
-					vpninfo->ip_info.dns[j] = new_option->value;
+				if (!new_ip_info.dns[j]) {
+					new_ip_info.dns[j] = new_option->value;
 					break;
 				}
 			}
 		} else if (!strcmp(buf + 7, "NBNS")) {
 			int j;
 			for (j = 0; j < 3; j++) {
-				if (!vpninfo->ip_info.nbns[j]) {
-					vpninfo->ip_info.nbns[j] = new_option->value;
+				if (!new_ip_info.nbns[j]) {
+					new_ip_info.nbns[j] = new_option->value;
 					break;
 				}
 			}
 		} else if (!strcmp(buf + 7, "Default-Domain")) {
-			vpninfo->ip_info.domain = new_option->value;
+			new_ip_info.domain = new_option->value;
 		} else if (!strcmp(buf + 7, "MSIE-Proxy-PAC-URL")) {
-			vpninfo->ip_info.proxy_pac = new_option->value;
+			new_ip_info.proxy_pac = new_option->value;
 		} else if (!strcmp(buf + 7, "Banner")) {
-			vpninfo->banner = new_option->value;
+			banner = new_option->value;
 		} else if (!strcmp(buf + 7, "Split-DNS")) {
 			struct oc_split_include *dns = malloc(sizeof(*dns));
 			if (!dns)
 				continue;
 			dns->route = new_option->value;
-			dns->next = vpninfo->ip_info.split_dns;
-			vpninfo->ip_info.split_dns = dns;
+			dns->next = new_ip_info.split_dns;
+			new_ip_info.split_dns = dns;
 		} else if (!strcmp(buf + 7, "Split-Include") || !strcmp(buf + 7, "Split-Include-IP6")) {
 			struct oc_split_include *inc = malloc(sizeof(*inc));
 			if (!inc)
 				continue;
 			inc->route = new_option->value;
-			inc->next = vpninfo->ip_info.split_includes;
-			vpninfo->ip_info.split_includes = inc;
+			inc->next = new_ip_info.split_includes;
+			new_ip_info.split_includes = inc;
 		} else if (!strcmp(buf + 7, "Split-Exclude") || !strcmp(buf + 7, "Split-Exclude-IP6")) {
 			struct oc_split_include *exc = malloc(sizeof(*exc));
 			if (!exc)
 				continue;
 			exc->route = new_option->value;
-			exc->next = vpninfo->ip_info.split_excludes;
-			vpninfo->ip_info.split_excludes = exc;
+			exc->next = new_ip_info.split_excludes;
+			new_ip_info.split_excludes = exc;
 		}
 	}
 
 	if (!mtu) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("No MTU received. Aborting\n"));
-		return -EINVAL;
+		ret = -EINVAL;
+		goto err;
 	}
-	vpninfo->ip_info.mtu = mtu;
+	new_ip_info.mtu = mtu;
 
-	if (!vpninfo->ip_info.addr && !vpninfo->ip_info.addr6 &&
-	    !vpninfo->ip_info.netmask6) {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("No IP address received. Aborting\n"));
-		return -EINVAL;
-	}
-	if (mtu < 1280 &&
-	    (vpninfo->ip_info.addr6 || vpninfo->ip_info.netmask6)) {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("IPv6 configuration received but MTU %d is too small.\n"),
-			     mtu);
-	}
-	if (old_addr) {
-		if (!vpninfo->ip_info.addr || strcmp(old_addr, vpninfo->ip_info.addr)) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Reconnect gave different Legacy IP address (%s != %s)\n"),
-				     vpninfo->ip_info.addr, old_addr);
-			/* EPERM means that the retry loop will abort and won't keep trying. */
-			return -EPERM;
-		}
-	}
-	if (old_netmask) {
-		if (!vpninfo->ip_info.netmask || strcmp(old_netmask, vpninfo->ip_info.netmask)) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Reconnect gave different Legacy IP netmask (%s != %s)\n"),
-				     vpninfo->ip_info.netmask, old_netmask);
-			return -EPERM;
-		}
-	}
-	if (old_addr6) {
-		if (!vpninfo->ip_info.addr6 || strcmp(old_addr6, vpninfo->ip_info.addr6)) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Reconnect gave different IPv6 address (%s != %s)\n"),
-				     vpninfo->ip_info.addr6, old_addr6);
-			return -EPERM;
-		}
-	}
-	if (old_netmask6) {
-		if (!vpninfo->ip_info.netmask6 || strcmp(old_netmask6, vpninfo->ip_info.netmask6)) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Reconnect gave different IPv6 netmask (%s != %s)\n"),
-				     vpninfo->ip_info.netmask6, old_netmask6);
-			return -EPERM;
-		}
+	ret = install_vpn_opts(vpninfo, new_cstp_opts, &new_ip_info);
+	if (ret) {
+	err:
+		free_optlist(new_cstp_opts);
+		free_optlist(new_dtls_opts);
+		free_split_routes(&new_ip_info);
+		return ret;
 	}
 
-	free_optlist(old_dtls_opts);
-	free_optlist(old_cstp_opts);
+	/* DTLS opts are a special case. Perhaps should have been in the
+	 * CSTP opts list anyway. This all very Cisco-specific. */
+	free_optlist(vpninfo->dtls_options);
+	vpninfo->dtls_options = new_dtls_opts;
+	vpninfo->banner = banner;
+
 	vpn_progress(vpninfo, PRG_INFO, _("CSTP connected. DPD %d, Keepalive %d\n"),
 		     vpninfo->ssl_times.dpd, vpninfo->ssl_times.keepalive);
 
@@ -709,8 +716,8 @@ int cstp_connect(struct openconnect_info *vpninfo)
 
 	/* If *any* compression is enabled, we'll need a deflate_pkt to compress into */
 	if (deflate_bufsize > vpninfo->deflate_pkt_size) {
-		free(vpninfo->deflate_pkt);
-		vpninfo->deflate_pkt = malloc(sizeof(struct pkt) + deflate_bufsize);
+		free_pkt(vpninfo, vpninfo->deflate_pkt);
+		vpninfo->deflate_pkt = alloc_pkt(vpninfo, deflate_bufsize);
 		if (!vpninfo->deflate_pkt) {
 			vpninfo->deflate_pkt_size = 0;
 			vpn_progress(vpninfo, PRG_ERR,
@@ -755,7 +762,7 @@ int decompress_and_queue_packet(struct openconnect_info *vpninfo, int compr_type
 	   negotiated MTU after decompression. We reserve some extra
 	   space to handle that */
 	int receive_mtu = MAX(16384, vpninfo->ip_info.mtu);
-	struct pkt *new = malloc(sizeof(struct pkt) + receive_mtu);
+	struct pkt *new = alloc_pkt(vpninfo, receive_mtu);
 	const char *comprname = "";
 
 	if (!new)
@@ -876,7 +883,7 @@ int compress_packet(struct openconnect_info *vpninfo, int compr_type, struct pkt
 		if (this->len < 40)
 			return -EFBIG;
 
-		ret = LZ4_compress_default((void*)this->data, (void*)vpninfo->deflate_pkt->data,
+		ret = LZ4_compress_default((void *)this->data, (void *)vpninfo->deflate_pkt->data,
 					   this->len, this->len);
 		if (ret <= 0) {
 			if (ret == 0)
@@ -915,14 +922,14 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 		int len, payload_len;
 
 		if (!vpninfo->cstp_pkt) {
-			vpninfo->cstp_pkt = malloc(sizeof(struct pkt) + receive_mtu);
+			vpninfo->cstp_pkt = alloc_pkt(vpninfo, receive_mtu);
 			if (!vpninfo->cstp_pkt) {
 				vpn_progress(vpninfo, PRG_ERR, _("Allocation failed\n"));
 				break;
 			}
 		}
 
-		len = ssl_nonblock_read(vpninfo, vpninfo->cstp_pkt->cstp.hdr, receive_mtu + 8);
+		len = ssl_nonblock_read(vpninfo, 0, vpninfo->cstp_pkt->cstp.hdr, receive_mtu + 8);
 		if (!len)
 			break;
 		if (len < 0)
@@ -1033,7 +1040,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 		vpninfo->ssl_times.last_tx = time(NULL);
 		unmonitor_write_fd(vpninfo, ssl);
 
-		ret = ssl_nonblock_write(vpninfo,
+		ret = ssl_nonblock_write(vpninfo, 0,
 					 vpninfo->current_ssl_pkt->cstp.hdr,
 					 vpninfo->current_ssl_pkt->len + 8);
 		if (ret < 0)
@@ -1052,7 +1059,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 				return work_done;
 			default:
 				/* This should never happen */
-				;
+				break;
 			}
 		}
 
@@ -1065,12 +1072,12 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 		}
 		/* Don't free the 'special' packets */
 		if (vpninfo->current_ssl_pkt == vpninfo->deflate_pkt) {
-			free(vpninfo->pending_deflated_pkt);
+			free_pkt(vpninfo, vpninfo->pending_deflated_pkt);
 			vpninfo->pending_deflated_pkt = NULL;
 		} else if (vpninfo->current_ssl_pkt != &dpd_pkt &&
 			 vpninfo->current_ssl_pkt != &dpd_resp_pkt &&
 			 vpninfo->current_ssl_pkt != &keepalive_pkt)
-			free(vpninfo->current_ssl_pkt);
+			free_pkt(vpninfo, vpninfo->current_ssl_pkt);
 
 		vpninfo->current_ssl_pkt = NULL;
 	}
@@ -1214,7 +1221,7 @@ int cstp_bye(struct openconnect_info *vpninfo, const char *reason)
 	vpn_progress(vpninfo, PRG_INFO,
 		     _("Send BYE packet: %s\n"), reason);
 
-	ret = ssl_nonblock_write(vpninfo, bye_pkt, reason_len + 9);
+	ret = ssl_nonblock_write(vpninfo, 0, bye_pkt, reason_len + 9);
 	if (ret == reason_len + 9) {
 		ret = 0;
 	} else if (ret >= 0) {
@@ -1234,11 +1241,8 @@ void cstp_common_headers(struct openconnect_info *vpninfo, struct oc_text_buf *b
 	buf_append(buf, "Accept: */*\r\n");
 	buf_append(buf, "Accept-Encoding: identity\r\n");
 	buf_append(buf, "X-Transcend-Version: 1\r\n");
-	if (vpninfo->xmlpost) {
+	if (vpninfo->xmlpost)
 		buf_append(buf, "X-Aggregate-Auth: 1\r\n");
-		buf_append(buf, "X-AnyConnect-Platform: %s\r\n",
-			   vpninfo->platname);
-	}
 	if (vpninfo->try_http_auth)
 		buf_append(buf, "X-Support-HTTP-Auth: true\r\n");
 
