@@ -48,19 +48,22 @@
  * THE POSSIBILITY OF SUCH DAMAGE.
  ******************************************************************************/
 
-#include "config.h"
+#include <config.h>
 
 #include "openconnect-internal.h"
+
 #include "gnutls.h"
 
+#include <tss2/tss2_mu.h>
+#include <tss2/tss2_esys.h>
+#include <tss2/tss2_tctildr.h>
+
 #include <errno.h>
 #include <stdio.h>
 #include <string.h>
 
-#include <tss2/tss2_mu.h>
-#include <tss2/tss2_esys.h>
-
 struct oc_tpm2_ctx {
+	TSS2_TCTI_CONTEXT *tcti_ctx;
 	TPM2B_PUBLIC pub;
 	TPM2B_PRIVATE priv;
 	TPM2B_DIGEST userauth;
@@ -72,7 +75,7 @@ struct oc_tpm2_ctx {
 	unsigned int parent;
 };
 
-static TPM2B_PUBLIC primaryTemplate = {
+static const TPM2B_PUBLIC primaryTemplate = {
 	.publicArea = {
 		.type = TPM2_ALG_ECC,
 		.nameAlg = TPM2_ALG_SHA256,
@@ -109,7 +112,7 @@ static TPM2B_PUBLIC primaryTemplate = {
 	}
 };
 
-static TPM2B_PUBLIC primaryTemplate_legacy = {
+static const TPM2B_PUBLIC primaryTemplate_legacy = {
 	.publicArea = {
 		.type = TPM2_ALG_ECC,
 		.nameAlg = TPM2_ALG_SHA256,
@@ -144,7 +147,7 @@ static TPM2B_PUBLIC primaryTemplate_legacy = {
 	}
 };
 
-static TPM2B_SENSITIVE_CREATE primarySensitive = {
+static const TPM2B_SENSITIVE_CREATE primarySensitive = {
 	.sensitive = {
 		.userAuth = {
 			.size = 0,
@@ -154,17 +157,16 @@ static TPM2B_SENSITIVE_CREATE primarySensitive = {
 		}
 	}
 };
-static TPM2B_DATA allOutsideInfo = {
+static const TPM2B_DATA allOutsideInfo = {
 	.size = 0,
 };
-static TPML_PCR_SELECTION allCreationPCR = {
+static const TPML_PCR_SELECTION allCreationPCR = {
 	.count = 0,
 };
 
 
-/* Where do these error values come from? */
-#define KEY_AUTH_FAILED		0x9a2
-#define PARENT_AUTH_FAILED	0x98e
+#define rc_is_key_auth_failed(rc) (((rc) & 0xff) == TPM2_RC_BAD_AUTH)
+#define rc_is_parent_auth_failed(rc) (((rc) & 0xff) == TPM2_RC_AUTH_FAIL)
 
 static void install_tpm_passphrase(struct openconnect_info *vpninfo, TPM2B_DIGEST *auth, char *pass)
 {
@@ -182,14 +184,14 @@ static void install_tpm_passphrase(struct openconnect_info *vpninfo, TPM2B_DIGES
 	free_pass(&pass);
 }
 
-static int init_tpm2_primary(struct openconnect_info *vpninfo,
+static int init_tpm2_primary(struct openconnect_info *vpninfo, struct cert_info *certinfo,
 			     ESYS_CONTEXT *ctx, ESYS_TR *primaryHandle)
 {
 	TSS2_RC r;
 	const char *hierarchy_name;
 	ESYS_TR hierarchy;
 
-	switch(vpninfo->tpm2->parent) {
+	switch(certinfo->tpm2->parent) {
 	case TPM2_RH_OWNER:	hierarchy = ESYS_TR_RH_OWNER;	hierarchy_name = _("owner"); break;
 	case TPM2_RH_NULL:	hierarchy = ESYS_TR_RH_NULL;	hierarchy_name = _("null"); break;
 	case TPM2_RH_ENDORSEMENT:hierarchy = ESYS_TR_RH_ENDORSEMENT; hierarchy_name = _("endorsement"); break;
@@ -199,15 +201,18 @@ static int init_tpm2_primary(struct openconnect_info *vpninfo,
 
 	vpn_progress(vpninfo, PRG_DEBUG, _("Creating primary key under %s hierarchy.\n"), hierarchy_name);
  reauth:
-	if (vpninfo->tpm2->need_ownerauth) {
+	if (certinfo->tpm2->need_ownerauth) {
 		char *pass = NULL;
-		if (request_passphrase(vpninfo, "openconnect_tpm2_hierarchy", &pass,
-					   _("Enter TPM2 %s hierarchy password:"), hierarchy_name))
+		if (request_passphrase(vpninfo,
+				       certinfo_string(certinfo, "openconnect_tpm2_hierarchy",
+						       "openconnect_secondary_tpm2_hierarchy"),
+				       &pass,
+				       _("Enter TPM2 %s hierarchy password:"), hierarchy_name))
 			return -EPERM;
-		install_tpm_passphrase(vpninfo, &vpninfo->tpm2->ownerauth, pass);
-		vpninfo->tpm2->need_ownerauth = 0;
+		install_tpm_passphrase(vpninfo, &certinfo->tpm2->ownerauth, pass);
+		certinfo->tpm2->need_ownerauth = 0;
 	}
-	r = Esys_TR_SetAuth(ctx, hierarchy, &vpninfo->tpm2->ownerauth);
+	r = Esys_TR_SetAuth(ctx, hierarchy, &certinfo->tpm2->ownerauth);
 	if (r) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("TPM2 Esys_TR_SetAuth failed: 0x%x\n"),
@@ -217,13 +222,13 @@ static int init_tpm2_primary(struct openconnect_info *vpninfo,
 	r = Esys_CreatePrimary(ctx, hierarchy,
 			       ESYS_TR_PASSWORD, ESYS_TR_NONE, ESYS_TR_NONE,
 			       &primarySensitive,
-			       vpninfo->tpm2->legacy_srk ? &primaryTemplate_legacy : &primaryTemplate,
+			       certinfo->tpm2->legacy_srk ? &primaryTemplate_legacy : &primaryTemplate,
 			       &allOutsideInfo, &allCreationPCR,
 			       primaryHandle, NULL, NULL, NULL, NULL);
-	if (r == KEY_AUTH_FAILED) {
+	if (rc_is_key_auth_failed(r)) {
 		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("TPM2 Esys_CreatePrimary owner auth failed\n"));
-		vpninfo->tpm2->need_ownerauth = 1;
+		certinfo->tpm2->need_ownerauth = 1;
 		goto reauth;
 	} else if (r) {
 		vpn_progress(vpninfo, PRG_ERR,
@@ -238,7 +243,7 @@ static int init_tpm2_primary(struct openconnect_info *vpninfo,
 #define parent_is_persistent(parent) ((parent) >> TPM2_HR_SHIFT == TPM2_HT_PERSISTENT)
 
 static int init_tpm2_key(ESYS_CONTEXT **ctx, ESYS_TR *keyHandle,
-			 struct openconnect_info *vpninfo)
+			 struct openconnect_info *vpninfo, struct cert_info *certinfo)
 {
 	ESYS_TR parentHandle = ESYS_TR_NONE;
 	TSS2_RC r;
@@ -248,7 +253,7 @@ static int init_tpm2_key(ESYS_CONTEXT **ctx, ESYS_TR *keyHandle,
 	vpn_progress(vpninfo, PRG_DEBUG,
 		     _("Establishing connection with TPM.\n"));
 
-	r = Esys_Initialize(ctx, NULL, NULL);
+	r = Esys_Initialize(ctx, certinfo->tpm2->tcti_ctx, NULL);
 	if (r) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("TPM2 Esys_Initialize failed: 0x%x\n"),
@@ -267,41 +272,45 @@ static int init_tpm2_key(ESYS_CONTEXT **ctx, ESYS_TR *keyHandle,
 		goto error;
 	}
 
-	if (parent_is_generated(vpninfo->tpm2->parent)) {
-		if (init_tpm2_primary(vpninfo, *ctx, &parentHandle))
+	if (parent_is_generated(certinfo->tpm2->parent)) {
+		if (init_tpm2_primary(vpninfo, certinfo, *ctx, &parentHandle))
 			goto error;
 	} else {
-		r = Esys_TR_FromTPMPublic(*ctx, vpninfo->tpm2->parent,
+		r = Esys_TR_FromTPMPublic(*ctx, certinfo->tpm2->parent,
 					  ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE, &parentHandle);
 		if (r) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Esys_TR_FromTPMPublic failed for handle 0x%x: 0x%x\n"),
-				     vpninfo->tpm2->parent, r);
+				     certinfo->tpm2->parent, r);
 			goto error;
 		}
 		/* If we don't already have a password (and haven't already authenticated
 		 * successfully), check the NODA flag on the parent and demand one if DA
 		 * protection is enabled (since that strongly implies there is a non-empty
 		 * password). */
-		if (!vpninfo->tpm2->did_ownerauth && !vpninfo->tpm2->ownerauth.size) {
+		if (!certinfo->tpm2->did_ownerauth && !certinfo->tpm2->ownerauth.size) {
 			TPM2B_PUBLIC *pub = NULL;
 
 			r = Esys_ReadPublic(*ctx, parentHandle, ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE,
 					    &pub, NULL, NULL);
 			if (!r && !(pub->publicArea.objectAttributes & TPMA_OBJECT_NODA))
-				vpninfo->tpm2->need_ownerauth = 1;
-			free(pub);
+				certinfo->tpm2->need_ownerauth = 1;
+			Esys_Free(pub);
 		}
 	reauth:
-		if (vpninfo->tpm2->need_ownerauth) {
+		if (certinfo->tpm2->need_ownerauth) {
 			char *pass = NULL;
-			if (request_passphrase(vpninfo, "openconnect_tpm2_parent", &pass,
-					       _("Enter TPM2 parent key password:")))
+			if (request_passphrase(vpninfo,
+					       certinfo_string(certinfo, "openconnect_tpm2_parent",
+							       "openconnect_secondary_tpm2_parent"),
+					       &pass,
+					       certinfo_string(certinfo, _("Enter TPM2 parent key password:"),
+							       _("Enter secondary TPM2 parent key password:"))))
 				return -EPERM;
-			install_tpm_passphrase(vpninfo, &vpninfo->tpm2->ownerauth, pass);
-			vpninfo->tpm2->need_ownerauth = 0;
+			install_tpm_passphrase(vpninfo, &certinfo->tpm2->ownerauth, pass);
+			certinfo->tpm2->need_ownerauth = 0;
 		}
-		r = Esys_TR_SetAuth(*ctx, parentHandle, &vpninfo->tpm2->ownerauth);
+		r = Esys_TR_SetAuth(*ctx, parentHandle, &certinfo->tpm2->ownerauth);
 		if (r) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("TPM2 Esys_TR_SetAuth failed: 0x%x\n"),
@@ -314,12 +323,12 @@ static int init_tpm2_key(ESYS_CONTEXT **ctx, ESYS_TR *keyHandle,
 
 	r = Esys_Load(*ctx, parentHandle,
 		      ESYS_TR_PASSWORD, ESYS_TR_NONE, ESYS_TR_NONE,
-		      &vpninfo->tpm2->priv, &vpninfo->tpm2->pub,
+		      &certinfo->tpm2->priv, &certinfo->tpm2->pub,
 		      keyHandle);
-	if (r == PARENT_AUTH_FAILED) {
+	if (rc_is_parent_auth_failed(r)) {
 		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("TPM2 Esys_Load auth failed\n"));
-		vpninfo->tpm2->need_ownerauth = 1;
+		certinfo->tpm2->need_ownerauth = 1;
 		goto reauth;
 	}
 	if (r) {
@@ -328,9 +337,9 @@ static int init_tpm2_key(ESYS_CONTEXT **ctx, ESYS_TR *keyHandle,
 			     r);
 		goto error;
 	}
-	vpninfo->tpm2->did_ownerauth = 1;
+	certinfo->tpm2->did_ownerauth = 1;
 
-	if (parent_is_generated(vpninfo->tpm2->parent)) {
+	if (parent_is_generated(certinfo->tpm2->parent)) {
 		r = Esys_FlushContext(*ctx, parentHandle);
 		if (r) {
 			vpn_progress(vpninfo, PRG_ERR,
@@ -342,7 +351,7 @@ static int init_tpm2_key(ESYS_CONTEXT **ctx, ESYS_TR *keyHandle,
 
 	return 0;
  error:
-	if (parent_is_generated(vpninfo->tpm2->parent) && parentHandle != ESYS_TR_NONE)
+	if (parent_is_generated(certinfo->tpm2->parent) && parentHandle != ESYS_TR_NONE)
 		Esys_FlushContext(*ctx, parentHandle);
 	if (*keyHandle != ESYS_TR_NONE)
 		Esys_FlushContext(*ctx, *keyHandle);
@@ -352,27 +361,32 @@ static int init_tpm2_key(ESYS_CONTEXT **ctx, ESYS_TR *keyHandle,
 	return -EIO;
 }
 
-static int auth_tpm2_key(struct openconnect_info *vpninfo, ESYS_CONTEXT *ctx, ESYS_TR key_handle)
+static int auth_tpm2_key(struct openconnect_info *vpninfo, struct cert_info *certinfo,
+			 ESYS_CONTEXT *ctx, ESYS_TR key_handle)
 {
 	TSS2_RC r;
 
-	if (vpninfo->tpm2->need_userauth || vpninfo->cert_password) {
+	if (certinfo->tpm2->need_userauth) {
 		char *pass = NULL;
 
-		if (vpninfo->cert_password) {
-			pass = vpninfo->cert_password;
-			vpninfo->cert_password = NULL;
+		if (certinfo->password) {
+			pass = certinfo->password;
+			certinfo->password = NULL;
 		} else {
-			int err = request_passphrase(vpninfo, "openconnect_tpm2_key",
-						     &pass, _("Enter TPM2 key password:"));
+			int err = request_passphrase(vpninfo,
+						     certinfo_string(certinfo, "openconnect_tpm2_key",
+								     "openconnect_secondary_tpm2_key"),
+						     &pass,
+						     certinfo_string(certinfo, _("Enter TPM2 key password:"),
+								     _("Enter secondary TPM2 key password:")));
 			if (err)
 				return err;
 		}
-		install_tpm_passphrase(vpninfo, &vpninfo->tpm2->userauth, pass);
-		vpninfo->tpm2->need_userauth = 0;
+		install_tpm_passphrase(vpninfo, &certinfo->tpm2->userauth, pass);
+		certinfo->tpm2->need_userauth = 0;
 	}
 
-	r = Esys_TR_SetAuth(ctx, key_handle, &vpninfo->tpm2->userauth);
+	r = Esys_TR_SetAuth(ctx, key_handle, &certinfo->tpm2->userauth);
 	if (r) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("TPM2 Esys_TR_SetAuth failed: 0x%x\n"),
@@ -383,10 +397,11 @@ static int auth_tpm2_key(struct openconnect_info *vpninfo, ESYS_CONTEXT *ctx, ES
 }
 
 int tpm2_rsa_sign_hash_fn(gnutls_privkey_t key, gnutls_sign_algorithm_t algo,
-			  void *_vpninfo, unsigned int flags,
+			  void *_certinfo, unsigned int flags,
 			  const gnutls_datum_t *data, gnutls_datum_t *sig)
 {
-	struct openconnect_info *vpninfo = _vpninfo;
+	struct cert_info *certinfo = _certinfo;
+	struct openconnect_info *vpninfo = certinfo->vpninfo;
 	int ret = GNUTLS_E_PK_SIGN_FAILED;
 	ESYS_CONTEXT *ectx = NULL;
 	TPM2B_PUBLIC_KEY_RSA digest, *tsig = NULL;
@@ -396,27 +411,28 @@ int tpm2_rsa_sign_hash_fn(gnutls_privkey_t key, gnutls_sign_algorithm_t algo,
 	TSS2_RC r;
 
 	vpn_progress(vpninfo, PRG_DEBUG,
-		     _("TPM2 RSA sign function called for %d bytes.\n"),
-		     data->size);
+		     _("TPM2 RSA sign function called for %d bytes, algo %s\n"),
+		     data->size, gnutls_sign_get_name(algo));
 
-	digest.size = vpninfo->tpm2->pub.publicArea.unique.rsa.size;
+	digest.size = certinfo->tpm2->pub.publicArea.unique.rsa.size;
 
-	if (oc_pkcs1_pad(vpninfo, digest.buffer, digest.size, data))
+	if (oc_pad_rsasig(vpninfo, algo, digest.buffer, digest.size, data,
+			  certinfo->tpm2->pub.publicArea.parameters.rsaDetail.keyBits))
 		return GNUTLS_E_PK_SIGN_FAILED;
 
-	if (init_tpm2_key(&ectx, &key_handle, vpninfo))
+	if (init_tpm2_key(&ectx, &key_handle, vpninfo, certinfo))
 		goto out;
  reauth:
-	if (auth_tpm2_key(vpninfo, ectx, key_handle))
+	if (auth_tpm2_key(vpninfo, certinfo, ectx, key_handle))
 		goto out;
 
 	r = Esys_RSA_Decrypt(ectx, key_handle,
 			     ESYS_TR_PASSWORD, ESYS_TR_NONE, ESYS_TR_NONE,
 			     &digest, &inScheme, &label, &tsig);
-	if (r == KEY_AUTH_FAILED) {
+	if (rc_is_key_auth_failed(r)) {
 		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("TPM2 Esys_RSA_Decrypt auth failed\n"));
-		vpninfo->tpm2->need_userauth = 1;
+		certinfo->tpm2->need_userauth = 1;
 		goto reauth;
 	}
 	if (r) {
@@ -435,8 +451,7 @@ int tpm2_rsa_sign_hash_fn(gnutls_privkey_t key, gnutls_sign_algorithm_t algo,
 
 	ret = 0;
  out:
-	if (tsig)
-		free(tsig);
+	Esys_Free(tsig);
 
 	if (key_handle != ESYS_TR_NONE)
 		Esys_FlushContext(ectx, key_handle);
@@ -448,10 +463,11 @@ int tpm2_rsa_sign_hash_fn(gnutls_privkey_t key, gnutls_sign_algorithm_t algo,
 }
 
 int tpm2_ec_sign_hash_fn(gnutls_privkey_t key, gnutls_sign_algorithm_t algo,
-			 void *_vpninfo, unsigned int flags,
+			 void *_certinfo, unsigned int flags,
 			 const gnutls_datum_t *data, gnutls_datum_t *sig)
 {
-	struct openconnect_info *vpninfo = _vpninfo;
+	struct cert_info *certinfo = _certinfo;
+	struct openconnect_info *vpninfo = certinfo->vpninfo;
 	int ret = GNUTLS_E_PK_SIGN_FAILED;
 	ESYS_CONTEXT *ectx = NULL;
 	TPM2B_DIGEST digest;
@@ -468,35 +484,45 @@ int tpm2_ec_sign_hash_fn(gnutls_privkey_t key, gnutls_sign_algorithm_t algo,
 		     _("TPM2 EC sign function called for %d bytes.\n"),
 		     data->size);
 
-	switch (algo) {
-	case GNUTLS_SIGN_ECDSA_SHA1:   inScheme.details.ecdsa.hashAlg = TPM2_ALG_SHA1;   break;
-	case GNUTLS_SIGN_ECDSA_SHA256: inScheme.details.ecdsa.hashAlg = TPM2_ALG_SHA256; break;
-	case GNUTLS_SIGN_ECDSA_SHA384: inScheme.details.ecdsa.hashAlg = TPM2_ALG_SHA384; break;
-	case GNUTLS_SIGN_ECDSA_SHA512: inScheme.details.ecdsa.hashAlg = TPM2_ALG_SHA512; break;
+	/* FIPS-186-4 §6.4 says "When the length of the output of the hash
+	 * function is greater than the bit length of n, then the leftmost
+	 * n bits of the hash function output block shall be used in any
+	 * calculation using the hash function output during the generation
+	 * or verification of a digital signature."
+	 *
+	 * So GnuTLS is expected to *truncate* a larger hash to fit the
+	 * curve bit length, and then we lie to the TPM about which hash
+	 * it was because the TPM only really cares about the size of the
+	 * data anyway. */
+	switch (data->size) {
+	case SHA1_SIZE:   inScheme.details.ecdsa.hashAlg = TPM2_ALG_SHA1;   break;
+	case SHA256_SIZE: inScheme.details.ecdsa.hashAlg = TPM2_ALG_SHA256; break;
+	case SHA384_SIZE: inScheme.details.ecdsa.hashAlg = TPM2_ALG_SHA384; break;
+	case SHA512_SIZE: inScheme.details.ecdsa.hashAlg = TPM2_ALG_SHA512; break;
 	default:
 		vpn_progress(vpninfo, PRG_ERR,
-			     _("Unknown TPM2 EC digest size %d\n"),
-			     data->size);
+			     _("Unknown TPM2 EC digest size %d for algo 0x%x\n"),
+			     data->size, algo);
 		return GNUTLS_E_PK_SIGN_FAILED;
 	}
 
 	memcpy(digest.buffer, data->data, data->size);
 	digest.size = data->size;
 
-	if (init_tpm2_key(&ectx, &key_handle, vpninfo))
+	if (init_tpm2_key(&ectx, &key_handle, vpninfo, certinfo))
 		goto out;
  reauth:
-	if (auth_tpm2_key(vpninfo, ectx, key_handle))
+	if (auth_tpm2_key(vpninfo, certinfo, ectx, key_handle))
 		goto out;
 
 	r = Esys_Sign(ectx, key_handle,
 		      ESYS_TR_PASSWORD, ESYS_TR_NONE, ESYS_TR_NONE,
 		      &digest, &inScheme, &validation,
 		      &tsig);
-	if (r == KEY_AUTH_FAILED) {
+	if (rc_is_key_auth_failed(r)) {
 		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("TPM2 Esys_Sign auth failed\n"));
-		vpninfo->tpm2->need_userauth = 1;
+		certinfo->tpm2->need_userauth = 1;
 		goto reauth;
 	}
 	if (r) {
@@ -513,7 +539,7 @@ int tpm2_ec_sign_hash_fn(gnutls_privkey_t key, gnutls_sign_algorithm_t algo,
 
 	ret = gnutls_encode_rs_value(sig, &sig_r, &sig_s);
  out:
-	free(tsig);
+	Esys_Free(tsig);
 
 	if (key_handle != ESYS_TR_NONE)
 		Esys_FlushContext(ectx, key_handle);
@@ -524,8 +550,10 @@ int tpm2_ec_sign_hash_fn(gnutls_privkey_t key, gnutls_sign_algorithm_t algo,
 	return ret;
 }
 
-int install_tpm2_key(struct openconnect_info *vpninfo, gnutls_privkey_t *pkey, gnutls_datum_t *pkey_sig,
-		     unsigned int parent, int emptyauth, int legacy, gnutls_datum_t *privdata, gnutls_datum_t *pubdata)
+int install_tpm2_key(struct openconnect_info *vpninfo, struct cert_info *certinfo,
+		     gnutls_privkey_t *pkey, gnutls_datum_t *pkey_sig,
+		     unsigned int parent, int emptyauth, int legacy,
+		     gnutls_datum_t *privdata, gnutls_datum_t *pubdata)
 {
 	TSS2_RC r;
 
@@ -537,14 +565,32 @@ int install_tpm2_key(struct openconnect_info *vpninfo, gnutls_privkey_t *pkey, g
 		return -EINVAL;
 	}
 
-	vpninfo->tpm2 = calloc(1, sizeof(*vpninfo->tpm2));
-	if (!vpninfo->tpm2)
+	certinfo->tpm2 = calloc(1, sizeof(*certinfo->tpm2));
+	if (!certinfo->tpm2)
 		return -ENOMEM;
 
-	vpninfo->tpm2->parent = parent;
+	certinfo->tpm2->parent = parent;
+
+	/* This is the variable which the *IBM* TSS uses, to force it to use
+	 * the swtpm; it happens in the library automatically. To allow the
+	 * swtpm test to work on platforms where a real TPM is available,
+	 * emulate the same thing. Not really intended for production use. */
+	const char *tpm_type = getenv("TPM_INTERFACE_TYPE");
+	if (tpm_type && !strcmp(tpm_type, "socsim")) {
+		vpn_progress(vpninfo, PRG_DEBUG,
+			     _("Using SWTPM due to TPM_INTERFACE_TYPE environment variable\n"));
+
+		r = Tss2_TctiLdr_Initialize("swtpm", &certinfo->tpm2->tcti_ctx);
+		if (r) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("TSS2_TctiLdr_Initialize failed for swtpm: 0x%x\n"),
+				     r);
+			goto err_out;
+		}
+	}
 
 	r = Tss2_MU_TPM2B_PRIVATE_Unmarshal(privdata->data, privdata->size, NULL,
-					    &vpninfo->tpm2->priv);
+					    &certinfo->tpm2->priv);
 	if (r) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to import TPM2 private key data: 0x%x\n"),
@@ -553,7 +599,7 @@ int install_tpm2_key(struct openconnect_info *vpninfo, gnutls_privkey_t *pkey, g
 	}
 
 	r = Tss2_MU_TPM2B_PUBLIC_Unmarshal(pubdata->data, pubdata->size, NULL,
-					   &vpninfo->tpm2->pub);
+					   &certinfo->tpm2->pub);
 	if (r) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to import TPM2 public key data: 0x%x\n"),
@@ -561,29 +607,40 @@ int install_tpm2_key(struct openconnect_info *vpninfo, gnutls_privkey_t *pkey, g
 		goto err_out;
 	}
 
-	vpninfo->tpm2->need_userauth = !emptyauth;
-	vpninfo->tpm2->legacy_srk = legacy;
+	certinfo->tpm2->need_userauth = !emptyauth;
+	certinfo->tpm2->legacy_srk = legacy;
 
-	switch(vpninfo->tpm2->pub.publicArea.type) {
+	switch(certinfo->tpm2->pub.publicArea.type) {
 	case TPM2_ALG_RSA: return GNUTLS_PK_RSA;
 	case TPM2_ALG_ECC: return GNUTLS_PK_ECC;
 	}
 
 	vpn_progress(vpninfo, PRG_ERR,
 		     _("Unsupported TPM2 key type %d\n"),
-		     vpninfo->tpm2->pub.publicArea.type);
+		     certinfo->tpm2->pub.publicArea.type);
  err_out:
-	release_tpm2_ctx(vpninfo);
+	release_tpm2_ctx(vpninfo, certinfo);
 	return -EINVAL;
 }
 
-
-void release_tpm2_ctx(struct openconnect_info *vpninfo)
+uint16_t tpm2_key_curve(struct openconnect_info *vpninfo, struct cert_info *certinfo)
 {
-	if (vpninfo->tpm2) {
-		clear_mem(vpninfo->tpm2->ownerauth.buffer, sizeof(vpninfo->tpm2->ownerauth.buffer));
-		clear_mem(vpninfo->tpm2->userauth.buffer, sizeof(vpninfo->tpm2->userauth.buffer));
-		free(vpninfo->tpm2);
+	return certinfo->tpm2->pub.publicArea.parameters.eccDetail.curveID;
+}
+
+int tpm2_rsa_key_bits(struct openconnect_info *vpninfo, struct cert_info *certinfo)
+{
+	return certinfo->tpm2->pub.publicArea.parameters.rsaDetail.keyBits;
+}
+
+void release_tpm2_ctx(struct openconnect_info *vpninfo, struct cert_info *certinfo)
+{
+	if (certinfo->tpm2) {
+		clear_mem(certinfo->tpm2->ownerauth.buffer, sizeof(certinfo->tpm2->ownerauth.buffer));
+		clear_mem(certinfo->tpm2->userauth.buffer, sizeof(certinfo->tpm2->userauth.buffer));
+		if (certinfo->tpm2->tcti_ctx)
+			Tss2_TctiLdr_Finalize(&certinfo->tpm2->tcti_ctx);
+		free(certinfo->tpm2);
 	}
-	vpninfo->tpm2 = NULL;
+	certinfo->tpm2 = NULL;
 }
