@@ -17,6 +17,7 @@ package org.infradead.libopenconnect;
 
 import java.util.ArrayList;
 import java.util.HashMap;
+import java.time.Instant;
 
 public abstract class LibOpenConnect {
 
@@ -147,11 +148,16 @@ public abstract class LibOpenConnect {
 	public synchronized native void setClientCert(String cert, String sslKey);
 	public synchronized native void setReqMTU(int mtu);
 	public synchronized native void setPFS(boolean isEnabled);
+	public synchronized native int setAllowInsecureCrypto(boolean isEnabled);
 	public synchronized native void setSystemTrust(boolean isEnabled);
 	public synchronized native int setProtocol(String protocol);
+	public synchronized native int disableDTLS();
+	public synchronized native int disableIPv6();
+	public synchronized native void setCookie(String cookie);
 
 	/* connection info */
 
+	public synchronized native String getConnectUrl();
 	public synchronized native String getHostname();
 	public synchronized native String getDNSName();
 	public synchronized native String getUrlpath();
@@ -165,6 +171,7 @@ public abstract class LibOpenConnect {
 	public synchronized native String getDTLSCompression();
 	public synchronized native String getProtocol();
 	public synchronized native int getIdleTimeout();
+	public synchronized native Instant getAuthExpiration();
 
 	/* certificate info */
 
