@@ -15,17 +15,19 @@
 
 #include <config.h>
 
+#include "openconnect.h"
+
+#include <jni.h>
+
+#include <unistd.h>
+#include <sys/types.h>
+
 #include <errno.h>
 #include <stdarg.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <unistd.h>
-#include <sys/types.h>
-
-#include <jni.h>
-#include "openconnect.h"
 
 struct libctx {
 	JNIEnv *jenv;
@@ -49,7 +51,7 @@ static void throw_excep(JNIEnv *jenv, const char *exc, int line)
 		(*jenv)->ThrowNew(jenv, excep, msg);
 }
 
-#define OOM(jenv)	do { throw_excep(jenv, "java/lang/OutOfMemoryError", __LINE__); } while (0)
+#define OOM(jenv)	throw_excep(jenv, "java/lang/OutOfMemoryError", __LINE__)
 
 static struct libctx *getctx(JNIEnv *jenv, jobject jobj)
 {
@@ -75,7 +77,7 @@ static jstring dup_to_jstring(JNIEnv *jenv, const char *in)
 	/*
 	 * Many implementations of NewStringUTF() will return NULL on
 	 * NULL input, but that isn't guaranteed:
-	 * http://gcc.gnu.org/bugzilla/show_bug.cgi?id=35979
+	 * https://gcc.gnu.org/bugzilla/show_bug.cgi?id=35979
 	 */
 	return in ? (*jenv)->NewStringUTF(jenv, in) : NULL;
 }
@@ -850,7 +852,6 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setClien
 
 	release_cstring(ctx->jenv, jcert, cert);
 	release_cstring(ctx->jenv, jsslkey, sslkey);
-	return;
 }
 
 /* special handling: multiple string arguments */
@@ -1001,14 +1002,24 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_resetSSL
 	openconnect_reset_ssl(ctx->vpninfo);
 }
 
-JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_disableIPv6(
+JNIEXPORT int JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_disableIPv6(
 	JNIEnv *jenv, jobject jobj)
 {
 	struct libctx *ctx = getctx(jenv, jobj);
 
 	if (!ctx)
-		return;
-	openconnect_disable_ipv6(ctx->vpninfo);
+		return -EINVAL;
+	return openconnect_disable_ipv6(ctx->vpninfo);
+}
+
+JNIEXPORT int JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_disableDTLS(
+	JNIEnv *jenv, jobject jobj)
+{
+	struct libctx *ctx = getctx(jenv, jobj);
+
+	if (!ctx)
+		return -EINVAL;
+	return openconnect_disable_dtls(ctx->vpninfo);
 }
 
 JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setCertExpiryWarning(
@@ -1071,6 +1082,16 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setPFS(
 	openconnect_set_pfs(ctx->vpninfo, arg);
 }
 
+JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setAllowInsecureCrypto(
+	JNIEnv *jenv, jobject jobj, jboolean arg)
+{
+	struct libctx *ctx = getctx(jenv, jobj);
+
+	if (!ctx)
+		return -EINVAL;
+	return openconnect_set_allow_insecure_crypto(ctx->vpninfo, arg);
+}
+
 JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setSystemTrust(
 	JNIEnv *jenv, jobject jobj, jboolean arg)
 {
@@ -1154,6 +1175,36 @@ JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getIdleT
 	return openconnect_get_idle_timeout(ctx->vpninfo);
 }
 
+JNIEXPORT jobject JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getAuthExpiration(
+	JNIEnv *jenv, jobject jobj)
+{
+	struct libctx *ctx = getctx(jenv, jobj);
+	jmethodID mid;
+	jobject result;
+	jclass jcls;
+	time_t auth_expiration;
+
+	if (!ctx)
+		return NULL;
+
+	auth_expiration = openconnect_get_auth_expiration(ctx->vpninfo);
+	jcls = (*ctx->jenv)->FindClass(ctx->jenv, "java/time/Instant");
+	if (jcls == NULL)
+		goto err;
+	mid = (*jenv)->GetStaticMethodID(jenv, jcls, "ofEpochSecond", "(J)Ljava/time/Instant;");
+	if (!mid)
+		goto err;
+	result = (*jenv)->CallStaticObjectMethod(jenv, jcls, mid, auth_expiration);
+	if (result == NULL)
+		goto err;
+
+	return result;
+
+err:
+	return NULL;
+}
+
+
 /* simple cases: return a const string (no need to free it) */
 
 #define RETURN_STRING_START \
@@ -1251,6 +1302,14 @@ JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getPr
 	RETURN_STRING_END
 }
 
+JNIEXPORT jstring JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_getConnectUrl(
+	JNIEnv *jenv, jobject jobj)
+{
+	RETURN_STRING_START
+	buf = openconnect_get_connect_url(ctx->vpninfo);
+	RETURN_STRING_END
+}
+
 #define SET_STRING_START()			 \
 	struct libctx *ctx = getctx(jenv, jobj); \
 	const char *arg = NULL;			 \
@@ -1353,6 +1412,14 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setUrlpa
 	SET_STRING_END();
 }
 
+JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setCookie(
+	JNIEnv *jenv, jobject jobj, jstring jarg)
+{
+	SET_STRING_START_VOID()
+	openconnect_set_cookie(ctx->vpninfo, arg);
+	SET_STRING_END();
+}
+
 JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setLocalName(
 	JNIEnv *jenv, jobject jobj, jstring jarg)
 {
