@@ -18,12 +18,11 @@
 
 #include <config.h>
 
-#include <string.h>
-#include <errno.h>
-#include <stdlib.h>
-#include <unistd.h>
-#include <fcntl.h>
-#include <ctype.h>
+#include "openconnect-internal.h"
+
+#if defined(OPENCONNECT_GNUTLS)
+#include "gnutls.h"
+#endif
 
 #ifdef HAVE_LIBSTOKEN
 #include <stoken.h>
@@ -32,16 +31,18 @@
 #include <libxml/tree.h>
 #include <zlib.h>
 
-#include "openconnect-internal.h"
-
-#if defined(OPENCONNECT_GNUTLS)
-#include "gnutls.h"
-#endif
-
 #if defined(OPENCONNECT_OPENSSL)
 #include <openssl/bio.h>
 #endif
 
+#include <unistd.h>
+#include <fcntl.h>
+
+#include <string.h>
+#include <errno.h>
+#include <stdlib.h>
+#include <ctype.h>
+
 struct openconnect_info *openconnect_vpninfo_new(const char *useragent,
 						 openconnect_validate_peer_cert_vfn validate_peer_cert,
 						 openconnect_write_new_config_vfn write_new_config,
@@ -66,12 +67,16 @@ struct openconnect_info *openconnect_vpninfo_new(const char *useragent,
 		vpninfo->ic_legacy_to_utf8 = (iconv_t)-1;
 	}
 #endif
+#ifdef HAVE_VHOST
+	vpninfo->vhost_fd = vpninfo->vhost_call_fd = vpninfo->vhost_kick_fd = -1;
+#endif
 #ifndef _WIN32
 	vpninfo->tun_fd = -1;
 #endif
+	init_pkt_queue(&vpninfo->free_queue);
 	init_pkt_queue(&vpninfo->incoming_queue);
 	init_pkt_queue(&vpninfo->outgoing_queue);
-	init_pkt_queue(&vpninfo->oncp_control_queue);
+	init_pkt_queue(&vpninfo->tcp_control_queue);
 	vpninfo->dtls_tos_current = 0;
 	vpninfo->dtls_pass_tos = 0;
 	vpninfo->ssl_fd = vpninfo->dtls_fd = -1;
@@ -81,6 +86,7 @@ struct openconnect_info *openconnect_vpninfo_new(const char *useragent,
 	vpninfo->req_compr = COMPR_STATELESS;
 	vpninfo->max_qlen = 10;
 	vpninfo->localname = strdup("localhost");
+	vpninfo->port = 443;
 	vpninfo->useragent = openconnect_create_useragent(useragent);
 	vpninfo->validate_peer_cert = validate_peer_cert;
 	vpninfo->write_new_config = write_new_config;
@@ -93,7 +99,9 @@ struct openconnect_info *openconnect_vpninfo_new(const char *useragent,
 	vpninfo->proxy_auth[AUTH_TYPE_BASIC].state = AUTH_DEFAULT_DISABLED;
 	vpninfo->http_auth[AUTH_TYPE_BASIC].state = AUTH_DEFAULT_DISABLED;
 	openconnect_set_reported_os(vpninfo, NULL);
-
+#ifdef HAVE_EPOLL
+	vpninfo->epoll_fd = epoll_create1(EPOLL_CLOEXEC);
+#endif
 	if (!vpninfo->localname || !vpninfo->useragent)
 		goto err;
 
@@ -115,12 +123,14 @@ static const struct vpn_proto openconnect_protos[] = {
 		.name = "anyconnect",
 		.pretty_name = N_("Cisco AnyConnect or openconnect"),
 		.description = N_("Compatible with Cisco AnyConnect SSL VPN, as well as ocserv"),
+		.proto = PROTO_ANYCONNECT,
 		.flags = OC_PROTO_PROXY | OC_PROTO_CSD | OC_PROTO_AUTH_CERT | OC_PROTO_AUTH_OTP | OC_PROTO_AUTH_STOKEN,
 		.vpn_close_session = cstp_bye,
 		.tcp_connect = cstp_connect,
 		.tcp_mainloop = cstp_mainloop,
 		.add_http_headers = cstp_common_headers,
 		.obtain_cookie = cstp_obtain_cookie,
+		.secure_cookie = "webvpn",
 		.udp_protocol = "DTLS",
 #ifdef HAVE_DTLS
 		.udp_setup = dtls_setup,
@@ -132,12 +142,14 @@ static const struct vpn_proto openconnect_protos[] = {
 		.name = "nc",
 		.pretty_name = N_("Juniper Network Connect"),
 		.description = N_("Compatible with Juniper Network Connect"),
-		.flags = OC_PROTO_PROXY | OC_PROTO_CSD | OC_PROTO_AUTH_CERT | OC_PROTO_AUTH_OTP | OC_PROTO_PERIODIC_TROJAN,
+		.proto = PROTO_NC,
+		.flags = OC_PROTO_PROXY | OC_PROTO_CSD | OC_PROTO_AUTH_CERT | OC_PROTO_AUTH_OTP | OC_PROTO_AUTH_STOKEN | OC_PROTO_PERIODIC_TROJAN,
 		.vpn_close_session = oncp_bye,
 		.tcp_connect = oncp_connect,
 		.tcp_mainloop = oncp_mainloop,
 		.add_http_headers = oncp_common_headers,
 		.obtain_cookie = oncp_obtain_cookie,
+		.secure_cookie = "DSID",
 		.udp_protocol = "ESP",
 #ifdef HAVE_ESP
 		.udp_setup = esp_setup,
@@ -151,6 +163,7 @@ static const struct vpn_proto openconnect_protos[] = {
 		.name = "gp",
 		.pretty_name = N_("Palo Alto Networks GlobalProtect"),
 		.description = N_("Compatible with Palo Alto Networks (PAN) GlobalProtect SSL VPN"),
+		.proto = PROTO_GPST,
 		.flags = OC_PROTO_PROXY | OC_PROTO_CSD | OC_PROTO_AUTH_CERT | OC_PROTO_AUTH_OTP | OC_PROTO_AUTH_STOKEN | OC_PROTO_PERIODIC_TROJAN,
 		.vpn_close_session = gpst_bye,
 		.tcp_connect = gpst_setup,
@@ -170,7 +183,8 @@ static const struct vpn_proto openconnect_protos[] = {
 		.name = "pulse",
 		.pretty_name = N_("Pulse Connect Secure"),
 		.description = N_("Compatible with Pulse Connect Secure SSL VPN"),
-		.flags = OC_PROTO_PROXY,
+		.proto = PROTO_PULSE,
+		.flags = OC_PROTO_PROXY | OC_PROTO_AUTH_CERT | OC_PROTO_AUTH_OTP | OC_PROTO_AUTH_STOKEN,
 		.vpn_close_session = pulse_bye,
 		.tcp_connect = pulse_connect,
 		.tcp_mainloop = pulse_mainloop,
@@ -184,16 +198,84 @@ static const struct vpn_proto openconnect_protos[] = {
 		.udp_shutdown = esp_shutdown,
 		.udp_send_probes = oncp_esp_send_probes,
 		.udp_catch_probe = oncp_esp_catch_probe,
+#endif
+	}, {
+		.name = "f5",
+		.pretty_name = N_("F5 BIG-IP SSL VPN"),
+		.description = N_("Compatible with F5 BIG-IP SSL VPN"),
+		.proto = PROTO_F5,
+		.flags = OC_PROTO_PROXY | OC_PROTO_AUTH_CERT,
+		.vpn_close_session = f5_bye,
+		.tcp_connect = f5_connect,
+		.tcp_mainloop = ppp_tcp_mainloop,
+		.add_http_headers = http_common_headers,
+		.obtain_cookie = f5_obtain_cookie,
+		.secure_cookie = "MRHSession",
+		.udp_protocol = "DTLS",
+#ifdef HAVE_DTLS
+		.udp_setup = dtls_setup,
+		.udp_mainloop = ppp_udp_mainloop,
+		.udp_close = dtls_close,
+		.udp_shutdown = dtls_shutdown,
+		.udp_catch_probe = f5_dtls_catch_probe,
+#endif
+	}, {
+		.name = "fortinet",
+		.pretty_name = N_("Fortinet SSL VPN"),
+		.description = N_("Compatible with FortiGate SSL VPN"),
+		.proto = PROTO_FORTINET,
+		.flags = OC_PROTO_PROXY | OC_PROTO_AUTH_CERT | OC_PROTO_AUTH_OTP | OC_PROTO_AUTH_STOKEN,
+		.vpn_close_session = fortinet_bye,
+		.tcp_connect = fortinet_connect,
+		.tcp_mainloop = ppp_tcp_mainloop,
+		.add_http_headers = fortinet_common_headers,
+		.obtain_cookie = fortinet_obtain_cookie,
+		.secure_cookie = "SVPNCOOKIE",
+		.udp_protocol = "DTLS",
+#ifdef HAVE_DTLS
+		.udp_setup = dtls_setup,
+		.udp_mainloop = ppp_udp_mainloop,
+		.udp_close = dtls_close,
+		.udp_shutdown = dtls_shutdown,
+		.udp_catch_probe = fortinet_dtls_catch_svrhello,
+#endif
+	}, {
+		.name = "nullppp",
+		.pretty_name = N_("PPP over TLS"),
+		.description = N_("Unauthenticated RFC1661/RFC1662 PPP over TLS, for testing"),
+		.proto = PROTO_NULLPPP,
+		.flags = OC_PROTO_PROXY | OC_PROTO_HIDDEN,
+		.tcp_connect = nullppp_connect,
+		.tcp_mainloop = nullppp_mainloop,
+		.add_http_headers = http_common_headers,
+		.obtain_cookie = nullppp_obtain_cookie,
+	}, {
+		.name = "array",
+		.pretty_name = N_("Array SSL VPN"),
+		.description = N_("Compatible with Array Networks SSL VPN"),
+		.proto = PROTO_ARRAY,
+		.flags = OC_PROTO_PROXY,
+		.vpn_close_session = array_bye,
+		.tcp_connect = array_connect,
+		.tcp_mainloop = array_mainloop,
+		.add_http_headers = http_common_headers,
+		.obtain_cookie = array_obtain_cookie,
+		.udp_protocol = "DTLS",
+#ifdef HAVE_DTLS
+		.udp_setup = dtls_setup,
+		.udp_mainloop = array_dtls_mainloop,
+		.udp_close = dtls_close,
+		.udp_shutdown = dtls_shutdown,
 #endif
 	},
 };
 
-#define NR_PROTOS (sizeof(openconnect_protos)/sizeof(*openconnect_protos))
+#define NR_PROTOS ARRAY_SIZE(openconnect_protos)
 
 int openconnect_get_supported_protocols(struct oc_vpn_proto **protos)
 {
 	struct oc_vpn_proto *pr;
-	int i;
+	int i, j;
 
         /* The original version of this function included an all-zero
          * sentinel value at the end of the array, so we must continue
@@ -205,13 +287,16 @@ int openconnect_get_supported_protocols(struct oc_vpn_proto **protos)
 	if (!pr)
 		return -ENOMEM;
 
-	for (i = 0; i < NR_PROTOS; i++) {
-		pr[i].name = openconnect_protos[i].name;
-		pr[i].pretty_name = _(openconnect_protos[i].pretty_name);
-		pr[i].description = _(openconnect_protos[i].description);
-		pr[i].flags = openconnect_protos[i].flags;
+	for (i = j = 0; i < NR_PROTOS; i++) {
+		if (!(openconnect_protos[i].flags & OC_PROTO_HIDDEN)) {
+			pr[j].name = openconnect_protos[i].name;
+			pr[j].pretty_name = _(openconnect_protos[i].pretty_name);
+			pr[j].description = _(openconnect_protos[i].description);
+			pr[j].flags = openconnect_protos[i].flags;
+			j++;
+		}
 	}
-	return i;
+	return j;
 }
 
 void openconnect_free_supported_protocols(struct oc_vpn_proto *protos)
@@ -258,8 +343,9 @@ int openconnect_setup_dtls(struct openconnect_info *vpninfo,
 			   int attempt_period)
 
 {
+	vpninfo->dtls_attempt_period = attempt_period;
 	if (vpninfo->proto->udp_setup)
-		return vpninfo->proto->udp_setup(vpninfo, attempt_period);
+		return vpninfo->proto->udp_setup(vpninfo);
 
 	vpn_progress(vpninfo, PRG_ERR,
 		     _("Built against SSL library with no Cisco DTLS support\n"));
@@ -281,36 +367,47 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 
 int openconnect_make_cstp_connection(struct openconnect_info *vpninfo)
 {
-	return vpninfo->proto->tcp_connect(vpninfo);
+	int result = vpninfo->proto->tcp_connect(vpninfo);
+
+	/* ssl_times.last_tx should be set to show that a connection has been setup */
+	if (result == 0 && vpninfo->ssl_times.last_tx == 0)
+		vpninfo->ssl_times.last_tx = time(NULL);
+	return result;
 }
 
 int openconnect_set_reported_os(struct openconnect_info *vpninfo,
 				const char *os)
 {
+	const char *allowed[] = {"linux", "linux-64", "win", "mac-intel", "android", "apple-ios"};
+
 	if (!os) {
 #if defined(__APPLE__)
+#  include <TargetConditionals.h>
+#  if TARGET_OS_IOS
+		/* We need to use Apple's boolean "target" defines to distinguish iOS from
+		 * desktop MacOS. See  https://stackoverflow.com/a/5920028 and
+		 * https://github.com/mstg/iOS-full-sdk/blob/master/iPhoneOS9.3.sdk/usr/include/TargetConditionals.h#L64-L71
+		 */
+		os = "apple-ios";
+#  else
 		os = "mac-intel";
+#  endif
 #elif defined(__ANDROID__)
 		os = "android";
+#elif defined(_WIN32)
+		os = "win";
 #else
 		os = sizeof(long) > 4 ? "linux-64" : "linux";
 #endif
 	}
 
-	if (!strcmp(os, "mac-intel"))
-		vpninfo->csd_xmltag = "csdMac";
-	else if (!strcmp(os, "linux") || !strcmp(os, "linux-64"))
-		vpninfo->csd_xmltag = "csdLinux";
-	else if (!strcmp(os, "android") || !strcmp(os, "apple-ios")) {
-		vpninfo->csd_xmltag = "csdLinux";
-		vpninfo->csd_nostub = 1;
-	} else if (!strcmp(os, "win"))
-		vpninfo->csd_xmltag = "csd";
-	else
-		return -EINVAL;
-
-	STRDUP(vpninfo->platname, os);
-	return 0;
+	for (int i = 0; i < ARRAY_SIZE(allowed); i++) {
+		if (!strcmp(os, allowed[i])) {
+			STRDUP(vpninfo->platname, os);
+			return 0;
+		}
+	}
+	return -EINVAL;
 }
 
 int openconnect_set_mobile_info(struct openconnect_info *vpninfo,
@@ -333,6 +430,55 @@ int openconnect_set_version_string(struct openconnect_info *vpninfo,
 	return 0;
 }
 
+const char *add_option_dup(struct oc_vpn_option **list,
+			   const char *opt,
+			   const char *val, int val_len)
+{
+	const char *ret;
+	char *new_val;
+
+	if (val_len >= 0)
+		new_val = strndup(val, val_len);
+	else
+		new_val = strdup(val);
+
+	ret = add_option_steal(list, opt, &new_val);
+	free(new_val);
+	return ret;
+}
+
+const char *add_option_steal(struct oc_vpn_option **list,
+			     const char *opt, char **val)
+{
+	struct oc_vpn_option *new = malloc(sizeof(*new));
+	if (!new)
+		return NULL;
+
+	new->option = strdup(opt);
+	if (!new->option) {
+		free(new);
+		return NULL;
+	}
+	new->value = *val;
+	*val = NULL;
+	new->next = *list;
+	*list = new;
+
+	return new->value;
+}
+
+const char *add_option_ipaddr(struct oc_vpn_option **list,
+			      const char *opt, int af, void *addr)
+{
+	char buf[40];
+
+	if (!inet_ntop(af, addr, buf, sizeof(buf)))
+		return NULL;
+
+	return add_option_dup(list, opt, buf, -1);
+}
+
+
 void free_optlist(struct oc_vpn_option *opt)
 {
 	struct oc_vpn_option *next;
@@ -345,6 +491,92 @@ void free_optlist(struct oc_vpn_option *opt)
 	}
 }
 
+int install_vpn_opts(struct openconnect_info *vpninfo, struct oc_vpn_option *opt,
+		     struct oc_ip_info *ip_info)
+{
+	/* XX: remove protocol-specific exceptions here, once we can test them
+	 * with F5 reconnections in addition to Juniper reconnections. See:
+	 * https://gitlab.com/openconnect/openconnect/-/merge_requests/293#note_702388182
+	 */
+	if (!ip_info->addr && !ip_info->addr6 && !ip_info->netmask6) {
+		if (vpninfo->proto->proto == PROTO_F5) {
+			/* F5 doesn't get its IP address until it actually establishes the
+			 * PPP connection. */
+		} else if (vpninfo->proto->proto == PROTO_NC && vpninfo->ip_info.addr) {
+			/* Juniper doesn't necessarily resend the Legacy IP address in the
+			 * event of a rekey/reconnection. */
+			ip_info->addr = add_option_dup(&opt, "ipaddr", vpninfo->ip_info.addr, -1);
+			if (!ip_info->netmask && vpninfo->ip_info.netmask)
+				ip_info->netmask = add_option_dup(&opt, "netmask", vpninfo->ip_info.netmask, -1);
+			vpn_progress(vpninfo, PRG_DEBUG,
+				     _("No IP address received with Juniper rekey/reconnection.\n"));
+			goto after_ip_checks;
+		} else {
+			/* For all other protocols, not receiving any IP address is an error */
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("No IP address received. Aborting\n"));
+			return -EINVAL;
+		}
+	}
+
+	if (vpninfo->ip_info.addr) {
+		if (!ip_info->addr || strcmp(ip_info->addr, vpninfo->ip_info.addr)) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Reconnect gave different Legacy IP address (%s != %s)\n"),
+				     ip_info->addr, vpninfo->ip_info.addr);
+			/* EPERM means that the retry loop will abort and won't keep trying. */
+			return -EPERM;
+		}
+	}
+	if (vpninfo->ip_info.netmask) {
+		if (!ip_info->netmask || strcmp(ip_info->netmask, vpninfo->ip_info.netmask)) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Reconnect gave different Legacy IP netmask (%s != %s)\n"),
+				     ip_info->netmask, vpninfo->ip_info.netmask);
+			return -EPERM;
+		}
+	}
+	if (vpninfo->ip_info.addr6) {
+		if (!ip_info->addr6 || strcmp(ip_info->addr6, vpninfo->ip_info.addr6)) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Reconnect gave different IPv6 address (%s != %s)\n"),
+				     ip_info->addr6, vpninfo->ip_info.addr6);
+			return -EPERM;
+		}
+	}
+	if (vpninfo->ip_info.netmask6) {
+		if (!ip_info->netmask6 || strcmp(ip_info->netmask6, vpninfo->ip_info.netmask6)) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Reconnect gave different IPv6 netmask (%s != %s)\n"),
+				     ip_info->netmask6, vpninfo->ip_info.netmask6);
+			return -EPERM;
+		}
+	}
+
+ after_ip_checks:
+	/* Preserve gateway_addr and MTU if they were set */
+	ip_info->gateway_addr = vpninfo->ip_info.gateway_addr;
+	if (!ip_info->mtu)
+		ip_info->mtu = vpninfo->ip_info.mtu;
+
+	if (ip_info->mtu && ip_info->mtu < 1280 &&
+	    (ip_info->addr6 || ip_info->netmask6)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("IPv6 configuration received but MTU %d is too small.\n"),
+			     ip_info->mtu);
+	}
+
+	/* Free the original options */
+	free_split_routes(&vpninfo->ip_info);
+	free_optlist(vpninfo->cstp_options);
+
+	/* Install the new options */
+	vpninfo->cstp_options = opt;
+	vpninfo->ip_info = *ip_info;
+
+	return 0;
+}
+
 void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 {
 	openconnect_close_https(vpninfo, 1);
@@ -357,6 +589,10 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 		closesocket(vpninfo->cmd_fd_write);
 	}
 
+	free(vpninfo->ppp);
+	buf_free(vpninfo->ppp_tls_connect_req);
+	buf_free(vpninfo->ppp_dtls_connect_req);
+
 #ifdef HAVE_ICONV
 	if (vpninfo->ic_utf8_to_legacy != (iconv_t)-1)
 		iconv_close(vpninfo->ic_utf8_to_legacy);
@@ -371,6 +607,7 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 		CloseHandle(vpninfo->ssl_event);
 	if (vpninfo->dtls_event)
 		CloseHandle(vpninfo->dtls_event);
+	free(vpninfo->ifname_w);
 #endif
 	free(vpninfo->peer_addr);
 	free(vpninfo->ip_info.gateway_addr);
@@ -379,9 +616,10 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	free_optlist(vpninfo->cookies);
 	free_optlist(vpninfo->cstp_options);
 	free_optlist(vpninfo->dtls_options);
-	free_split_routes(vpninfo);
+	free_split_routes(&vpninfo->ip_info);
 	free(vpninfo->hostname);
 	free(vpninfo->unique_hostname);
+	buf_free(vpninfo->connect_urlbuf);
 	free(vpninfo->urlpath);
 	free(vpninfo->redirect_url);
 	free_pass(&vpninfo->cookie);
@@ -389,12 +627,13 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	free(vpninfo->proxy);
 	free(vpninfo->proxy_user);
 	free_pass(&vpninfo->proxy_pass);
-	free_pass(&vpninfo->cert_password);
+	free_pass(&vpninfo->certinfo[0].password);
 	free(vpninfo->vpnc_script);
 	free(vpninfo->cafile);
 	free(vpninfo->ifname);
 	free(vpninfo->dtls_cipher);
 	free(vpninfo->peer_cert_hash);
+	free(vpninfo->ciphersuite_config);
 #if defined(OPENCONNECT_OPENSSL)
 	free(vpninfo->cstp_cipher);
 #if defined(HAVE_BIO_METH_FREE)
@@ -434,9 +673,9 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	/* These are const in openconnect itself, but for consistency of
 	   the library API we do take ownership of the strings we're given,
 	   and thus we have to free them too. */
-	if (vpninfo->cert != vpninfo->sslkey)
-		free((void *)vpninfo->sslkey);
-	free((void *)vpninfo->cert);
+	if (vpninfo->certinfo[0].cert != vpninfo->certinfo[0].key)
+		free((void *)vpninfo->certinfo[0].key);
+	free((void *)vpninfo->certinfo[0].cert);
 	if (vpninfo->peer_cert) {
 #if defined(OPENCONNECT_OPENSSL)
 		X509_free(vpninfo->peer_cert);
@@ -490,14 +729,56 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	inflateEnd(&vpninfo->inflate_strm);
 	deflateEnd(&vpninfo->deflate_strm);
 
-	free(vpninfo->deflate_pkt);
-	free(vpninfo->tun_pkt);
-	free(vpninfo->dtls_pkt);
-	free(vpninfo->cstp_pkt);
+#ifdef HAVE_EPOLL
+	if (vpninfo->epoll_fd >= 0)
+		close(vpninfo->epoll_fd);
+#endif
+
+	free_pkt(vpninfo, vpninfo->deflate_pkt);
+	free_pkt(vpninfo, vpninfo->tun_pkt);
+	free_pkt(vpninfo, vpninfo->dtls_pkt);
+	free_pkt(vpninfo, vpninfo->cstp_pkt);
+	struct pkt *pkt;
+	while ((pkt = dequeue_packet(&vpninfo->free_queue)))
+		free(pkt);
+
 	free(vpninfo->bearer_token);
 	free(vpninfo);
 }
 
+
+const char *openconnect_get_connect_url(struct openconnect_info *vpninfo)
+{
+	struct oc_text_buf *urlbuf = vpninfo->connect_urlbuf;
+
+	if (!urlbuf)
+		urlbuf = buf_alloc();
+
+	buf_append(urlbuf, "https://%s", vpninfo->hostname);
+	if (vpninfo->port != 443)
+		buf_append(urlbuf, ":%d", vpninfo->port);
+	buf_append(urlbuf, "/");
+
+	/* Other protocols don't care and just leave noise from the
+	 * authentication process in ->urlpath. Pulse does care, and
+	 * you have to *connect* to a given usergroup at the correct
+	 * path, not just authenticate.
+	 *
+	 * https://gitlab.gnome.org/GNOME/NetworkManager-openconnect/-/issues/53
+	 * https://gitlab.gnome.org/GNOME/NetworkManager-openconnect/-/merge_requests/22
+	 */
+	if (vpninfo->proto->proto == PROTO_PULSE)
+		buf_append(urlbuf, "%s", vpninfo->urlpath);
+	if (buf_error(urlbuf)) {
+		buf_free(urlbuf);
+		vpninfo->connect_urlbuf = NULL;
+		return NULL;
+	}
+
+	vpninfo->connect_urlbuf = urlbuf;
+	return urlbuf->data;
+}
+
 const char *openconnect_get_hostname(struct openconnect_info *vpninfo)
 {
 	return vpninfo->unique_hostname?:vpninfo->hostname;
@@ -556,9 +837,40 @@ void openconnect_set_xmlsha1(struct openconnect_info *vpninfo,
 	memcpy(&vpninfo->xmlsha1, xmlsha1, size);
 }
 
-void openconnect_disable_ipv6(struct openconnect_info *vpninfo)
+int openconnect_disable_ipv6(struct openconnect_info *vpninfo)
 {
+	/* This prevents disabling IPv6 when the connection is
+	 * currently connected or has been connected previously.
+	 *
+	 * XX: It would be better to allow it when currently
+	 * disconnected, but we currently have no way to indicate
+	 * a state in which IP and routing configuration are
+	 * unconfigured state. (Neither a closed TLS socket
+	 * nor tunnel socket is a reliable indicator.)
+	 */
+	if (!vpninfo->disable_ipv6 &&
+	    vpninfo->ssl_times.last_tx != 0)
+		return -EINVAL;
 	vpninfo->disable_ipv6 = 1;
+	return 0;
+}
+
+int openconnect_disable_dtls(struct openconnect_info *vpninfo)
+{
+	/* This disables DTLS or ESP. It is prevented when the
+	 * connection is currently connected or has been
+	 * connected previously.
+	 *
+	 * XX: It would be better to allow it when DTLS is not
+	 * in use, but other than DTLS already being disabled,
+	 * we currently do not have a reliable indicator of
+	 * this.
+	 */
+	if (vpninfo->dtls_state != DTLS_NOSECRET
+	    || vpninfo->ssl_times.last_tx != 0)
+		return -EINVAL;
+	vpninfo->dtls_state = DTLS_DISABLED;
+	return 0;
 }
 
 int openconnect_set_cafile(struct openconnect_info *vpninfo, const char *cafile)
@@ -603,6 +915,11 @@ int openconnect_get_idle_timeout(struct openconnect_info *vpninfo)
 	return vpninfo->idle_timeout;
 }
 
+time_t openconnect_get_auth_expiration(struct openconnect_info *vpninfo)
+{
+	return vpninfo->auth_expiration;
+}
+
 int openconnect_get_ip_info(struct openconnect_info *vpninfo,
 			    const struct oc_ip_info **info,
 			    const struct oc_vpn_option **cstp_options,
@@ -641,15 +958,15 @@ int openconnect_set_client_cert(struct openconnect_info *vpninfo,
 	UTF8CHECK(sslkey);
 
 	/* Avoid freeing it twice if it's the same */
-	if (vpninfo->sslkey == vpninfo->cert)
-		vpninfo->sslkey = NULL;
+	if (vpninfo->certinfo[0].key == vpninfo->certinfo[0].cert)
+		vpninfo->certinfo[0].key = NULL;
 
-	STRDUP(vpninfo->cert, cert);
+	STRDUP(vpninfo->certinfo[0].cert, cert);
 
 	if (sslkey) {
-		STRDUP(vpninfo->sslkey, sslkey);
+		STRDUP(vpninfo->certinfo[0].key, sslkey);
 	} else {
-		vpninfo->sslkey = vpninfo->cert;
+		vpninfo->certinfo[0].key = vpninfo->certinfo[0].cert;
 	}
 
 	return 0;
@@ -671,6 +988,15 @@ void openconnect_clear_cookie(struct openconnect_info *vpninfo)
 		memset(vpninfo->cookie, 0, strlen(vpninfo->cookie));
 }
 
+int openconnect_set_cookie(struct openconnect_info *vpninfo,
+			    const char *cookie)
+{
+	UTF8CHECK(cookie);
+
+	STRDUP(vpninfo->cookie, cookie);
+	return 0;
+}
+
 void openconnect_reset_ssl(struct openconnect_info *vpninfo)
 {
 	vpninfo->got_cancel_cmd = 0;
@@ -722,7 +1048,7 @@ void openconnect_set_cert_expiry_warning(struct openconnect_info *vpninfo,
 
 int openconnect_set_key_password(struct openconnect_info *vpninfo, const char *pass)
 {
-	STRDUP(vpninfo->cert_password, pass);
+	STRDUP(vpninfo->certinfo[0].password, pass);
 
 	return 0;
 }
@@ -732,6 +1058,15 @@ void openconnect_set_pfs(struct openconnect_info *vpninfo, unsigned val)
 	vpninfo->pfs = val;
 }
 
+int openconnect_set_allow_insecure_crypto(struct openconnect_info *vpninfo, unsigned val)
+{
+	int ret = can_enable_insecure_crypto();
+	if (ret)
+		return ret;
+	vpninfo->allow_insecure_crypto = val;
+	return 0;
+}
+
 void openconnect_set_cancel_fd(struct openconnect_info *vpninfo, int fd)
 {
 	vpninfo->cmd_fd = fd;
@@ -762,6 +1097,7 @@ OPENCONNECT_CMD_SOCKET openconnect_setup_cmd_pipe(struct openconnect_info *vpnin
 	}
 	vpninfo->cmd_fd = pipefd[0];
 	vpninfo->cmd_fd_write = pipefd[1];
+	vpninfo->need_poll_cmd_fd = 1;
 	return vpninfo->cmd_fd_write;
 }
 
@@ -981,6 +1317,14 @@ int openconnect_setup_tun_device(struct openconnect_info *vpninfo,
 	STRDUP(vpninfo->ifname, ifname);
 
 	prepare_script_env(vpninfo);
+
+	/* XX: vpninfo->ifname will only be non-NULL here if set by the -i option,
+	   which only works on some platforms (see os_setup_tun implementations) */
+	legacy_ifname = vpninfo->ifname ? openconnect_utf8_to_legacy(vpninfo, vpninfo->ifname) : NULL;
+	script_setenv(vpninfo, "TUNDEV", legacy_ifname, 0, 0);
+	if (legacy_ifname != vpninfo->ifname)
+		free(legacy_ifname);
+
 	script_config_tun(vpninfo, "pre-init");
 
 	tun_fd = os_setup_tun(vpninfo);
@@ -991,10 +1335,13 @@ int openconnect_setup_tun_device(struct openconnect_info *vpninfo,
 	if (vpninfo->tun_idx != -1)
 		script_setenv_int(vpninfo, "TUNIDX", vpninfo->tun_idx);
 #endif
+
+	/* XX: os_setup_tun has set (or even changed) ifname */
 	legacy_ifname = openconnect_utf8_to_legacy(vpninfo, vpninfo->ifname);
 	script_setenv(vpninfo, "TUNDEV", legacy_ifname, 0, 0);
 	if (legacy_ifname != vpninfo->ifname)
 		free(legacy_ifname);
+
 	script_config_tun(vpninfo, "connect");
 
 	return openconnect_setup_tun_fd(vpninfo, tun_fd);
@@ -1025,7 +1372,7 @@ const char *openconnect_get_dtls_compression(struct openconnect_info * vpninfo)
 
 const char *openconnect_get_dtls_cipher(struct openconnect_info *vpninfo)
 {
-	if (vpninfo->dtls_state != DTLS_CONNECTED || !vpninfo->dtls_ssl) {
+	if (vpninfo->dtls_state < DTLS_CONNECTED || !vpninfo->dtls_ssl) {
 #if defined(OPENCONNECT_GNUTLS)
 		gnutls_free(vpninfo->dtls_cipher_desc);
 #else
@@ -1038,7 +1385,7 @@ const char *openconnect_get_dtls_cipher(struct openconnect_info *vpninfo)
 	 * one is enabled. */
 	if (vpninfo->dtls_cipher_desc == NULL) {
 #if defined(OPENCONNECT_GNUTLS)
-        vpninfo->dtls_cipher_desc = get_gnutls_cipher(vpninfo->dtls_ssl);
+		vpninfo->dtls_cipher_desc = get_gnutls_cipher(vpninfo->dtls_ssl);
 #else
 		if (asprintf(&vpninfo->dtls_cipher_desc, "%s-%s",
 		             SSL_get_version(vpninfo->dtls_ssl), SSL_get_cipher_name(vpninfo->dtls_ssl)) < 0)
@@ -1122,9 +1469,12 @@ int openconnect_check_peer_cert_hash(struct openconnect_info *vpninfo,
 		if (len < 0)
 			return len;
 
-		if (openconnect_sha1(sha1_bin, cert, len))
+		if (openconnect_sha1(sha1_bin, cert, len)) {
+			free(cert);
 			return -EIO;
+		}
 
+		free(cert);
 		fingerprint = openconnect_bin2hex(NULL, sha1_bin, sizeof(sha1_bin));
 	}
 
@@ -1200,6 +1550,10 @@ int process_auth_form(struct openconnect_info *vpninfo, struct oc_auth_form *for
 		vpn_progress(vpninfo, PRG_ERR, _("No form handler; cannot authenticate.\n"));
 		return OC_FORM_RESULT_ERR;
 	}
+	if (!form->auth_id) {
+		vpn_progress(vpninfo, PRG_ERR, _("No form ID. This is a bug in OpenConnect's authentication code.\n"));
+		return OC_FORM_RESULT_ERR;
+	}
 
 retry:
 	auth_choice = NULL;
@@ -1252,3 +1606,15 @@ retry:
 
 	return ret;
 }
+
+void openconnect_set_webview_callback(struct openconnect_info *vpninfo,
+				      openconnect_open_webview_vfn webview_fn)
+{
+	return;
+}
+
+int openconnect_webview_load_changed(struct openconnect_info *vpninfo,
+				      const struct oc_webview_result *result)
+{
+	return -EOPNOTSUPP;
+}
