@@ -22,73 +22,19 @@
 
 #include <config.h>
 
+#include "openconnect-internal.h"
+
 #include <unistd.h>
 #include <fcntl.h>
+#include <sys/types.h>
+
 #include <time.h>
 #include <string.h>
 #include <ctype.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include <sys/types.h>
 #include <stdarg.h>
-#include <sys/types.h>
-
-#include "openconnect-internal.h"
-
-static int parse_cookie(struct openconnect_info *vpninfo)
-{
-	char *p = vpninfo->cookie;
-
-	/* We currenly expect the "cookie" to be contain multiple cookies:
-	 * DSSignInUrl=/; DSID=xxx; DSFirstAccess=xxx; DSLastAccess=xxx
-	 * Process those into vpninfo->cookies unless we already had them
-	 * (in which case they'll may be newer. */
-	while (p && *p) {
-		char *semicolon = strchr(p, ';');
-		char *equals;
-
-		if (semicolon)
-			*semicolon = 0;
-
-		equals = strchr(p, '=');
-		if (!equals) {
-			vpn_progress(vpninfo, PRG_ERR, _("Invalid cookie '%s'\n"), p);
-			return -EINVAL;
-		}
-		*equals = 0;
-		http_add_cookie(vpninfo, p, equals+1, 0);
-		*equals = '=';
-
-		p = semicolon;
-		if (p) {
-			*p = ';';
-			p++;
-			while (*p && isspace((int)(unsigned char)*p))
-				p++;
-		}
-	}
-
-	return 0;
-}
-
-static void buf_append_be16(struct oc_text_buf *buf, uint16_t val)
-{
-	unsigned char b[2];
-
-	store_be16(b, val);
-
-	buf_append_bytes(buf, b, 2);
-}
-
-static void buf_append_le16(struct oc_text_buf *buf, uint16_t val)
-{
-	unsigned char b[2];
-
-	store_le16(b, val);
-
-	buf_append_bytes(buf, b, 2);
-}
 
 static void buf_append_tlv(struct oc_text_buf *buf, uint16_t val, uint32_t len, void *data)
 {
@@ -115,39 +61,9 @@ static const char authpkt_tail[] = { 0xbb, 0x01, 0x00, 0x00, 0x00, 0x00 };
 
 #define GRP_ATTR(g, a) (((g) << 16) | (a))
 
-/* We behave like CSTP — create a linked list in vpninfo->cstp_options
- * with the strings containing the information we got from the server,
- * and oc_ip_info contains const copies of those pointers. */
-
-static const char *add_option(struct openconnect_info *vpninfo, const char *opt,
-			      const char *val, int val_len)
-{
-	struct oc_vpn_option *new = malloc(sizeof(*new));
-	if (!new)
-		return NULL;
-
-	new->option = strdup(opt);
-	if (!new->option) {
-		free(new);
-		return NULL;
-	}
-	if (val_len >= 0)
-		new->value = strndup(val, val_len);
-	else
-		new->value = strdup(val);
-	if (!new->value) {
-		free(new->option);
-		free(new);
-		return NULL;
-	}
-	new->next = vpninfo->cstp_options;
-	vpninfo->cstp_options = new;
-
-	return new->value;
-}
-
-static int process_attr(struct openconnect_info *vpninfo, int group, int attr,
-			unsigned char *data, int attrlen)
+static int process_attr(struct openconnect_info *vpninfo,
+			struct oc_vpn_option **new_opts, struct oc_ip_info *new_ip_info,
+			int group, int attr, unsigned char *data, int attrlen)
 {
 	char buf[80];
 	int i;
@@ -161,10 +77,10 @@ static int process_attr(struct openconnect_info *vpninfo, int group, int attr,
 				     attrlen, group, attr);
 			return -EINVAL;
 		}
-		vpninfo->ip_info.mtu = load_be32(data);
+		new_ip_info->mtu = load_be32(data);
 		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("Received MTU %d from server\n"),
-			     vpninfo->ip_info.mtu);
+			     new_ip_info->mtu);
 		break;
 
 	case GRP_ATTR(2, 1):
@@ -175,8 +91,8 @@ static int process_attr(struct openconnect_info *vpninfo, int group, int attr,
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received DNS server %s\n"), buf);
 
 		for (i = 0; i < 3; i++) {
-			if (!vpninfo->ip_info.dns[i]) {
-				vpninfo->ip_info.dns[i] = add_option(vpninfo, "DNS", buf, -1);
+			if (!new_ip_info->dns[i]) {
+				new_ip_info->dns[i] = add_option_dup(new_opts, "DNS", buf, -1);
 				break;
 			}
 		}
@@ -185,9 +101,9 @@ static int process_attr(struct openconnect_info *vpninfo, int group, int attr,
 	case GRP_ATTR(2, 2):
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received DNS search domain %.*s\n"),
 			     attrlen, (char *)data);
-		vpninfo->ip_info.domain = add_option(vpninfo, "search", (char *)data, attrlen);
-		if (vpninfo->ip_info.domain) {
-			char *p = (char *)vpninfo->ip_info.domain;
+		new_ip_info->domain = add_option_dup(new_opts, "search", (char *)data, attrlen);
+		if (new_ip_info->domain) {
+			char *p = (char *)new_ip_info->domain;
 			while ((p = strchr(p, ',')))
 				*p = ' ';
 		}
@@ -199,7 +115,7 @@ static int process_attr(struct openconnect_info *vpninfo, int group, int attr,
 		snprintf(buf, sizeof(buf), "%d.%d.%d.%d", data[0], data[1], data[2], data[3]);
 
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received internal IP address %s\n"), buf);
-		vpninfo->ip_info.addr = add_option(vpninfo, "ipaddr", buf, -1);
+		new_ip_info->addr = add_option_dup(new_opts, "ipaddr", buf, -1);
 		break;
 
 	case GRP_ATTR(1, 2):
@@ -208,7 +124,7 @@ static int process_attr(struct openconnect_info *vpninfo, int group, int attr,
 		snprintf(buf, sizeof(buf), "%d.%d.%d.%d", data[0], data[1], data[2], data[3]);
 
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received netmask %s\n"), buf);
-		vpninfo->ip_info.netmask = add_option(vpninfo, "netmask", buf, -1);
+		new_ip_info->netmask = add_option_dup(new_opts, "netmask", buf, -1);
 		break;
 
 	case GRP_ATTR(1, 3):
@@ -219,7 +135,7 @@ static int process_attr(struct openconnect_info *vpninfo, int group, int attr,
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received internal gateway address %s\n"), buf);
 		/* Hm, what are we supposed to do with this? It's a tunnel;
 		   having a gateway is meaningless. */
-		add_option(vpninfo, "ipaddr", buf, -1);
+		add_option_dup(new_opts, "ipaddr", buf, -1);
 		break;
 
 	case GRP_ATTR(3, 3): {
@@ -234,10 +150,10 @@ static int process_attr(struct openconnect_info *vpninfo, int group, int attr,
 			break;
 		inc = malloc(sizeof(*inc));
 		if (inc) {
-			inc->route = add_option(vpninfo, "split-include", buf, -1);
+			inc->route = add_option_dup(new_opts, "split-include", buf, -1);
 			if (inc->route) {
-				inc->next = vpninfo->ip_info.split_includes;
-				vpninfo->ip_info.split_includes = inc;
+				inc->next = new_ip_info->split_includes;
+				new_ip_info->split_includes = inc;
 			} else
 				free(inc);
 		}
@@ -256,10 +172,10 @@ static int process_attr(struct openconnect_info *vpninfo, int group, int attr,
 			break;
 		exc = malloc(sizeof(*exc));
 		if (exc) {
-			exc->route = add_option(vpninfo, "split-exclude", buf, -1);
+			exc->route = add_option_dup(new_opts, "split-exclude", buf, -1);
 			if (exc->route) {
-				exc->next = vpninfo->ip_info.split_excludes;
-				vpninfo->ip_info.split_excludes = exc;
+				exc->next = new_ip_info->split_excludes;
+				new_ip_info->split_excludes = exc;
 			} else
 				free(exc);
 		}
@@ -274,8 +190,8 @@ static int process_attr(struct openconnect_info *vpninfo, int group, int attr,
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received WINS server %s\n"), buf);
 
 		for (i = 0; i < 3; i++) {
-			if (!vpninfo->ip_info.nbns[i]) {
-				vpninfo->ip_info.nbns[i] = add_option(vpninfo, "WINS", buf, -1);
+			if (!new_ip_info->nbns[i]) {
+				new_ip_info->nbns[i] = add_option_dup(new_opts, "WINS", buf, -1);
 				break;
 			}
 		}
@@ -455,13 +371,17 @@ static const struct pkt esp_enable_pkt = {
 
 static int queue_esp_control(struct openconnect_info *vpninfo, int enable)
 {
-	struct pkt *new = malloc(sizeof(*new) + 13);
+	struct pkt *new = alloc_pkt(vpninfo, esp_enable_pkt.len);
 	if (!new)
 		return -ENOMEM;
 
-	memcpy(new, &esp_enable_pkt, sizeof(*new) + 13);
+	new->oncp = esp_enable_pkt.oncp;
+	new->len = esp_enable_pkt.len;
+
+	memcpy(new->data, esp_enable_pkt.data, esp_enable_pkt.len);
 	new->data[12] = enable;
-	queue_packet(&vpninfo->oncp_control_queue, new);
+
+	queue_packet(&vpninfo->tcp_control_queue, new);
 	return 0;
 }
 
@@ -481,6 +401,8 @@ static int parse_conf_pkt(struct openconnect_info *vpninfo, unsigned char *bytes
 	int kmplen, kmpend, grouplen, groupend, group, attr, attrlen;
 	int ofs = 0;
 	int split_enc_hmac_keys = 0;
+	struct oc_vpn_option *new_opts = NULL;
+	struct oc_ip_info new_ip_info = {};
 
 	kmplen = load_be16(bytes + ofs + 18);
 	kmpend = ofs + kmplen;
@@ -489,6 +411,9 @@ static int parse_conf_pkt(struct openconnect_info *vpninfo, unsigned char *bytes
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to parse KMP message\n"));
 		dump_buf_hex(vpninfo, PRG_ERR, '<', bytes, pktlen);
+	einval:
+		free_optlist(new_opts);
+		free_split_routes(&new_ip_info);
 		return -EINVAL;
 	}
 
@@ -511,7 +436,7 @@ static int parse_conf_pkt(struct openconnect_info *vpninfo, unsigned char *bytes
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Received non-ESP TLVs (group %d) in ESP negotiation KMP\n"),
 				     group);
-			return -EINVAL;
+			goto einval;
 		}
 
 		while (ofs < groupend) {
@@ -522,7 +447,8 @@ static int parse_conf_pkt(struct openconnect_info *vpninfo, unsigned char *bytes
 			ofs += 6;
 			if (attrlen + ofs > groupend)
 				goto eparse;
-			if (process_attr(vpninfo, group, attr, bytes + ofs, attrlen))
+			if (process_attr(vpninfo, &new_opts, &new_ip_info,
+					 group, attr, bytes + ofs, attrlen))
 				goto eparse;
 			if (GRP_ATTR(group, attr)==GRP_ATTR(7, 2))
 				split_enc_hmac_keys = 1;
@@ -535,6 +461,13 @@ static int parse_conf_pkt(struct openconnect_info *vpninfo, unsigned char *bytes
 	if (split_enc_hmac_keys)
 		memcpy(vpninfo->esp_out.hmac_key, vpninfo->esp_out.enc_key + vpninfo->enc_key_len, vpninfo->hmac_key_len);
 
+	int ret = install_vpn_opts(vpninfo, new_opts, &new_ip_info);
+	if (ret) {
+		free_optlist(new_opts);
+		free_split_routes(&new_ip_info);
+		return ret;
+	}
+
 	return 0;
 }
 
@@ -545,11 +478,9 @@ int oncp_connect(struct openconnect_info *vpninfo)
 	struct oc_text_buf *reqbuf;
 	unsigned char bytes[65536];
 
-	/* XXX: We should do what cstp_connect() does to check that configuration
-	   hasn't changed on a reconnect. */
-
 	if (!vpninfo->cookies) {
-		ret = parse_cookie(vpninfo);
+		/* XX: This will happen if authentication was separate/external */
+		ret = internal_split_cookies(vpninfo, 0, "DSID");
 		if (ret)
 			return ret;
 	}
@@ -578,6 +509,8 @@ int oncp_connect(struct openconnect_info *vpninfo)
 		ret = buf_error(reqbuf);
 		goto out;
 	}
+	if (vpninfo->dump_http_traffic)
+		dump_buf(vpninfo, '>', reqbuf->data);
 	ret = vpninfo->ssl_write(vpninfo, reqbuf->data, reqbuf->pos);
 	if (ret < 0)
 		goto out;
@@ -590,7 +523,7 @@ int oncp_connect(struct openconnect_info *vpninfo)
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Unexpected %d result from server\n"),
 			     ret);
-		ret = -EINVAL;
+		ret = ((ret >= 400 && ret <= 499) || ret == 302) ? -EPERM : -EINVAL;
 		goto out;
 	}
 
@@ -782,13 +715,16 @@ int oncp_connect(struct openconnect_info *vpninfo)
 	vpn_progress(vpninfo, PRG_DEBUG, _("oNCP negotiation request outgoing:\n"));
 	dump_buf_hex(vpninfo, PRG_DEBUG, '>', (void *)reqbuf->data, reqbuf->pos);
 	ret = vpninfo->ssl_write(vpninfo, reqbuf->data, reqbuf->pos);
-	if (ret == reqbuf->pos)
-		ret = 0;
-	else if (ret >= 0) {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("Short write in oNCP negotiation\n"));
-		ret = -EIO;
+	if (ret != reqbuf->pos) {
+		if (ret >= 0) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Short write in oNCP negotiation\n"));
+			ret = -EIO;
+		}
+		goto out;
 	}
+	ret = 0;
+
  out:
 	if (ret)
 		openconnect_close_https(vpninfo, 0);
@@ -799,8 +735,8 @@ int oncp_connect(struct openconnect_info *vpninfo)
 	}
 	buf_free(reqbuf);
 
-	vpninfo->oncp_rec_size = 0;
-	free(vpninfo->cstp_pkt);
+	vpninfo->partial_rec_size = 0;
+	free_pkt(vpninfo, vpninfo->cstp_pkt);
 	vpninfo->cstp_pkt = NULL;
 
 	return ret;
@@ -831,7 +767,7 @@ static int oncp_receive_espkeys(struct openconnect_info *vpninfo, int len)
 		store_le16(vpninfo->cstp_pkt->oncp.rec,
 			   (p - vpninfo->cstp_pkt->oncp.kmp));
 
-		queue_packet(&vpninfo->oncp_control_queue, vpninfo->cstp_pkt);
+		queue_packet(&vpninfo->tcp_control_queue, vpninfo->cstp_pkt);
 		vpninfo->cstp_pkt = NULL;
 
 		print_esp_keys(vpninfo, _("new incoming"), esp);
@@ -850,10 +786,10 @@ static int oncp_record_read(struct openconnect_info *vpninfo, void *buf, int len
 {
 	int ret;
 
-	if (!vpninfo->oncp_rec_size) {
+	if (!vpninfo->partial_rec_size) {
 		unsigned char lenbuf[2];
 
-		ret = ssl_nonblock_read(vpninfo, lenbuf, 2);
+		ret = ssl_nonblock_read(vpninfo, 0, lenbuf, 2);
 		if (ret <= 0)
 			return ret;
 		if (ret == 1) {
@@ -863,34 +799,39 @@ static int oncp_record_read(struct openconnect_info *vpninfo, void *buf, int len
 				     _("Read only 1 byte of oNCP length field\n"));
 			return -EIO;
 		}
-		vpninfo->oncp_rec_size = load_le16(lenbuf);
-		if (!vpninfo->oncp_rec_size) {
-			ret = ssl_nonblock_read(vpninfo, lenbuf, 1);
+		vpninfo->partial_rec_size = load_le16(lenbuf);
+		if (!vpninfo->partial_rec_size) {
+			ret = ssl_nonblock_read(vpninfo, 0, lenbuf, 1);
 			if (ret == 1) {
 				if (lenbuf[0] == 1) {
 					vpn_progress(vpninfo, PRG_ERR,
 						     _("Server terminated connection (session expired)\n"));
 					vpninfo->quit_reason = "VPN session expired";
+				} else if (lenbuf[0] == 8) {
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Server terminated connection (idle timeout)\n"));
+					vpninfo->quit_reason = "Idle timeout";
 				} else {
 					vpn_progress(vpninfo, PRG_ERR,
 						     _("Server terminated connection (reason: %d)\n"),
 						     lenbuf[0]);
 					vpninfo->quit_reason = "Server terminated connection";
 				}
+				return -EPIPE;
 			} else {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Server sent zero-length oNCP record\n"));
 				vpninfo->quit_reason = "Zero-length oNCP record";
+				return -EIO;
 			}
-			return -EIO;
 		}
 	}
-	if (len > vpninfo->oncp_rec_size)
-		len = vpninfo->oncp_rec_size;
+	if (len > vpninfo->partial_rec_size)
+		len = vpninfo->partial_rec_size;
 
-	ret = ssl_nonblock_read(vpninfo, buf, len);
+	ret = ssl_nonblock_read(vpninfo, 0, buf, len);
 	if (ret > 0)
-		vpninfo->oncp_rec_size -= ret;
+		vpninfo->partial_rec_size -= ret;
 	return ret;
 }
 
@@ -917,13 +858,13 @@ int oncp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 	while (readable) {
 		int len, kmp, kmplen, iplen;
 		/* Some servers send us packets that are larger than
-		   negitiated MTU. We reserve some estra space to
+		   negotiated MTU. We reserve some extra space to
 		   handle that */
 		int receive_mtu = MAX(16384, vpninfo->ip_info.mtu);
 
 		len = receive_mtu + vpninfo->pkt_trailer;
 		if (!vpninfo->cstp_pkt) {
-			vpninfo->cstp_pkt = malloc(sizeof(struct pkt) + len);
+			vpninfo->cstp_pkt = alloc_pkt(vpninfo, len);
 			if (!vpninfo->cstp_pkt) {
 				vpn_progress(vpninfo, PRG_ERR, _("Allocation failed\n"));
 				break;
@@ -1028,7 +969,8 @@ int oncp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 			}
 
 			/* OK, we have a whole packet, and we have stuff after it */
-			queue_new_packet(&vpninfo->incoming_queue, vpninfo->cstp_pkt->data, iplen);
+			queue_new_packet(vpninfo, &vpninfo->incoming_queue,
+					 vpninfo->cstp_pkt->data, iplen);
 			kmplen -= iplen;
 			if (kmplen) {
 				/* Still data packets to come in this KMP300 */
@@ -1095,7 +1037,7 @@ int oncp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 			     vpninfo->current_ssl_pkt->oncp.rec,
 			     vpninfo->current_ssl_pkt->len + 22);
 
-		ret = ssl_nonblock_write(vpninfo,
+		ret = ssl_nonblock_write(vpninfo, 0,
 					 vpninfo->current_ssl_pkt->oncp.rec,
 					 vpninfo->current_ssl_pkt->len + 22);
 		if (ret < 0) {
@@ -1129,7 +1071,7 @@ int oncp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 				return work_done;
 			default:
 				/* This should never happen */
-				;
+				break;
 			}
 #else
 			return work_done;
@@ -1145,7 +1087,7 @@ int oncp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 		}
 		/* Don't free the 'special' packets */
 		if (vpninfo->current_ssl_pkt == vpninfo->deflate_pkt) {
-			free(vpninfo->pending_deflated_pkt);
+			free_pkt(vpninfo, vpninfo->pending_deflated_pkt);
 			vpninfo->pending_deflated_pkt = NULL;
 		} else if (vpninfo->current_ssl_pkt == &esp_enable_pkt) {
 			/* Only set the ESP state to connected and actually start
@@ -1153,10 +1095,10 @@ int oncp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 			   *sent* over the TCP channel. */
 			vpn_progress(vpninfo, PRG_TRACE,
 				     _("Sent ESP enable control packet\n"));
-			vpninfo->dtls_state = DTLS_CONNECTED;
+			vpninfo->dtls_state = DTLS_ESTABLISHED;
 			work_done = 1;
 		} else {
-			free(vpninfo->current_ssl_pkt);
+			free_pkt(vpninfo, vpninfo->current_ssl_pkt);
 		}
 		vpninfo->current_ssl_pkt = NULL;
 	}
@@ -1219,7 +1161,7 @@ int oncp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 	case KA_KEEPALIVE:
 		/* No need to send an explicit keepalive
 		   if we have real data to send */
-		if (vpninfo->dtls_state != DTLS_CONNECTED && vpninfo->outgoing_queue)
+		if (vpninfo->dtls_state != DTLS_ESTABLISHED && vpninfo->outgoing_queue)
 			break;
 
 		vpn_progress(vpninfo, PRG_DEBUG, _("Send CSTP Keepalive\n"));
@@ -1235,17 +1177,17 @@ int oncp_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 	 * via ESP once the enable message has been *sent* over the
 	 * TCP channel. Assign it directly to current_ssl_pkt so that
 	 * we can use it in-place and match against it above. */
-	if (vpninfo->dtls_state == DTLS_CONNECTING) {
+	if (vpninfo->dtls_state == DTLS_CONNECTED) {
 		vpninfo->current_ssl_pkt = (struct pkt *)&esp_enable_pkt;
 		goto handle_outgoing;
 	}
 
-	vpninfo->current_ssl_pkt = dequeue_packet(&vpninfo->oncp_control_queue);
+	vpninfo->current_ssl_pkt = dequeue_packet(&vpninfo->tcp_control_queue);
 	if (vpninfo->current_ssl_pkt)
 		goto handle_outgoing;
 
 	/* Service outgoing packet queue, if no DTLS */
-	while (vpninfo->dtls_state != DTLS_CONNECTED &&
+	while (vpninfo->dtls_state != DTLS_ESTABLISHED &&
 	       (vpninfo->current_ssl_pkt = dequeue_packet(&vpninfo->outgoing_queue))) {
 		struct pkt *this = vpninfo->current_ssl_pkt;
 
@@ -1279,7 +1221,7 @@ int oncp_bye(struct openconnect_info *vpninfo, const char *reason)
 
 	orig_path = vpninfo->urlpath;
 	vpninfo->urlpath = strdup("dana-na/auth/logout.cgi"); /* redirect segfaults without strdup */
-	ret = do_https_request(vpninfo, "GET", NULL, NULL, &res_buf, 0);
+	ret = do_https_request(vpninfo, "GET", NULL, NULL, &res_buf, NULL, HTTP_NO_FLAGS);
 	free(vpninfo->urlpath);
 	vpninfo->urlpath = orig_path;
 
@@ -1296,7 +1238,7 @@ int oncp_bye(struct openconnect_info *vpninfo, const char *reason)
 void oncp_esp_close(struct openconnect_info *vpninfo)
 {
 	/* Tell server to stop sending on ESP channel */
-	if (vpninfo->dtls_state >= DTLS_CONNECTING)
+	if (vpninfo->dtls_state >= DTLS_CONNECTED)
 		queue_esp_control(vpninfo, 0);
 	esp_close(vpninfo);
 }
@@ -1319,11 +1261,11 @@ int oncp_esp_send_probes(struct openconnect_info *vpninfo)
 		monitor_except_fd(vpninfo, dtls);
 	}
 
-	pkt = malloc(sizeof(*pkt) + 1 + vpninfo->pkt_trailer);
+	pkt = alloc_pkt(vpninfo, 1 + vpninfo->pkt_trailer);
 	if (!pkt)
 		return -ENOMEM;
 
-	for (seq=1; seq <= (vpninfo->dtls_state==DTLS_CONNECTED ? 1 : 2); seq++) {
+	for (seq=1; seq <= (vpninfo->dtls_state==DTLS_ESTABLISHED ? 1 : 2); seq++) {
 		pkt->len = 1;
 		pkt->data[0] = 0;
 		pktlen = construct_esp_packet(vpninfo, pkt,
@@ -1332,7 +1274,7 @@ int oncp_esp_send_probes(struct openconnect_info *vpninfo)
 		    send(vpninfo->dtls_fd, (void *)&pkt->esp, pktlen, 0) < 0)
 			vpn_progress(vpninfo, PRG_DEBUG, _("Failed to send ESP probe\n"));
 	}
-	free(pkt);
+	free_pkt(vpninfo, pkt);
 
 	vpninfo->dtls_times.last_tx = time(&vpninfo->new_dtls_started);
 
