@@ -17,20 +17,20 @@
 
 #include <config.h>
 
-#include <errno.h>
-#include <sys/types.h>
+#include "openconnect-internal.h"
+
 #include <unistd.h>
+#include <sys/types.h>
 #include <fcntl.h>
-#include <string.h>
-#include <errno.h>
-#include <stdlib.h>
-#include <stdio.h>
 #ifndef _WIN32
 #include <netinet/in.h>
 #include <sys/socket.h>
 #endif
 
-#include "openconnect-internal.h"
+#include <errno.h>
+#include <string.h>
+#include <stdlib.h>
+#include <stdio.h>
 
 /* In the very early days there were cases where this wasn't found in
  * the header files but it did still work somehow. I forget the details
@@ -331,9 +331,15 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 	int dtlsver = DTLS1_BAD_VER;
 	const char *cipher = vpninfo->dtls_cipher;
 
+	if (!cipher) {
+		dtlsver = 0;
 #ifdef HAVE_DTLS12
+		/* The F5 BIG-IP server before v16, will crap itself if we
+		 * even *try* to do DTLSv1.2 */
+		if (!vpninfo->dtls12)
+			dtlsver = DTLS1_VERSION;
 	/* These things should never happen unless they're supported */
-	if (vpninfo->cisco_dtls12) {
+	} else if (vpninfo->dtls12) {
 		dtlsver = DTLS1_2_VERSION;
 	} else if (!strcmp(cipher, "OC-DTLS1_2-AES128-GCM")) {
 		dtlsver = DTLS1_2_VERSION;
@@ -345,19 +351,17 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 	} else if (!strcmp(cipher, "PSK-NEGOTIATE")) {
 		dtlsver = 0; /* Let it negotiate */
 #endif
-	}
 #endif
+	}
 
 	if (!vpninfo->dtls_ctx) {
 #ifdef HAVE_DTLS12
 		/* If we can use SSL_CTX_set_min_proto_version, do so. */
 		dtls_method = DTLS_client_method();
-#endif
-#ifndef HAVE_SSL_CTX_PROTOVER
-		/* If !HAVE_DTLS12, dtlsver *MUST* be DTLS1_BAD_VER because it's set
-		 * at the top of the function and nothing can change it. */
-		if (dtlsver == DTLS1_BAD_VER)
-			dtls_method = DTLSv1_client_method();
+#elif !defined(HAVE_SSL_CTX_PROTOVER)
+		dtls_method = DTLSv1_client_method();
+#else
+#error "This can't happen as DTLS1.2 came before SSL_CTX_set_mXX_proto_version()"
 #endif
 
 		vpninfo->dtls_ctx = SSL_CTX_new(dtls_method);
@@ -386,75 +390,99 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 		if (dtlsver == DTLS1_BAD_VER)
 			SSL_CTX_set_options(vpninfo->dtls_ctx, SSL_OP_CISCO_ANYCONNECT);
 #endif
+		/* If we don't readahead, then we do short reads and throw
+		   away the tail of data packets. */
+		SSL_CTX_set_read_ahead(vpninfo->dtls_ctx, 1);
+
+		if (vpninfo->proto->proto == PROTO_ANYCONNECT) {
+			/* All the AnyConnect hackery about saved sessions and PSK */
 #if defined (HAVE_DTLS12) && !defined(OPENSSL_NO_PSK)
-		if (!dtlsver) {
-			SSL_CTX_set_psk_client_callback(vpninfo->dtls_ctx, psk_callback);
-			/* For PSK we override the DTLS master secret with one derived
-			 * from the HTTPS session. */
-			if (!SSL_export_keying_material(vpninfo->https_ssl,
-							vpninfo->dtls_secret, PSK_KEY_SIZE,
-							PSK_LABEL, PSK_LABEL_SIZE, NULL, 0, 0)) {
+			if (!dtlsver) {
+				SSL_CTX_set_psk_client_callback(vpninfo->dtls_ctx, psk_callback);
+				/* For PSK we override the DTLS master secret with one derived
+				 * from the HTTPS session. */
+				if (!SSL_export_keying_material(vpninfo->https_ssl,
+								vpninfo->dtls_secret, PSK_KEY_SIZE,
+								PSK_LABEL, PSK_LABEL_SIZE, NULL, 0, 0)) {
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Failed to generate DTLS key\n"));
+					openconnect_report_ssl_errors(vpninfo);
+					SSL_CTX_free(vpninfo->dtls_ctx);
+					vpninfo->dtls_ctx = NULL;
+					vpninfo->dtls_attempt_period = 0;
+					return -EINVAL;
+				}
+				/* For SSL_CTX_set_cipher_list() */
+				cipher = "PSK";
+			}
+#endif /* OPENSSL_NO_PSK */
+#ifdef SSL_OP_NO_ENCRYPT_THEN_MAC
+			/*
+			 * I'm fairly sure I wasn't lying when I said I had tested
+			 * https://github.com/openssl/openssl/commit/e23d5071ec4c7aa6bb2b
+			 * against GnuTLS both with and without EtM in 2016.
+			 *
+			 * Nevertheless, in 2019 it seems to be failing to negotiate
+			 * at least for DTLS1_BAD_VER against ocserv with GnuTLS 3.6.7:
+			 * https://gitlab.com/gnutls/gnutls/issues/139 — I think because
+			 * GnuTLS isn't actually doing EtM after negotiating it (like
+			 * OpenSSL 1.1.0 used to).
+			 *
+			 * Just turn it off. Real Cisco servers don't do it for
+			 * DTLS1_BAD_VER, and against ocserv (and newer Cisco) we should
+			 * be using DTLSv1.2 with AEAD ciphersuites anyway so EtM is
+			 * irrelevant.
+			 */
+			SSL_CTX_set_options(vpninfo->dtls_ctx, SSL_OP_NO_ENCRYPT_THEN_MAC);
+#endif
+#ifdef SSL_OP_LEGACY_SERVER_CONNECT
+			/*
+			 * Since https://github.com/openssl/openssl/pull/15127, OpenSSL
+			 * *requires* secure renegotiation support by default. For interop
+			 * with Cisco's resumed DTLS sessions, we have to turn that off.
+			 */
+			if (dtlsver)
+				SSL_CTX_set_options(vpninfo->dtls_ctx, SSL_OP_LEGACY_SERVER_CONNECT);
+#endif
+#ifdef SSL_OP_NO_EXTENDED_MASTER_SECRET
+			/* RFC7627 says:
+			 *
+			 *   If the original session did not use the "extended_master_secret"
+			 *   extension but the new ClientHello contains the extension, then the
+			 *   server MUST NOT perform the abbreviated handshake.  Instead, it
+			 *   SHOULD continue with a full handshake (as described in
+			 *   Section 5.2) to negotiate a new session.
+			 *
+			 * Now that would be distinctly suboptimal, since we have no way to do
+			 * a full handshake (we even explicitly protect against it, in case a
+			 * MITM server attempts to hijack our deliberately-resumed session).
+			 *
+			 * So where OpenSSL provides the choice, tell it not to use extms on
+			 * resumed sessions.
+			 */
+			if (dtlsver)
+				SSL_CTX_set_options(vpninfo->dtls_ctx, SSL_OP_NO_EXTENDED_MASTER_SECRET);
+#endif
+			if (!SSL_CTX_set_cipher_list(vpninfo->dtls_ctx, cipher)) {
 				vpn_progress(vpninfo, PRG_ERR,
-					     _("Failed to generate DTLS key\n"));
-				openconnect_report_ssl_errors(vpninfo);
+					     _("Set DTLS cipher list failed\n"));
 				SSL_CTX_free(vpninfo->dtls_ctx);
 				vpninfo->dtls_ctx = NULL;
 				vpninfo->dtls_attempt_period = 0;
 				return -EINVAL;
 			}
-			/* For SSL_CTX_set_cipher_list() */
-			cipher = "PSK";
-		}
-#endif /* OPENSSL_NO_PSK */
-#ifdef SSL_OP_NO_ENCRYPT_THEN_MAC
-		/*
-		 * I'm fairly sure I wasn't lying when I said I had tested
-		 * https://github.com/openssl/openssl/commit/e23d5071ec4c7aa6bb2b
-		 * against GnuTLS both with and without EtM in 2016.
-		 *
-		 * Nevertheless, in 2019 it seems to be failing to negotiate
-		 * at least for DTLS1_BAD_VER against ocserv with GnuTLS 3.6.7:
-		 * https://gitlab.com/gnutls/gnutls/issues/139 — I think because
-		 * GnuTLS isn't actually doing EtM after negotiating it (like
-		 * OpenSSL 1.1.0 used to).
-		 *
-		 * Just turn it off. Real Cisco servers don't do it for
-		 * DTLS1_BAD_VER, and against ocserv (and newer Cisco) we should
-		 * be using DTLSv1.2 with AEAD ciphersuites anyway so EtM is
-		 * irrelevant.
-		 */
-		SSL_CTX_set_options(vpninfo->dtls_ctx, SSL_OP_NO_ENCRYPT_THEN_MAC);
-#endif
-#ifdef SSL_OP_NO_EXTENDED_MASTER_SECRET
-		/* RFC7627 says:
-		 *
-		 *   If the original session did not use the "extended_master_secret"
-		 *   extension but the new ClientHello contains the extension, then the
-		 *   server MUST NOT perform the abbreviated handshake.  Instead, it
-		 *   SHOULD continue with a full handshake (as described in
-		 *   Section 5.2) to negotiate a new session.
-		 *
-		 * Now that would be distinctly suboptimal, since we have no way to do
-		 * a full handshake (we even explicitly protect against it, in case a
-		 * MITM server attempts to hijack our deliberately-resumed session).
-		 *
-		 * So where OpenSSL provides the choice, tell it not to use extms on
-		 * resumed sessions.
-		 */
-		if (dtlsver)
-			SSL_CTX_set_options(vpninfo->dtls_ctx, SSL_OP_NO_EXTENDED_MASTER_SECRET);
-#endif
-		/* If we don't readahead, then we do short reads and throw
-		   away the tail of data packets. */
-		SSL_CTX_set_read_ahead(vpninfo->dtls_ctx, 1);
-
-		if (!SSL_CTX_set_cipher_list(vpninfo->dtls_ctx, cipher)) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Set DTLS cipher list failed\n"));
-			SSL_CTX_free(vpninfo->dtls_ctx);
-			vpninfo->dtls_ctx = NULL;
-			vpninfo->dtls_attempt_period = 0;
-			return -EINVAL;
+		} else {
+			/* Protocols other than AnyConnect are plain DTLS and do
+			 * need to check the server certificate properly (which
+			 * AnyConnect can skip because it all depends on PSK or
+			 * session resumption anyway */
+			int ret = openconnect_install_ctx_verify(vpninfo, vpninfo->dtls_ctx);
+			if (ret) {
+				SSL_CTX_free(vpninfo->dtls_ctx);
+				vpninfo->dtls_ctx = NULL;
+				vpninfo->dtls_attempt_period = 0;
+				return ret;
+			}
 		}
 	}
 
@@ -462,8 +490,20 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 	SSL_set_connect_state(dtls_ssl);
 	SSL_set_app_data(dtls_ssl, vpninfo);
 
-
-	if (dtlsver) {
+	/*
+	 * AnyConnect always sets cipher; nothing else does. Checking this
+	 * instead of vpninfo->proto->proto == PROTO_ANYCONNECT makes the
+	 * static analyser less unhappy because it doesn't know that, so it
+	 * would think we can dereference cipher when it's NULL.
+	 */
+	if (!cipher) {
+		/* Non-AnyConnect protocols need to verify the peer */
+		SSL_set_verify(dtls_ssl, SSL_VERIFY_PEER, NULL);
+		/* Where they only do DTLSv1, they also don't cope with secure renegotiation */
+		if (dtlsver == DTLS1_VERSION)
+			SSL_set_options(dtls_ssl, SSL_OP_LEGACY_SERVER_CONNECT);
+	} else if (dtlsver) {
+		/* This is the actual Cisco AnyConnect method, using session resume */
 		STACK_OF(SSL_CIPHER) *ciphers = SSL_get_ciphers(dtls_ssl);
 		const SSL_CIPHER *ssl_ciph = NULL;
 		int i;
@@ -497,11 +537,11 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 
 		if (!SSL_set_session(dtls_ssl, dtls_session)) {
 			vpn_progress(vpninfo, PRG_ERR,
-				     _("SSL_set_session() failed with old protocol version 0x%x\n"
+				     _("SSL_set_session() failed with DTLS protocol version 0x%x\n"
 				       "Are you using a version of OpenSSL older than 0.9.8m?\n"
 				       "See http://rt.openssl.org/Ticket/Display.html?id=1751\n"
 				       "Use the --no-dtls command line option to avoid this message\n"),
-				     DTLS1_BAD_VER);
+				     dtlsver);
 			SSL_CTX_free(vpninfo->dtls_ctx);
 			SSL_free(dtls_ssl);
 			vpninfo->dtls_ctx = NULL;
@@ -513,10 +553,13 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 		SSL_SESSION_free(dtls_session);
 
 	} else if (vpninfo->dtls_app_id_size > 0) {
+		/*
+		 * For ocserv PSK-NEGOTIATE we abuse the session resume
+		 * protocol just to pass an 'App ID' in our ClientHello
+		 * in the session identifier. This session doesn't exist
+		 * and isn't actually going to be resumed at all.
+		 */
 		const uint8_t cs[2] = {0x00, 0x2F}; /* RSA-AES-128 */
-		/* we generate a session with a random key which cannot be resumed;
-		 * we want to set the client identifier we received from the server
-		 * as a session ID. */
 		dtls_session = generate_dtls_session(vpninfo, DTLS1_VERSION,
 						     SSL_CIPHER_find(dtls_ssl, cs),
 						     1);
@@ -527,7 +570,7 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 			vpninfo->dtls_attempt_period = 0;
 			return -EINVAL;
 		}
-	
+
 		if (!SSL_set_session(dtls_ssl, dtls_session)) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("SSL_set_session() failed\n"));
@@ -540,10 +583,24 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 		}
 		/* We don't need our own refcount on it any more */
 		SSL_SESSION_free(dtls_session);
+	} /* else it's ocserv PSK-NEGOTIATE without an App-ID */
+
+	dtls_bio = BIO_new_dgram(dtls_fd, BIO_NOCLOSE);
+	if (!dtls_bio || !BIO_ctrl(dtls_bio, BIO_CTRL_DGRAM_SET_CONNECTED,
+				  0, (char *)vpninfo->dtls_addr)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Create DTLS dgram BIO failed\n"));
+
+		if (dtls_bio)
+			BIO_free(dtls_bio);
+
+		SSL_CTX_free(vpninfo->dtls_ctx);
+		SSL_free(dtls_ssl);
+		vpninfo->dtls_ctx = NULL;
+		vpninfo->dtls_attempt_period = 0;
+		return -EIO;
 	}
 
-
-	dtls_bio = BIO_new_socket(dtls_fd, BIO_NOCLOSE);
 	/* Set non-blocking */
 	BIO_set_nbio(dtls_bio, 1);
 	SSL_set_bio(dtls_ssl, dtls_bio, dtls_bio);
@@ -553,14 +610,28 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 	return 0;
 }
 
-int dtls_try_handshake(struct openconnect_info *vpninfo)
+int dtls_try_handshake(struct openconnect_info *vpninfo, int *timeout)
 {
 	int ret = SSL_do_handshake(vpninfo->dtls_ssl);
 
 	if (ret == 1) {
 		const char *c;
 
-		if (!strcmp(vpninfo->dtls_cipher, "PSK-NEGOTIATE")) {
+		if (!vpninfo->dtls_cipher) {
+			/* Anonymous DTLS (PPP protocols) will set vpninfo->ip_info.mtu
+			 * in PPP negotiation.
+			 *
+			 * XX: Needs forthcoming overhaul to detect MTU correctly and offer
+			 * reasonable MRU values during PPP negotiation.
+			 */
+			int data_mtu = vpninfo->cstp_basemtu = 1500;
+			if (vpninfo->peer_addr->sa_family == IPPROTO_IPV6)
+				data_mtu -= 40; /* IPv6 header */
+			else
+				data_mtu -= 20; /* Legacy IP header */
+			data_mtu -= 8; /* UDP header */
+			dtls_set_mtu(vpninfo, data_mtu);
+		} else if (!strcmp(vpninfo->dtls_cipher, "PSK-NEGOTIATE")) {
 			/* For PSK-NEGOTIATE, we have to determine the tunnel MTU
 			 * for ourselves based on the base MTU */
 			int data_mtu = vpninfo->cstp_basemtu;
@@ -612,7 +683,7 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 				     _("DTLS connection compression using %s.\n"), c);
 		}
 
-		vpninfo->dtls_times.last_rekey = vpninfo->dtls_times.last_rx = 
+		vpninfo->dtls_times.last_rekey = vpninfo->dtls_times.last_rx =
 			vpninfo->dtls_times.last_tx = time(NULL);
 
 		/* From about 8.4.1(11) onwards, the ASA seems to get
@@ -640,7 +711,7 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 		   version, warn about it. */
 		if (SSLeay() < 0x1000005fL) {
 			vpn_progress(vpninfo, PRG_ERR,
-				     _("Your OpenSSL is older than the one you built against, so DTLS may fail!"));
+				     _("Your OpenSSL is older than the one you built against, so DTLS may fail!\n"));
 		}
 #elif defined(HAVE_DTLS1_STOP_TIMER)
 		/*
@@ -663,7 +734,7 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 			 &(vpninfo->dtls_ssl->d1->next_timeout));
 #elif defined(BIO_CTRL_DGRAM_SET_TIMEOUT)
 		/*
-		 * OK, here it gets more fun... this shoul handle the case
+		 * OK, here it gets more fun... this should handle the case
 		 * of older OpenSSL which has the Cisco DTLS compatibility
 		 * backported, but *not* the fix for RT#1922.
 		 */
@@ -683,11 +754,26 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 
 	ret = SSL_get_error(vpninfo->dtls_ssl, ret);
 	if (ret == SSL_ERROR_WANT_WRITE || ret == SSL_ERROR_WANT_READ) {
-		static int badossl_bitched = 0;
-		if (time(NULL) < vpninfo->new_dtls_started + 12)
+		int quit_time = vpninfo->new_dtls_started + 12 - time(NULL);
+
+		if (quit_time > 0) {
+			if (timeout) {
+				struct timeval tv;
+				if (SSL_ctrl(vpninfo->dtls_ssl, DTLS_CTRL_GET_TIMEOUT, 0, &tv)) {
+					unsigned timeout_ms = tv.tv_sec * 1000 + tv.tv_usec / 1000;
+					if (*timeout > timeout_ms)
+						*timeout = timeout_ms;
+				}
+
+				if (*timeout > quit_time * 1000)
+					*timeout = quit_time * 1000;
+			}
 			return 0;
-		if (((OPENSSL_VERSION_NUMBER >= 0x100000b0L && OPENSSL_VERSION_NUMBER <= 0x100000c0L) || \
-		     (OPENSSL_VERSION_NUMBER >= 0x10001040L && OPENSSL_VERSION_NUMBER <= 0x10001060L) || \
+		}
+
+		static int badossl_bitched; /* static variable initialised to 0 */
+		if (((OPENSSL_VERSION_NUMBER >= 0x100000b0L && OPENSSL_VERSION_NUMBER <= 0x100000c0L) ||
+		     (OPENSSL_VERSION_NUMBER >= 0x10001040L && OPENSSL_VERSION_NUMBER <= 0x10001060L) ||
 		     OPENSSL_VERSION_NUMBER == 0x10002000L) && !badossl_bitched) {
 			badossl_bitched = 1;
 			vpn_progress(vpninfo, PRG_ERR, _("DTLS handshake timed out\n"));
@@ -705,6 +791,8 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 
 	vpninfo->dtls_state = DTLS_SLEEPING;
 	time(&vpninfo->new_dtls_started);
+	if (timeout && *timeout > vpninfo->dtls_attempt_period * 1000)
+		*timeout = vpninfo->dtls_attempt_period * 1000;
 	return -EINVAL;
 }
 
@@ -754,6 +842,8 @@ void gather_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *b
 		return;
 	}
 
+	int aes128_gcm = 0, aes256_gcm = 0;
+
 	ciphers = SSL_get1_supported_ciphers(ssl);
 	for (i = 0; i < sk_SSL_CIPHER_num(ciphers); i++) {
 		const SSL_CIPHER *ciph = sk_SSL_CIPHER_value(ciphers, i);
@@ -769,7 +859,15 @@ void gather_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *b
 			buf_append(buf12, "%s%s",
 				   (buf_error(buf12) || !buf12->pos) ? "" : ":",
 				   name);
+			/* The OC-specific names for the DTLSv1.2 AES-GCM ciphersuites
+			 * need to be added to the X-DTLS-CipherSuite: header too. */
+			if (!strcmp(name, "AES128-GCM-SHA256")) {
+				aes128_gcm = 1;
+			} else if (!strcmp(name, "AES256-GCM-SHA384")) {
+				aes256_gcm = 1;
+			}
 		}
+
 	}
 	sk_SSL_CIPHER_free(ciphers);
 	SSL_free(ssl);
@@ -778,6 +876,10 @@ void gather_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *b
 	/* All DTLSv1 suites are also supported in DTLSv1.2 */
 	if (!buf_error(buf))
 		buf_append(buf12, ":%s", buf->data);
+	if (aes128_gcm)
+		buf_append(buf, ":OC-DTLS1_2-AES128-GCM");
+	if (aes256_gcm)
+		buf_append(buf, ":OC-DTLS1_2-AES256-GCM");
 #ifndef OPENSSL_NO_PSK
 	buf_append(buf, ":PSK-NEGOTIATE");
 #endif
