@@ -17,19 +17,19 @@
 
 #include <config.h>
 
+#include "openconnect-internal.h"
+
 #include <unistd.h>
 #include <fcntl.h>
+#include <sys/types.h>
+
 #include <time.h>
 #include <string.h>
 #include <ctype.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include <sys/types.h>
 #include <stdarg.h>
-#include <sys/types.h>
-
-#include "openconnect-internal.h"
 
 #define VENDOR_JUNIPER 0xa4c
 #define VENDOR_JUNIPER2 0x583
@@ -95,24 +95,6 @@
 #define TTLS_MOREFRAGS	(1<<6)
 #define TTLS_START	(1<<5)
 
-static void buf_append_be16(struct oc_text_buf *buf, uint16_t val)
-{
-	unsigned char b[2];
-
-	store_be16(b, val);
-
-	buf_append_bytes(buf, b, 2);
-}
-
-static void buf_append_be32(struct oc_text_buf *buf, uint32_t val)
-{
-	unsigned char b[4];
-
-	store_be32(b, val);
-
-	buf_append_bytes(buf, b, 4);
-}
-
 static void buf_append_ift_hdr(struct oc_text_buf *buf, uint32_t vendor, uint32_t type)
 {
 	uint32_t b[4];
@@ -235,38 +217,8 @@ static int valid_ift_auth_eap_exj1(unsigned char *bytes, int len)
 	return 1;
 }
 
-/* We behave like CSTP — create a linked list in vpninfo->cstp_options
- * with the strings containing the information we got from the server,
- * and oc_ip_info contains const copies of those pointers. */
-
-static const char *add_option(struct openconnect_info *vpninfo, const char *opt,
-			      const char *val, int val_len)
-{
-	struct oc_vpn_option *new = malloc(sizeof(*new));
-	if (!new)
-		return NULL;
-
-	new->option = strdup(opt);
-	if (!new->option) {
-		free(new);
-		return NULL;
-	}
-	if (val_len >= 0)
-		new->value = strndup(val, val_len);
-	else
-		new->value = strdup(val);
-	if (!new->value) {
-		free(new->option);
-		free(new);
-		return NULL;
-	}
-	new->next = vpninfo->cstp_options;
-	vpninfo->cstp_options = new;
-
-	return new->value;
-}
-
-static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
+static int process_attr(struct openconnect_info *vpninfo, struct oc_vpn_option **new_opts,
+			struct oc_ip_info *new_ip_info, uint16_t type,
 			unsigned char *data, int attrlen)
 {
 	struct oc_split_include *xc;
@@ -281,7 +233,7 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 		snprintf(buf, sizeof(buf), "%d.%d.%d.%d", data[0], data[1], data[2], data[3]);
 
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received internal Legacy IP address %s\n"), buf);
-		vpninfo->ip_info.addr = add_option(vpninfo, "ipaddr", buf, -1);
+		new_ip_info->addr = add_option_dup(new_opts, "ipaddr", buf, -1);
 		break;
 
 	case 0x0002:
@@ -290,7 +242,7 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 		snprintf(buf, sizeof(buf), "%d.%d.%d.%d", data[0], data[1], data[2], data[3]);
 
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received netmask %s\n"), buf);
-		vpninfo->ip_info.netmask = add_option(vpninfo, "netmask", buf, -1);
+		new_ip_info->netmask = add_option_dup(new_opts, "netmask", buf, -1);
 		break;
 
 	case 0x0003:
@@ -301,8 +253,8 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received DNS server %s\n"), buf);
 
 		for (i = 0; i < 3; i++) {
-			if (!vpninfo->ip_info.dns[i]) {
-				vpninfo->ip_info.dns[i] = add_option(vpninfo, "DNS", buf, -1);
+			if (!new_ip_info->dns[i]) {
+				new_ip_info->dns[i] = add_option_dup(new_opts, "DNS", buf, -1);
 				break;
 			}
 		}
@@ -316,8 +268,8 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received WINS server %s\n"), buf);
 
 		for (i = 0; i < 3; i++) {
-			if (!vpninfo->ip_info.nbns[i]) {
-				vpninfo->ip_info.nbns[i] = add_option(vpninfo, "WINS", buf, -1);
+			if (!new_ip_info->nbns[i]) {
+				new_ip_info->nbns[i] = add_option_dup(new_opts, "WINS", buf, -1);
 				break;
 			}
 		}
@@ -331,12 +283,13 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 				     _("Failed to handle IPv6 address\n"));
 			return -EINVAL;
 		}
-		vpninfo->ip_info.addr6 = add_option(vpninfo, "ip6addr", buf, -1);
-
-		i = strlen(buf);
-		snprintf(buf + i, sizeof(buf) - i, "/%d", data[16]);
-		vpninfo->ip_info.netmask6 = add_option(vpninfo, "ip6netmask", buf, -1);
+		if (!vpninfo->disable_ipv6) {
+			new_ip_info->addr6 = add_option_dup(new_opts, "ip6addr", buf, -1);
 
+			i = strlen(buf);
+			snprintf(buf + i, sizeof(buf) - i, "/%d", data[16]);
+			new_ip_info->netmask6 = add_option_dup(new_opts, "ip6netmask", buf, -1);
+		}
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received internal IPv6 address %s\n"), buf);
 		break;
 
@@ -350,8 +303,8 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 		}
 
 		for (i = 0; i < 3; i++) {
-			if (!vpninfo->ip_info.dns[i]) {
-				vpninfo->ip_info.dns[i] = add_option(vpninfo, "DNS", buf, -1);
+			if (!new_ip_info->dns[i]) {
+				new_ip_info->dns[i] = add_option_dup(new_opts, "DNS", buf, -1);
 				break;
 			}
 		}
@@ -372,10 +325,10 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 
 		xc = malloc(sizeof(*xc));
 		if (xc) {
-			xc->route =  add_option(vpninfo, "split-include6", buf, -1);
+			xc->route =  add_option_dup(new_opts, "split-include6", buf, -1);
 			if (xc->route) {
-				xc->next = vpninfo->ip_info.split_includes;
-				vpninfo->ip_info.split_includes = xc;
+				xc->next = new_ip_info->split_includes;
+				new_ip_info->split_includes = xc;
 			} else
 				free(xc);
 		}
@@ -395,10 +348,10 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 
 		xc = malloc(sizeof(*xc));
 		if (xc) {
-			xc->route =  add_option(vpninfo, "split-exclude6", buf, -1);
+			xc->route =  add_option_dup(new_opts, "split-exclude6", buf, -1);
 			if (xc->route) {
-				xc->next = vpninfo->ip_info.split_excludes;
-				vpninfo->ip_info.split_excludes = xc;
+				xc->next = new_ip_info->split_excludes;
+				new_ip_info->split_excludes = xc;
 			} else
 				free(xc);
 		}
@@ -413,22 +366,22 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 				     attrlen, type);
 			return -EINVAL;
 		}
-		vpninfo->ip_info.mtu = load_be32(data);
+		new_ip_info->mtu = load_be32(data);
 		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("Received MTU %d from server\n"),
-			     vpninfo->ip_info.mtu);
+			     new_ip_info->mtu);
 		break;
 
 	case 0x4006:
 		if (!attrlen)
 			goto badlen;
 		if (!data[attrlen-1])
-		    attrlen--;
+			attrlen--;
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received DNS search domain %.*s\n"),
 			     attrlen, (char *)data);
-		vpninfo->ip_info.domain = add_option(vpninfo, "search", (char *)data, attrlen);
-		if (vpninfo->ip_info.domain) {
-			char *p = (char *)vpninfo->ip_info.domain;
+		new_ip_info->domain = add_option_dup(new_opts, "search", (char *)data, attrlen);
+		if (new_ip_info->domain) {
+			char *p = (char *)new_ip_info->domain;
 			while ((p = strchr(p, ',')))
 				*p = ' ';
 		}
@@ -442,7 +395,7 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 		vpn_progress(vpninfo, PRG_DEBUG, _("Received internal gateway address %s\n"), buf);
 		/* Hm, what are we supposed to do with this? It's a tunnel;
 		   having a gateway is meaningless. */
-		add_option(vpninfo, "ipaddr", buf, -1);
+		add_option_dup(new_opts, "gateway", buf, -1);
 		break;
 
 	case 0x4010: {
@@ -537,24 +490,22 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 		vpn_progress(vpninfo, PRG_DEBUG, _("ESP only: %d\n"),
 			     data[0]);
 		break;
-#if 0
-	case GRP_ATTR(7, 1):
-		if (attrlen != 4)
+
+	case 0x401e:
+		if (attrlen != 16)
 			goto badlen;
-		memcpy(&vpninfo->esp_out.spi, data, 4);
-		vpn_progress(vpninfo, PRG_DEBUG, _("ESP SPI (outbound): %x\n"),
-			     load_be32(data));
+		if (!inet_ntop(AF_INET6, data, buf, sizeof(buf))) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to handle IPv6 address\n"));
+			return -EINVAL;
+		}
+
+		vpn_progress(vpninfo, PRG_DEBUG, _("Received internal gateway IPv6 address %s\n"), buf);
+		/* Hm, what are we supposed to do with this? It's a tunnel;
+		   having a gateway is meaningless. */
+		add_option_dup(new_opts, "gateway6", buf, -1);
 		break;
 
-	case GRP_ATTR(7, 2):
-		if (attrlen != 0x40)
-			goto badlen;
-		/* data contains enc_key and hmac_key concatenated */
-		memcpy(vpninfo->esp_out.enc_key, data, 0x40);
-		vpn_progress(vpninfo, PRG_DEBUG, _("%d bytes of ESP secrets\n"),
-			     attrlen);
-		break;
-#endif
 	/* 0x4022: disable proxy
 	   0x400a: preserve proxy
 	   0x4008: proxy (string)
@@ -564,8 +515,6 @@ static int process_attr(struct openconnect_info *vpninfo, uint16_t type,
 	   0x401f:  tunnel routes with subnet access (also 4001 set)
 	   0x4020: Enforce IPv4
 	   0x4021: Enforce IPv6
-	   0x401e: Server IPv6 address
-	   0x000f: IPv6 netmask?
 	*/
 
 	default:
@@ -1382,9 +1331,6 @@ static int pulse_authenticate(struct openconnect_info *vpninfo, int connecting)
 	char *user2_prompt = NULL, *pass2_prompt = NULL;
 	int prompt_flags = PROMPT_PRIMARY | PROMPT_USERNAME | PROMPT_PASSWORD;
 
-	/* XXX: We should do what cstp_connect() does to check that configuration
-	   hasn't changed on a reconnect. */
-
 	ret = openconnect_open_https(vpninfo);
 	if (ret)
 		return ret;
@@ -1713,6 +1659,9 @@ static int pulse_authenticate(struct openconnect_info *vpninfo, int connecting)
 			if (failcode == 0x0d) {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Authentication failure: Account locked out\n"));
+			} else if (failcode == 0x0e) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Authentication failure: Client certificate required\n"));
 			} else {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Authentication failure: Code 0x%02x\n"),
@@ -1736,27 +1685,26 @@ static int pulse_authenticate(struct openconnect_info *vpninfo, int connecting)
 			val = load_be32(avp_p);
 
 			switch (val) {
-		case 1: /* Prompt for both username and password. */
-			prompt_flags = PROMPT_PASSWORD | PROMPT_USERNAME;
-			break;
-		case 3: /* Prompt for password.*/
-		case 15:
-			prompt_flags = PROMPT_PASSWORD;
-			break;
-		case 5: /* Prompt for username.*/
-			prompt_flags = PROMPT_USERNAME;
-			break;
-
-		default:
-			/* It does no harm to submit both, as anything unwanted is ignored. */
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Unknown D73 prompt value 0x%x. Will prompt for both username and password.\n"),
-				     val);
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Please report this value and the behaviour of the official client.\n"));
-			prompt_flags = PROMPT_PASSWORD | PROMPT_USERNAME;
-			break;
-		}
+			case 1: /* Prompt for both username and password. */
+				prompt_flags = PROMPT_PASSWORD | PROMPT_USERNAME;
+				break;
+			case 3: /* Prompt for password.*/
+			case 15:
+				prompt_flags = PROMPT_PASSWORD;
+				break;
+			case 5: /* Prompt for username.*/
+				prompt_flags = PROMPT_USERNAME;
+				break;
+			default:
+				/* It does no harm to submit both, as anything unwanted is ignored. */
+				vpn_progress(vpninfo, PRG_ERR,
+				             _("Unknown D73 prompt value 0x%x. Will prompt for both username and password.\n"),
+				             val);
+				vpn_progress(vpninfo, PRG_ERR,
+				             _("Please report this value and the behaviour of the official client.\n"));
+				prompt_flags = PROMPT_PASSWORD | PROMPT_USERNAME;
+				break;
+			}
 		} else if (avp_vendor == VENDOR_JUNIPER2 && avp_code == 0xd7b) {
 			free(signin_prompt);
 			signin_prompt = strndup(avp_p, avp_len);
@@ -1764,14 +1712,30 @@ static int pulse_authenticate(struct openconnect_info *vpninfo, int connecting)
 			realms_found++;
 		} else if (avp_vendor == VENDOR_JUNIPER2 && avp_code == 0xd4f) {
 			realm_entry++;
+		} else if (avp_vendor == VENDOR_JUNIPER2 && avp_code == 0xd5c) {
+			if (avp_len != 4)
+				goto auth_unknown;
+			uint32_t val = load_be32(avp_p);
+
+			if (val)
+				vpninfo->auth_expiration = time(NULL) + val;
+		} else if (avp_vendor == VENDOR_JUNIPER2 && avp_code == 0xd75) {
+			if (avp_len != 4)
+				goto auth_unknown;
+			uint32_t val = load_be32(avp_p);
+
+			if (val)
+				vpninfo->idle_timeout = val;
 		} else if (avp_vendor == VENDOR_JUNIPER2 && avp_code == 0xd53) {
 			free(vpninfo->cookie);
 			vpninfo->cookie = strndup(avp_p, avp_len);
 			cookie_found = 1;
+			/* DSID cookie may be needed for fallback to oNCP/Juniper logout */
+			http_add_cookie(vpninfo, "DSID", vpninfo->cookie, 1 /* replace */);
 		} else if (!avp_vendor && avp_code == AVP_CODE_EAP_MESSAGE) {
 			char *avp_c = avp_p;
 
-			/* EAP within AVP within EAP within IF-T/TLS. Chewck EAP header. */
+			/* EAP within AVP within EAP within IF-T/TLS. Check EAP header. */
 			if (avp_len < 5 || avp_c[0] != EAP_REQUEST ||
 			    load_be16(avp_c + 2) != avp_len)
 				goto auth_unknown;
@@ -2229,6 +2193,40 @@ int pulse_obtain_cookie(struct openconnect_info *vpninfo)
 	return pulse_authenticate(vpninfo, 0);
 }
 
+/* Handler for config attributes, see handle_main_config_packet */
+static int handle_attr_elements(struct openconnect_info *vpninfo,
+				unsigned char *bytes, int len,
+				struct oc_vpn_option **new_opts,
+				struct oc_ip_info *new_ip_info) {
+	unsigned char *p = bytes;
+	int l = len;
+
+	/* No idea what this is */
+	if (l < 8 || load_be32(p + 4) != 0x03000000)
+		return -EINVAL;
+	p += 8;
+	l -= 8;
+
+	while (l) {
+		if (l < 4)
+			return -EINVAL;
+
+		uint16_t type = load_be16(p);
+		uint16_t attrlen = load_be16(p+2);
+
+		if (attrlen + 4 > l)
+			return -EINVAL;
+
+		p += 4;
+		l -= 4;
+		process_attr(vpninfo, new_opts, new_ip_info, type, p, attrlen);
+		p += attrlen;
+		l -= attrlen;
+	}
+
+	return 0;
+}
+
 /* Example config packet:
    < 0000: 00 00 0a 4c 00 00 00 01  00 00 01 80 00 00 01 fb  |...L............|
    < 0010: 00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  |................|
@@ -2290,26 +2288,57 @@ static int handle_main_config_packet(struct openconnect_info *vpninfo,
 	int routes_len = 0;
 	int l;
 	unsigned char *p;
+	int offset = 0x2c;
+	struct oc_vpn_option *new_opts = NULL;
+	struct oc_ip_info new_ip_info = {};
 
-	/* First part of header, similar to ESP, has already been checked */
-	if (len < 0x31 ||
-	    /* Start of routing information */
-	    load_be16(bytes + 0x2c) != 0x2e00 ||
-	    /* Routing length at 0x2e makes sense */
-	    (routes_len = load_be16(bytes + 0x2e)) != ((int)bytes[0x30] * 0x10 + 8) ||
-	    /* Make sure the next length field (at 0xa4 in the above example) is present */
-	    len < 0x2c + routes_len + 4||
-	    /* Another length field, must match to end of packet */
-	    load_be32(bytes + 0x2c + routes_len) + routes_len + 0x2c != len) {
-	bad_config:
+	if (len < 0x31) {
+		bad_config:
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Unexpected Pulse config packet:\n"));
 		dump_buf_hex(vpninfo, PRG_ERR, '<', (void *)bytes, len);
+		free_optlist(new_opts);
+		free_split_routes(&new_ip_info);
 		return -EINVAL;
 	}
-	p = bytes + 0x34;
-	routes_len -= 8; /* The header including length and number of routes */
 
+	/* On Pulse 9.1R14, we see packet type 0x2e20f0000, whereas earlier
+	 * versions had 0x2c20f0000.
+	 * With the newer packet type, we seem to have a leading set of
+	 * attribute elements. Example:
+	 *     2c 00           (fixed)
+	 *     00 0d           (length 13)
+	 *     03 00 00 00     (fixed)
+	 *     40 25 00 01 01  (unknown attr 0x4025, length 1, value 0x01)
+	*/
+	if (bytes[0x20] == 0x2e) {
+		/* Length of attributes section */
+		int attr_len = load_be16(bytes + 0x2e);
+
+		/* Start of attributes */
+		if (load_be16(bytes + 0x2c) != 0x2c00 ||
+		    len < 0x2c + attr_len + 4 ||
+		    /* Process the attributes */
+		    handle_attr_elements(vpninfo, bytes + 0x2c, attr_len,
+					 &new_opts, &new_ip_info) < 0) {
+			goto bad_config;
+		}
+		offset += attr_len;
+	}
+
+	/* First part of header, similar to ESP, has already been checked */
+	if (len < offset + 5 ||
+	    /* Start of routing information */
+	    load_be16(bytes + offset) != 0x2e00 ||
+	    /* Routing length at offset+2 makes sense */
+	    (routes_len = load_be16(bytes + offset + 2)) != ((int)bytes[offset + 4] * 0x10 + 8) ||
+	    /* Make sure the next length field (at 0xa4 in the above example) is present */
+	    len < offset + routes_len + 4 ||
+	    /* Another length field, must match to end of packet */
+	    load_be32(bytes + offset + routes_len) + routes_len + offset != len) {
+	}
+	p = bytes + offset + 8;
+	routes_len -= 8; /* The header including length and number of routes */
 	/* We know it's a multiple of 0x10 now. We checked. */
 	while (routes_len) {
 		char buf[80];
@@ -2339,10 +2368,10 @@ static int handle_main_config_packet(struct openconnect_info *vpninfo,
 			vpn_progress(vpninfo, PRG_DEBUG, _("Received split include route %s\n"), buf);
 			inc = malloc(sizeof(*inc));
 			if (inc) {
-				inc->route = add_option(vpninfo, "split-include", buf, -1);
+				inc->route = add_option_dup(&new_opts, "split-include", buf, -1);
 				if (inc->route) {
-					inc->next = vpninfo->ip_info.split_includes;
-					vpninfo->ip_info.split_includes = inc;
+					inc->next = new_ip_info.split_includes;
+					new_ip_info.split_includes = inc;
 				} else
 					free(inc);
 			}
@@ -2352,10 +2381,10 @@ static int handle_main_config_packet(struct openconnect_info *vpninfo,
 			vpn_progress(vpninfo, PRG_DEBUG, _("Received split exclude route %s\n"), buf);
 			exc = malloc(sizeof(*exc));
 			if (exc) {
-				exc->route = add_option(vpninfo, "split-exclude", buf, -1);
+				exc->route = add_option_dup(&new_opts, "split-exclude", buf, -1);
 				if (exc->route) {
-					exc->next = vpninfo->ip_info.split_excludes;
-					vpninfo->ip_info.split_excludes = exc;
+					exc->next = new_ip_info.split_excludes;
+					new_ip_info.split_excludes = exc;
 				} else
 					free(exc);
 			}
@@ -2372,28 +2401,15 @@ static int handle_main_config_packet(struct openconnect_info *vpninfo,
 	/* p now points at the length field of the final elements, which
 	   was already checked. */
 	l = load_be32(p);
-	/* No idea what this is */
-	if (l < 8 || load_be32(p + 4) != 0x03000000)
+	if (handle_attr_elements(vpninfo, p, l, &new_opts, &new_ip_info) < 0)
 		goto bad_config;
-	p += 8;
-	l -= 8;
 
-	while (l) {
-		uint16_t type = load_be16(p);
-		uint16_t attrlen = load_be16(p+2);
-
-		if (attrlen + 4 > l)
-			goto bad_config;
-
-		p += 4;
-		l -= 4;
-		process_attr(vpninfo, type, p, attrlen);
-		p += attrlen;
-		l -= attrlen;
-		if (l && l < 4)
-			goto bad_config;
+	int ret = install_vpn_opts(vpninfo, new_opts, &new_ip_info);
+	if (ret) {
+		free_optlist(new_opts);
+		free_split_routes(&new_ip_info);
 	}
-	return 0;
+	return ret;
 }
 
 /* Example ESP config packet:
@@ -2555,6 +2571,7 @@ int pulse_connect(struct openconnect_info *vpninfo)
 
 		switch(load_be32(bytes + 0x20)) {
 		case 0x2c20f000:
+		case 0x2e20f000: /* Variant seen on Pulse 9.1R14 */
 			ret = handle_main_config_packet(vpninfo, bytes, ret);
 			if (ret)
 				return ret;
@@ -2591,16 +2608,20 @@ int pulse_connect(struct openconnect_info *vpninfo)
 
 	if (!vpninfo->ip_info.mtu ||
 	    (!vpninfo->ip_info.addr && !vpninfo->ip_info.addr6)) {
-		vpn_progress(vpninfo, PRG_ERR, "Insufficient configuration found\n");
+		vpn_progress(vpninfo, PRG_ERR, _("Insufficient configuration found\n"));
 		return -EINVAL;
 	}
 
+	/* This should never happen, but be defensive and shut Coverity up */
+	if (vpninfo->ssl_fd == -1)
+		return -EIO;
+
 	ret = 0;
 	monitor_fd_new(vpninfo, ssl);
 	monitor_read_fd(vpninfo, ssl);
 	monitor_except_fd(vpninfo, ssl);
 
-	free(vpninfo->cstp_pkt);
+	free_pkt(vpninfo, vpninfo->cstp_pkt);
 	vpninfo->cstp_pkt = NULL;
 
 	return ret;
@@ -2630,7 +2651,7 @@ int pulse_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 		int len, payload_len;
 
 		if (!pkt) {
-			pkt = vpninfo->cstp_pkt = malloc(sizeof(struct pkt) + receive_mtu);
+			pkt = vpninfo->cstp_pkt = alloc_pkt(vpninfo, receive_mtu);
 			if (!pkt) {
 				vpn_progress(vpninfo, PRG_ERR, _("Allocation failed\n"));
 				break;
@@ -2638,7 +2659,7 @@ int pulse_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 		}
 
 		/* Receive packet header, if there's anything there... */
-		len = ssl_nonblock_read(vpninfo, &pkt->pulse.vendor, 16);
+		len = ssl_nonblock_read(vpninfo, 0, &pkt->pulse.vendor, 16);
 		if (!len)
 			break;
 		if (len < 0)
@@ -2659,7 +2680,7 @@ int pulse_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 		} else
 			len = load_be32(&pkt->pulse.len) - 0x10;
 
-		payload_len = ssl_nonblock_read(vpninfo, &pkt->data, len);
+		payload_len = ssl_nonblock_read(vpninfo, 0, &pkt->data, len);
 		if (payload_len != load_be32(&pkt->pulse.len) - 0x10) {
 			if (payload_len < 0)
 				len = 0x10;
@@ -2705,12 +2726,35 @@ int pulse_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 			}
 			vpninfo->cstp_pkt = NULL;
 			pkt->len = load_be32(&pkt->pulse.len) - 16;
-			queue_packet(&vpninfo->oncp_control_queue, pkt);
+			queue_packet(&vpninfo->tcp_control_queue, pkt);
 
 			print_esp_keys(vpninfo, _("new incoming"), &vpninfo->esp_in[vpninfo->current_esp_in]);
 			print_esp_keys(vpninfo, _("new outgoing"), &vpninfo->esp_out);
 			continue;
 
+		case 0x93: {
+			/* Expected contents are "errorType=%d errorString=%s\n". Known values:
+			 * 6: "agentd error" (another session started and kicked this one off)
+			 * 7: "session has been terminated" (by client)
+			 * 8: "session timed out" (idle timeout)
+			 */
+			if (payload_len < 12 || strncmp("errorType=", (const char *)pkt->data, 10))
+				goto unknown_pkt;
+			pkt->data[payload_len - 1] = '\0'; /* overwrite final '\n' */
+
+			char *endp;
+			unsigned long reason = strtol((const char *)pkt->data + 10, &endp, 10);
+			if (strncmp(" errorString=", endp, 13))
+				goto unknown_pkt;
+
+			urldecode_inplace(endp+1);
+
+			vpn_progress(vpninfo, PRG_ERR, _("Pulse fatal error (reason: %ld): %s\n"),
+				     reason, endp+13);
+			vpninfo->quit_reason = strdup(endp+13);
+			return -EPIPE;
+		}
+
 		case 0x96:
 			/* It sends the licence information once the connection is set up. For
 			 * now, abuse this to deal with the race condition in ESP setup — it looks
@@ -2744,7 +2788,7 @@ int pulse_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 			     (void *)&vpninfo->current_ssl_pkt->pulse.vendor,
 			     vpninfo->current_ssl_pkt->len + 16);
 
-		ret = ssl_nonblock_write(vpninfo,
+		ret = ssl_nonblock_write(vpninfo, 0,
 					 &vpninfo->current_ssl_pkt->pulse.vendor,
 					 vpninfo->current_ssl_pkt->len + 16);
 		if (ret < 0) {
@@ -2778,7 +2822,7 @@ int pulse_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 				return work_done;
 			default:
 				/* This should never happen */
-				;
+				break;
 			}
 #else
 			return work_done;
@@ -2794,10 +2838,10 @@ int pulse_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 		}
 		/* Don't free the 'special' packets */
 		if (vpninfo->current_ssl_pkt == vpninfo->deflate_pkt) {
-			free(vpninfo->pending_deflated_pkt);
+			free_pkt(vpninfo, vpninfo->pending_deflated_pkt);
 			vpninfo->pending_deflated_pkt = NULL;
 		} else
-			free(vpninfo->current_ssl_pkt);
+			free_pkt(vpninfo, vpninfo->current_ssl_pkt);
 
 		vpninfo->current_ssl_pkt = NULL;
 	}
@@ -2861,7 +2905,7 @@ int pulse_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 	case KA_KEEPALIVE:
 		/* No need to send an explicit keepalive
 		   if we have real data to send */
-		if (vpninfo->dtls_state != DTLS_CONNECTED &&
+		if (vpninfo->dtls_state != DTLS_ESTABLISHED &&
 		    vpninfo->outgoing_queue.head)
 			break;
 
@@ -2874,15 +2918,15 @@ int pulse_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 		;
 	}
 #endif
-	if (vpninfo->dtls_state == DTLS_CONNECTING) {
+	if (vpninfo->dtls_state == DTLS_CONNECTED) {
 		/* We don't currently do anything to make the server start sending
 		 * data packets in ESP instead of over IF-T/TLS. Just go straight
 		 * to CONNECTED mode. */
-		vpninfo->dtls_state = DTLS_CONNECTED;
+		vpninfo->dtls_state = DTLS_ESTABLISHED;
 		work_done = 1;
 	}
 
-	vpninfo->current_ssl_pkt = dequeue_packet(&vpninfo->oncp_control_queue);
+	vpninfo->current_ssl_pkt = dequeue_packet(&vpninfo->tcp_control_queue);
 	if (vpninfo->current_ssl_pkt) {
 		/* Anything on the control queue will have the rest of its
 		   header filled in already. */
@@ -2891,7 +2935,7 @@ int pulse_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 	}
 
 	/* Service outgoing packet queue, if no DTLS */
-	while (vpninfo->dtls_state != DTLS_CONNECTED &&
+	while (vpninfo->dtls_state != DTLS_ESTABLISHED &&
 	       (vpninfo->current_ssl_pkt = dequeue_packet(&vpninfo->outgoing_queue))) {
 		struct pkt *this = vpninfo->current_ssl_pkt;
 
@@ -2914,14 +2958,17 @@ int pulse_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 
 int pulse_bye(struct openconnect_info *vpninfo, const char *reason)
 {
+	int ret = -1;
 	if (vpninfo->ssl_fd != -1) {
 		struct oc_text_buf *buf = buf_alloc();
 		buf_append_ift_hdr(buf, VENDOR_JUNIPER, 0x89);
 		if (!buf_error(buf))
-			send_ift_packet(vpninfo, buf);
+			ret = send_ift_packet(vpninfo, buf);
 		buf_free(buf);
-
 		openconnect_close_https(vpninfo, 0);
 	}
-	return 0;
+	/* Try Juniper logout if tunnel was already closed */
+	if (ret < 0)
+		ret = oncp_bye(vpninfo, reason);
+	return ret;
 }
