@@ -17,21 +17,25 @@
 
 #include <config.h>
 
+#include "openconnect-internal.h"
+
+#include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <string.h>
 #include <fcntl.h>
-#include <unistd.h>
-#ifndef _WIN32
+#ifdef _WIN32
+#include <process.h>
+#define getpid _getpid
+#else
 #include <sys/wait.h>
 #endif
+
 #include <errno.h>
 #include <ctype.h>
+#include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
 
-#include "openconnect-internal.h"
-
 int script_setenv(struct openconnect_info *vpninfo,
 		  const char *opt, const char *val, int trunc, int append)
 {
@@ -68,15 +72,40 @@ int script_setenv_int(struct openconnect_info *vpninfo, const char *opt, int val
 	return script_setenv(vpninfo, opt, buf, 0, 0);
 }
 
+static inline int cls(uint32_t word)
+{
+#if defined(HAVE_BUILTIN_CLZ) && UINT_MAX == UINT32_MAX
+	word = ~word;
+	if (!word)
+		return 32;
+	return __builtin_clz(word);
+#else
+	int masklen = 0;
+	while (word & 0x80000000) {
+		word <<= 1;
+		masklen++;
+	}
+	return masklen;
+#endif
+}
+
 static int netmasklen(struct in_addr addr)
 {
-	int masklen;
+	return cls(ntohl(addr.s_addr));
+}
 
-	for (masklen = 0; masklen < 32; masklen++) {
-		if (ntohl(addr.s_addr) >= (0xffffffff << masklen))
-			break;
+static int netmasklen6(struct in6_addr *addr)
+{
+	int masklen;
+	uint32_t *p = (uint32_t *)(addr->s6_addr);
+
+	for (masklen = 0; masklen < 128; p++, masklen += 32) {
+		uint32_t v = ntohl(*p);
+		if (~v == 0)
+			continue;
+		return masklen + cls(v);
 	}
-	return 32 - masklen;
+	return 128;
 }
 
 static uint32_t netmaskbits(int masklen)
@@ -93,7 +122,7 @@ static int process_split_xxclude(struct openconnect_info *vpninfo,
 {
 	struct in_addr net_addr, mask_addr;
 	const char *in_ex = include ? "IN" : "EX";
-	char envname[80], uptoslash[20];
+	char envname[80], uptoslash[20], abuf[INET_ADDRSTRLEN];
 	const char *slash;
 	char *endp;
 	int masklen;
@@ -106,10 +135,30 @@ static int process_split_xxclude(struct openconnect_info *vpninfo,
 			 *v6_incs);
 		script_setenv(vpninfo, envname, route, slash ? slash - route : 0, 0);
 
+		/* Accept IPv6 netmask in several forms */
 		snprintf(envname, 79, "CISCO_IPV6_SPLIT_%sC_%d_MASKLEN", in_ex,
 			 *v6_incs);
-		script_setenv(vpninfo, envname, slash ? slash + 1 : "128", 0, 0);
-
+		if (!slash) {
+			/* no mask (same as /128) */
+			script_setenv_int(vpninfo, envname, 128);
+		} else if ((masklen = strtol(slash+1, &endp, 10))<=128 && (*endp=='\0' || isspace(*endp))) {
+			/* mask is /N */
+			script_setenv_int(vpninfo, envname, masklen);
+		} else {
+			/* mask is /dead:beef:: */
+			struct in6_addr a;
+			if (inet_pton(AF_INET6, slash+1, &a) <= 0)
+				goto bad;
+			masklen = netmasklen6(&a);
+			/* something invalid like /ffff::1 */
+			for (int ii = (masklen >> 3); ii < 16; ii++) {
+				if (ii == (masklen >> 3) && (~a.s6_addr[ii] & 0xff) != (0xff >> (masklen & 0x07)))
+					goto bad;
+				else if (a.s6_addr[ii] != 0)
+					goto bad;
+			}
+			script_setenv_int(vpninfo, envname, masklen);
+		}
 		(*v6_incs)++;
 		return 0;
 	}
@@ -156,21 +205,22 @@ static int process_split_xxclude(struct openconnect_info *vpninfo,
 	/* Fix incorrectly-set host bits */
 	if (net_addr.s_addr & ~mask_addr.s_addr) {
 		net_addr.s_addr &= mask_addr.s_addr;
+		inet_ntop(AF_INET, &net_addr, abuf, sizeof(abuf));
 		if (include)
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("WARNING: Split include \"%s\" has host bits set, replacing with \"%s/%d\".\n"),
-				     route, inet_ntoa(net_addr), masklen);
+				     route, abuf, masklen);
 		else
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("WARNING: Split exclude \"%s\" has host bits set, replacing with \"%s/%d\".\n"),
-				     route, inet_ntoa(net_addr), masklen);
+				     route, abuf, masklen);
 	}
 
 	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_ADDR", in_ex, *v4_incs);
-	script_setenv(vpninfo, envname, inet_ntoa(net_addr), 0, 0);
+	script_setenv(vpninfo, envname, inet_ntop(AF_INET, &net_addr, abuf, sizeof(abuf)), 0, 0);
 
 	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_MASK", in_ex, *v4_incs);
-	script_setenv(vpninfo, envname, inet_ntoa(mask_addr), 0, 0);
+	script_setenv(vpninfo, envname, inet_ntop(AF_INET, &mask_addr, abuf, sizeof(abuf)), 0, 0);
 
 	snprintf(envname, 79, "CISCO_SPLIT_%sC_%d_MASKLEN", in_ex, *v4_incs);
 	script_setenv_int(vpninfo, envname, masklen);
@@ -255,6 +305,8 @@ void prepare_script_env(struct openconnect_info *vpninfo)
 	script_setenv(vpninfo, "CISCO_SPLIT_EXC", NULL, 0, 0);
 
 	script_setenv_int(vpninfo, "INTERNAL_IP4_MTU", vpninfo->ip_info.mtu);
+	script_setenv_int(vpninfo, "VPNPID", (int)getpid());
+	script_setenv_int(vpninfo, "LOG_LEVEL", vpninfo->verbose);
 
 	if (vpninfo->idle_timeout)
 		script_setenv_int(vpninfo, "IDLE_TIMEOUT", vpninfo->idle_timeout);
@@ -277,13 +329,13 @@ void prepare_script_env(struct openconnect_info *vpninfo)
 					     _("Ignoring legacy network because netmask \"%s\" is invalid.\n"),
 					     vpninfo->ip_info.netmask);
 			else {
-				char *netaddr;
+				char netaddr[INET_ADDRSTRLEN];
 				int masklen = netmasklen(mask);
 
 				if (netmaskbits(masklen) != mask.s_addr)
 					goto bad_netmask;
 				addr.s_addr &= mask.s_addr;
-				netaddr = inet_ntoa(addr);
+				inet_ntop(AF_INET, &addr, netaddr, sizeof(netaddr));
 
 				script_setenv(vpninfo, "INTERNAL_IP4_NETADDR", netaddr, 0, 0);
 				script_setenv(vpninfo, "INTERNAL_IP4_NETMASK", vpninfo->ip_info.netmask, 0, 0);
@@ -291,15 +343,18 @@ void prepare_script_env(struct openconnect_info *vpninfo)
 			}
 		}
 	}
-	if (vpninfo->ip_info.addr6) {
+
+	if (vpninfo->ip_info.addr6)
 		script_setenv(vpninfo, "INTERNAL_IP6_ADDRESS", vpninfo->ip_info.addr6, 0, 0);
+	if (vpninfo->ip_info.netmask6)
 		script_setenv(vpninfo, "INTERNAL_IP6_NETMASK", vpninfo->ip_info.netmask6, 0, 0);
-	} else if (vpninfo->ip_info.netmask6) {
-               char *slash = strchr(vpninfo->ip_info.netmask6, '/');
-               script_setenv(vpninfo, "INTERNAL_IP6_NETMASK", vpninfo->ip_info.netmask6, 0, 0);
-               if (slash)
-                       script_setenv(vpninfo, "INTERNAL_IP6_ADDRESS", vpninfo->ip_info.netmask6,
-				     slash - vpninfo->ip_info.netmask6, 0);
+	/* The 'netmask6' is actually the address *and* netmask. From which we
+	 * obtain just the address on its own, if we don't have it separately */
+	if (vpninfo->ip_info.netmask6 && !vpninfo->ip_info.addr6) {
+		char *slash = strchr(vpninfo->ip_info.netmask6, '/');
+		if (slash)
+			script_setenv(vpninfo, "INTERNAL_IP6_ADDRESS", vpninfo->ip_info.netmask6,
+				      slash - vpninfo->ip_info.netmask6, 0);
 	}
 
 	if (vpninfo->ip_info.dns[0])
@@ -391,27 +446,27 @@ void prepare_script_env(struct openconnect_info *vpninfo)
 	setenv_cstp_opts(vpninfo);
 }
 
-void free_split_routes(struct openconnect_info *vpninfo)
+void free_split_routes(struct oc_ip_info *ip_info)
 {
 	struct oc_split_include *inc;
 
-	for (inc = vpninfo->ip_info.split_includes; inc; ) {
+	for (inc = ip_info->split_includes; inc; ) {
 		struct oc_split_include *next = inc->next;
 		free(inc);
 		inc = next;
 	}
-	for (inc = vpninfo->ip_info.split_excludes; inc; ) {
+	for (inc = ip_info->split_excludes; inc; ) {
 		struct oc_split_include *next = inc->next;
 		free(inc);
 		inc = next;
 	}
-	for (inc = vpninfo->ip_info.split_dns; inc; ) {
+	for (inc = ip_info->split_dns; inc; ) {
 		struct oc_split_include *next = inc->next;
 		free(inc);
 		inc = next;
 	}
-	vpninfo->ip_info.split_dns = vpninfo->ip_info.split_includes =
-		vpninfo->ip_info.split_excludes = NULL;
+	ip_info->split_dns = ip_info->split_includes =
+		ip_info->split_excludes = NULL;
 }
 
 
@@ -497,7 +552,7 @@ int script_config_tun(struct openconnect_info *vpninfo, const char *reason)
 	char *cmd;
 	PROCESS_INFORMATION pi;
 	STARTUPINFOW si;
-	DWORD cpflags;
+	DWORD cpflags, exit_status;
 
 	if (!vpninfo->vpnc_script || vpninfo->script_tun)
 		return 0;
@@ -535,11 +590,28 @@ int script_config_tun(struct openconnect_info *vpninfo, const char *reason)
 	if (CreateProcessW(NULL, script_w, NULL, NULL, FALSE, cpflags,
 			   script_env, NULL, &si, &pi)) {
 		ret = WaitForSingleObject(pi.hProcess,10000);
+		if (!GetExitCodeProcess(pi.hProcess, &exit_status)) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to get script exit status: %s\n"),
+				     openconnect__win32_strerror(GetLastError()));
+			ret = -EIO;
+		} else if (exit_status > 0 && exit_status != STILL_ACTIVE) {
+			/* STILL_ACTIVE == 259. That means that a perfectly normal positive integer return value overlaps with
+			 * an exceptional condition. Don't blame me. I didn't design this.
+			 * https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-getexitcodeprocess#remarks
+			 */
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Script '%s' returned error %ld\n"),
+				     vpninfo->vpnc_script, exit_status);
+			ret = -EIO;
+		}
 		CloseHandle(pi.hThread);
 		CloseHandle(pi.hProcess);
-		if (ret == WAIT_TIMEOUT)
+		if (ret == WAIT_TIMEOUT || exit_status == STILL_ACTIVE) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Script did not complete within 10 seconds.\n"));
 			ret = -ETIMEDOUT;
-		else
+		} else if (ret != -EIO)
 			ret = 0;
 	} else {
 		ret = -EIO;
