@@ -18,15 +18,15 @@
 
 #include <config.h>
 
+#include "openconnect-internal.h"
+
+#include <stoken.h>
+
 #include <ctype.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include <stoken.h>
-
-#include "openconnect-internal.h"
-
 #ifndef STOKEN_CHECK_VER
 #define STOKEN_CHECK_VER(x,y) 0
 #endif
@@ -121,10 +121,8 @@ static int decrypt_stoken(struct openconnect_info *vpninfo)
 	while (1) {
 		nuke_opt_values(opts);
 
-		if (!opts[0].type) {
-			/* don't bug the user if there's nothing to enter */
-			ret = 0;
-		} else {
+		/* don't bug the user if there's nothing to enter */
+		if (opts[0].type) {
 			int some_empty = 0, all_empty = 1;
 
 			/* < 0 for error; 1 if cancelled */
@@ -335,4 +333,3 @@ int do_gen_stoken_code(struct openconnect_info *vpninfo,
 		return -ENOMEM;
 	return 0;
 }
-
