@@ -20,7 +20,7 @@
  * AnyConnect VPN protocol.
  *
  * This is designed to exercise the code paths in
- * http://git.infradead.org/users/dwmw2/openconnect.git/blob/HEAD:/dtls.c
+ * https://git.infradead.org/users/dwmw2/openconnect.git/blob/HEAD:/dtls.c
  * which have frequently been affected by regressions in DTLS1_BAD_VER
  * support.
  *
@@ -37,6 +37,8 @@
  */
 #include <string.h>
 
+#define OPENSSL_SUPPRESS_DEPRECATED
+
 #include <openssl/bio.h>
 #include <openssl/crypto.h>
 #include <openssl/evp.h>
@@ -49,7 +51,7 @@
 #define DTLS1_BAD_VER 0x100
 #endif
 
-/* PACKET functions lifted from OpenSSL 1.1's ssl/packet_locl.h. Permisson
+/* PACKET functions lifted from OpenSSL 1.1's ssl/packet_locl.h. Permission
  * requested in https://github.com/openssl/openssl/pull/1296 for reuse here
  * as an OpenConnect test case. */
 /*
@@ -268,7 +270,7 @@ static inline int PACKET_get_length_prefixed_1(PACKET *pkt,
     return 1;
 }
 
-#define OSSL_NELEM(x)    (sizeof(x)/sizeof(x[0]))
+#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
 
 /* For DTLS1_BAD_VER packets the MAC doesn't include the handshake header */
 #define MAC_OFFSET (DTLS1_RT_HEADER_LENGTH + DTLS1_HM_HEADER_LENGTH)
@@ -788,7 +790,8 @@ int main(int argc, char *argv[])
     ctx = SSL_CTX_new(DTLS_client_method());
     if (ctx == NULL ||
 	!SSL_CTX_set_min_proto_version(ctx, DTLS1_BAD_VER) ||
-	!SSL_CTX_set_max_proto_version(ctx, DTLS1_BAD_VER)) {
+	!SSL_CTX_set_max_proto_version(ctx, DTLS1_BAD_VER) ||
+	!SSL_CTX_set_options(ctx, SSL_OP_LEGACY_SERVER_CONNECT)) {
         printf("Failed to allocate SSL_CTX\n");
         goto end_md;
     }
@@ -867,7 +870,7 @@ int main(int argc, char *argv[])
        specific but useful anyway for the general case. It's been broken
        before, and in fact was broken even for a basic 0, 2, 1 test case
        when this test was first added.... */
-    for (i = 0; i < (int)OSSL_NELEM(tests); i++) {
+    for (i = 0; i < (int)ARRAY_SIZE(tests); i++) {
         unsigned long recv_buf[2];
 
         if (send_record(rbio, SSL3_RT_APPLICATION_DATA, tests[i].seq,
