@@ -32,10 +32,15 @@
 # Environment variables that may need customization (excerpted from
 # https://github.com/russdill/juniper-vpn-py/blame/master/README.host_checker):
 #
-# TNCC_DEVICE_ID: May need to be overriden to match a known value from a computer
-#   running the official Windows client software (obtained from the registry key
+# TNCC_DEVICE_ID: May need to be overridden to match a known value from a computer
+#   running the official client software (on Windows, obtained from the registry key
 #   \HKEY_CURRENT_USER\Software\Juniper Networks\Device Id)
 #
+# TNCC_USER_AGENT: May need to be overridden to match a known value from a computer
+#   running the official Windows client software. For historical reasons, the default
+#   value is 'Neoteris NC Http'; the value 'DSClient; PulseLinux' is known to be sent
+#   by the official Pulse Linux client.
+#
 # TNCC_FUNK: Set TNCC_FUNK=1 to force the use of client machine identification
 #   (known as "funk" to Juniper). This identification will include host platform,
 #   a list of network hardware/MAC addresses, and client certificates requested
@@ -55,7 +60,6 @@
 import sys
 import os
 import logging
-from io import StringIO
 from http.cookiejar import Cookie, CookieJar
 import struct
 import ssl
@@ -71,7 +75,8 @@ import xml.etree.ElementTree
 
 import mechanize
 try:
-    import asn1crypto.pem, asn1crypto.x509
+    import asn1crypto.pem
+    import asn1crypto.x509
 except ImportError:
     asn1crypto = None
 try:
@@ -99,11 +104,13 @@ def decode_0013(buf, indent):
         ret[cmd].append(out)
     return ret
 
+
 # 0012 - u32
 def decode_0012(buf, indent):
     logging.debug('%scmd 0012 (u32) %d bytes', indent, len(buf))
     return struct.unpack(">I", buf)
 
+
 # 0016 - zlib compressed message
 def decode_0016(buf, indent):
     logging.debug('%scmd 0016 (compressed message) %d bytes', indent, len(buf))
@@ -116,6 +123,7 @@ def decode_0016(buf, indent):
         ret[cmd].append(out)
     return ret
 
+
 # 0ce4 - encapsulation
 def decode_0ce4(buf, indent):
     logging.debug('%scmd 0ce4 (encapsulation) %d bytes', indent, len(buf))
@@ -126,6 +134,7 @@ def decode_0ce4(buf, indent):
         ret[cmd].append(out)
     return ret
 
+
 # 0ce5 - string without hex prefixer
 def decode_0ce5(buf, indent):
     s = struct.unpack(str(len(buf)) + "s", buf)[0]
@@ -134,19 +143,21 @@ def decode_0ce5(buf, indent):
     logging.debug('%s', s)
     return s
 
+
 # 0ce7 - string with hex prefixer
 def decode_0ce7(buf, indent):
     id, s = struct.unpack(">I" + str(len(buf) - 4) + "s", buf)
     logging.debug('%scmd 0ce7 (id %08x string) %d bytes', indent, id, len(buf))
 
     if s.startswith(b'COMPRESSED:'):
-        typ, length, data = s.split(':', 2)
+        typ, length, data = s.split(b':', 2)
         s = zlib.decompress(data)
 
     s = s.rstrip(b'\0')
     logging.debug('%s', s)
     return (id, s)
 
+
 # 0cf0 - encapsulation
 def decode_0cf0(buf, indent):
     logging.debug('%scmd 0cf0 (encapsulation) %d bytes', indent, len(buf))
@@ -155,6 +166,7 @@ def decode_0cf0(buf, indent):
     ret[cmd] = out
     return ret
 
+
 # 0cf1 - string without hex prefixer
 def decode_0cf1(buf, indent):
     s = struct.unpack(str(len(buf)) + "s", buf)[0]
@@ -163,12 +175,14 @@ def decode_0cf1(buf, indent):
     logging.debug('%s', s)
     return s
 
+
 # 0cf3 - u32
 def decode_0cf3(buf, indent):
     ret = struct.unpack(">I", buf)
     logging.debug('%scmd 0cf3 (u32) %d bytes - %d', indent, len(buf), ret[0])
     return ret
 
+
 def decode_packet(buf, indent=""):
     cmd, _1, _2, length, _3 = struct.unpack(">IBBHI", buf[:12])
     if length < 12:
@@ -203,6 +217,7 @@ def decode_packet(buf, indent=""):
 
     return length, cmd, data
 
+
 def encode_packet(cmd, align, buf):
     align = 4
     orig_len = len(buf)
@@ -211,42 +226,51 @@ def encode_packet(cmd, align, buf):
 
     return struct.pack(">IBBHI", cmd, 0xc0, 0x00, orig_len + 12, 0x0000583) + buf
 
+
 # 0013 - Message
 def encode_0013(buf):
     return encode_packet(0x0013, 4, buf)
 
+
 # 0012 - u32
 def encode_0012(i):
     return encode_packet(0x0012, 1, struct.pack("<I", i))
 
+
 # 0ce4 - encapsulation
 def encode_0ce4(buf):
     return encode_packet(0x0ce4, 4, buf)
 
+
 # 0ce5 - string without hex prefixer
 def encode_0ce5(s):
     return encode_packet(0x0ce5, 1, struct.pack(str(len(s)) + "s", s))
 
+
 # 0ce7 - string with hex prefixer
 def encode_0ce7(s, prefix):
     s += b'\0'
     return encode_packet(0x0ce7, 1, struct.pack(">I" + str(len(s)) + "sx",
-                                prefix, s))
+                         prefix, s))
+
 
 # 0cf0 - encapsulation
 def encode_0cf0(buf):
     return encode_packet(0x0cf0, 4, buf)
 
+
 # 0cf1 - string without hex prefixer
 def encode_0cf1(s):
     s += b'\0'
     return encode_packet(0x0ce5, 1, struct.pack(str(len(s)) + "s", s))
 
+
 # 0cf3 - u32
 def encode_0cf3(i):
     return encode_packet(0x0013, 1, struct.pack("<I", i))
 
-class x509cert(object):
+
+class x509cert:
 
     @staticmethod
     def decode_names(names):
@@ -269,16 +293,23 @@ class x509cert(object):
         self.not_after = tbs['validity']['not_after'].native.astimezone(datetime.timezone.utc).replace(tzinfo=None)
         self.subject = self.decode_names(tbs['subject'])
 
-class tncc(object):
-    def __init__(self, vpn_host, device_id=None, funk=None, platform=None, hostname=None, mac_addrs=[], certs=[], interval=None):
+
+class tncc:
+    def __init__(self, vpn_host, device_id=None, funk=None, platform=None, hostname=None, mac_addrs=None, certs=None, interval=None, user_agent=None):
         self.vpn_host = vpn_host
         self.path = '/dana-na/'
 
         self.funk = funk
         self.platform = platform
         self.hostname = hostname
-        self.mac_addrs = mac_addrs
-        self.avail_certs = certs
+        if mac_addrs is None:
+            self.mac_addrs = []
+        else:
+            self.mac_addrs = mac_addrs
+        if certs is None:
+            self.avail_certs = []
+        else:
+            self.avail_certs = certs
         self.interval = interval
 
         self.deviceid = device_id
@@ -296,7 +327,7 @@ class tncc(object):
 
         # Follows refresh 0 but not hangs on refresh > 0
         self.br.set_handle_refresh(mechanize._http.HTTPRefreshProcessor(),
-                              max_time=1)
+                                   max_time=1)
 
         # Want debugging messages?
         if debug:
@@ -304,7 +335,7 @@ class tncc(object):
             self.br.set_debug_redirects(True)
             self.br.set_debug_responses(True)
 
-        self.user_agent = 'Neoteris HC Http'
+        self.user_agent = user_agent
         self.br.addheaders = [('User-agent', self.user_agent)]
 
     def find_cookie(self, name):
@@ -315,10 +346,10 @@ class tncc(object):
 
     def set_cookie(self, name, value):
         cookie = Cookie(version=0, name=name, value=value,
-                port=None, port_specified=False, domain=self.vpn_host,
-                domain_specified=True, domain_initial_dot=False, path=self.path,
-                path_specified=True, secure=True, expires=None, discard=True,
-                comment=None, comment_url=None, rest=None, rfc2109=False)
+                        port=None, port_specified=False, domain=self.vpn_host,
+                        domain_specified=True, domain_initial_dot=False, path=self.path,
+                        path_specified=True, secure=True, expires=None, discard=True,
+                        comment=None, comment_url=None, rest=None, rfc2109=False)
         self.cj.set_cookie(cookie)
 
     def parse_response(self):
@@ -328,7 +359,7 @@ class tncc(object):
         for line in self.r.readlines():
             line = line.strip().decode()
             # Note that msg is too long and gets wrapped, handle it special
-            if last_key == 'msg' and len(line):
+            if last_key == 'msg' and line:
                 response['msg'] += line
             else:
                 key = ''
@@ -341,10 +372,12 @@ class tncc(object):
         logging.debug('Parsed response:\n\t%s', '\n\t'.join('%r: %r,' % pair for pair in response.items()))
         return response
 
-    def parse_policy_response(self, msg_data):
+    @staticmethod
+    def parse_policy_response(msg_data):
         # The decompressed data is HTMLish, decode it. The value="" of each
         # tag is the data we want.
         objs = []
+
         class ParamHTMLParser(HTMLParser.HTMLParser):
             def handle_starttag(self, tag, attrs):
                 if tag.lower() == 'param':
@@ -361,16 +394,18 @@ class tncc(object):
                                 except ValueError:
                                     pass
                             objs.append(d)
+
         p = ParamHTMLParser()
         p.feed(msg_data)
         p.close()
         return objs
 
-    def parse_funk_response(self, msg_data):
+    @staticmethod
+    def parse_funk_response(msg_data):
         e = xml.etree.ElementTree.fromstring(msg_data)
-        req_certs = dict()
+        req_certs = {}
         for cert in e.find('AttributeRequest').findall('CertData'):
-            dns = dict()
+            dns = {}
             cert_id = cert.attrib['Id']
             for attr in cert.findall('Attribute'):
                 name = attr.attrib['Name']
@@ -396,10 +431,10 @@ class tncc(object):
 
         msg += add_attr('Platform', self.platform)
         if self.hostname:
-            msg += add_attr(self.hostname, 'NETBIOSName') # Reversed
+            msg += add_attr(self.hostname, 'NETBIOSName')  # Reversed
 
         for mac in self.mac_addrs:
-            msg += add_attr(mac, 'MACAddress') # Reversed
+            msg += add_attr(mac, 'MACAddress')  # Reversed
 
         msg += "</ClientAttributes>  </FunkMessage>"
 
@@ -422,7 +457,8 @@ class tncc(object):
 
         return encode_0ce7(msg.encode(), MSG_FUNK)
 
-    def gen_policy_request(self):
+    @staticmethod
+    def gen_policy_request():
         policy_blocks = collections.OrderedDict({
             'policy_request': {
                 'message_version': '3'
@@ -443,12 +479,13 @@ class tncc(object):
 
         msg = ''
         for policy_key, policy_val in policy_blocks.items():
-            v = ''.join([ '%s=%s;' % (k, v) for k, v in policy_val.items()])
+            v = ''.join(['%s=%s;' % (k, v) for k, v in policy_val.items()])
             msg += '<parameter name="%s" value="%s">' % (policy_key, v)
 
         return encode_0ce7(msg.encode(), 0xa4c18)
 
-    def gen_policy_response(self, policy_objs):
+    @staticmethod
+    def gen_policy_response(policy_objs):
         # Make a set of policies
         policies = set()
         for entry in policy_objs:
@@ -477,11 +514,11 @@ class tncc(object):
         else:
             try:
                 self.cj.set_cookie(dspreauth)
-            except:
+            except Exception:
                 self.set_cookie('DSPREAUTH', dspreauth)
             try:
                 self.cj.set_cookie(dssignin)
-            except:
+            except Exception:
                 self.set_cookie('DSSIGNIN', dssignin)
 
         inner = self.gen_policy_request()
@@ -503,7 +540,7 @@ class tncc(object):
         if self.deviceid:
             post_attrs['deviceid'] = self.deviceid
 
-        post_data = ''.join([ '%s=%s;' % (k, v) for k, v in post_attrs.items()])
+        post_data = ''.join(['%s=%s;' % (k, v) for k, v in post_attrs.items()])
         self.r = self.br.open('https://' + self.vpn_host + self.path + 'hc/tnchcupdate.cgi', post_data)
 
         # Parse the data returned into a key/value dict
@@ -511,9 +548,9 @@ class tncc(object):
 
         if 'interval' in response:
             m = int(response['interval'])
-            logging.debug('Got interval of %d minutes' % m)
-            if self.interval is None or self.interval > m*60:
-                self.interval = m*60
+            logging.debug('Got interval of %d minutes', m)
+            if self.interval is None or self.interval > m * 60:
+                self.interval = m * 60
 
         # msg has the stuff we want, it's base64 encoded
         logging.debug('Receiving packet -')
@@ -525,7 +562,7 @@ class tncc(object):
 
         # Pull the data out of the 'value' key in the htmlish stuff returned
         policy_objs = []
-        req_certs = dict()
+        req_certs = {}
         for str_id, sub_str in sub_strings:
             if str_id == MSG_POLICY:
                 policy_objs += self.parse_policy_response(sub_str.decode())
@@ -541,7 +578,7 @@ class tncc(object):
                             logging.debug('\t%s %s', key, val)
 
         # Try to locate the required certificates
-        certs = dict()
+        certs = {}
         for cert_id, req_dns in req_certs.items():
             for cert in self.avail_certs:
                 fail = False
@@ -551,9 +588,9 @@ class tncc(object):
                             if dn_name == 'IssuerDN':
                                 assert val in cert.issuer[name]
                             else:
-                                logging.warn('Unknown DN type %s', str(dn_name))
+                                logging.warning('Unknown DN type %s', str(dn_name))
                                 raise Exception()
-                        except:
+                        except Exception:
                             fail = True
                             break
                     if fail:
@@ -562,7 +599,7 @@ class tncc(object):
                     certs[cert_id] = cert
                     break
             if cert_id not in certs:
-                logging.warn('Could not find certificate for %s', str(req_dns))
+                logging.warning('Could not find certificate for %s', str(req_dns))
 
         inner = b''
         if certs:
@@ -579,27 +616,28 @@ class tncc(object):
             'firsttime': '1'
         }
 
-        post_data = ''.join([ '%s=%s;' % (k, v) for k, v in post_attrs.items()])
+        post_data = ''.join(['%s=%s;' % (k, v) for k, v in post_attrs.items()])
         self.r = self.br.open('https://' + self.vpn_host + self.path + 'hc/tnchcupdate.cgi', post_data)
 
         # We have a new DSPREAUTH cookie
         return self.find_cookie('DSPREAUTH')
 
-class tncc_server(object):
+
+class tncc_server:
     def __init__(self, s, t):
         self.sock = s
         self.tncc = t
 
     def process_cmd(self):
-        buf = sock.recv(1024).decode('ascii')
-        if not len(buf):
+        buf = self.sock.recv(1024).decode('ascii')
+        if not buf:
             sys.exit(0)
         cmd, buf = buf.split('\n', 1)
         cmd = cmd.strip()
-        args = dict()
+        args = {}
         for n in buf.split('\n'):
             n = n.strip()
-            if len(n):
+            if n:
                 key, val = n.strip().split('=', 1)
                 args[key] = val
         if cmd == 'start':
@@ -607,16 +645,18 @@ class tncc_server(object):
             resp = ['200', '3', cookie.value]
             if self.tncc.interval is not None:
                 resp.append(str(self.tncc.interval))
-            sock.send(('\n'.join(resp) + '\n\n').encode('ascii'))
+            self.sock.send(('\n'.join(resp) + '\n\n').encode('ascii'))
         elif cmd == 'setcookie':
-            cookie = self.tncc.get_cookie(args['Cookie'],
-                                          self.tncc.find_cookie('DSSIGNIN'))
+            self.tncc.get_cookie(args['Cookie'],
+                                 self.tncc.find_cookie('DSSIGNIN'))
         else:
-            logging.warn('Unknown command %r' % cmd)
+            logging.warning('Unknown command %r', cmd)
+
 
 def fingerprint_checking_SSLSocket(_fingerprint):
     class SSLSocket(ssl.SSLSocket):
         fingerprint = _fingerprint
+
         def do_handshake(self, *args, **kw):
             res = super().do_handshake(*args, **kw)
             der_bytes = self.getpeercert(True)
@@ -625,8 +665,11 @@ def fingerprint_checking_SSLSocket(_fingerprint):
             pin_sha256 = base64.b64encode(hashlib.sha256(pubkey).digest()).decode()
             if pin_sha256 != self.fingerprint:
                 raise Exception("Server fingerprint %s does not match expected pin-sha256:%s" % (pin_sha256, self.fingerprint))
+            return res
+
     return SSLSocket
 
+
 if __name__ == "__main__":
     vpn_host = sys.argv[1]
 
@@ -636,52 +679,58 @@ if __name__ == "__main__":
 
     platform = os.environ.get('TNCC_PLATFORM', platform.system() + ' ' + platform.release())
 
+    user_agent = os.environ.get('TNCC_USER_AGENT', 'Neoteris HC Http')
+
     if 'TNCC_HWADDR' in os.environ:
         mac_addrs = [n.strip() for n in os.environ['TNCC_HWADDR'].split(',')]
     else:
         mac_addrs = []
         if netifaces is None:
-            logging.warn("No netifaces module; mac_addrs will be empty.")
+            logging.warning("No netifaces module; mac_addrs will be empty.")
         else:
             for iface in netifaces.interfaces():
                 try:
                     mac = netifaces.ifaddresses(iface)[netifaces.AF_LINK][0]['addr']
                     assert mac != '00:00:00:00:00:00'
                     mac_addrs.append(mac)
-                except:
+                except Exception:
                     pass
 
     hostname = os.environ.get('TNCC_HOSTNAME', socket.gethostname())
 
     fingerprint = os.environ.get('TNCC_SHA256')
     if not fingerprint:
-        logging.warn("TNCC_SHA256 not set, will not validate server certificate")
+        logging.warning("TNCC_SHA256 not set, will not validate server certificate")
     elif not asn1crypto:
-        logging.warn("asn1crypto module not available, will not validate server certificate")
+        logging.warning("asn1crypto module not available, will not validate server certificate")
     else:
-        # we need to monkey-patch this, because SSLContext.sslsocket_class isn't
-        # available until Python 3.7
-        # https://docs.python.org/3/library/ssl.html#ssl.SSLContext.wrap_socket
-        ssl.SSLSocket = fingerprint_checking_SSLSocket(fingerprint)
+        # For Python <3.7, we monkey-patch ssl.SSLSocket directly, because ssl.SSLContext.sslsocket_class
+        # isn't available until Python 3.7. For Python 3.7+, we set ssl.SSLContext.sslsocket_class
+        # to our modified version (which is sort of monkey-patching too).
+        # (see https://gist.github.com/dlenski/fc42156c00a615f4aa18a6d19d67e208)
+        if sys.version_info >= (3, 7):
+            ssl.SSLContext.sslsocket_class = fingerprint_checking_SSLSocket(fingerprint)
+        else:
+            ssl.SSLSocket = fingerprint_checking_SSLSocket(fingerprint)
 
     certs = []
-    if asn1crypto:
-        if 'TNCC_CERTS' in os.environ:
+    if 'TNCC_CERTS' in os.environ:
+        if asn1crypto:
             now = datetime.datetime.utcnow()
             for f in os.environ['TNCC_CERTS'].split(','):
                 cert = x509cert(f.strip())
                 if now < cert.not_before:
-                    logging.warn('WARNING: %s is not yet valid', f)
+                    logging.warning('WARNING: %s is not yet valid', f)
                 if now > cert.not_after:
-                    logging.warn('WARNING: %s is expired', f)
+                    logging.warning('WARNING: %s is expired', f)
                 certs.append(cert)
-    else:
-        raise Exception('TNCC_CERTS environment variable set, but asn1crypto module is not available')
+        else:
+            raise Exception('TNCC_CERTS environment variable set, but asn1crypto module is not available')
 
     # \HKEY_CURRENT_USER\Software\Juniper Networks\Device Id
     device_id = os.environ.get('TNCC_DEVICE_ID')
 
-    t = tncc(vpn_host, device_id, funk, platform, hostname, mac_addrs, certs, interval)
+    t = tncc(vpn_host, device_id, funk, platform, hostname, mac_addrs, certs, interval, user_agent)
     sock = socket.fromfd(0, socket.AF_UNIX, socket.SOCK_SEQPACKET)
     server = tncc_server(sock, t)
     while True:
