@@ -17,16 +17,20 @@
 
 #include <config.h>
 
+#include "openconnect-internal.h"
+
 #define WIN32_LEAN_AND_MEAN
 #include <windows.h>
 #include <winioctl.h>
+/* must precede iphlpapi.h, per https://docs.microsoft.com/en-us/windows/win32/api/iptypes/ns-iptypes-ip_adapter_addresses_lh */
+#include <winsock2.h>
+#include <ws2ipdef.h>
 #include <iphlpapi.h>
+#include <netioapi.h>
 
 #include <errno.h>
 #include <stdio.h>
 
-#include "openconnect-internal.h"
-
 /*
  * TAP-Windows support inspired by http://i3.cs.berkeley.edu/ (v0.2) with
  * permission.
@@ -55,18 +59,25 @@
 #define ADAPTERS_KEY CONTROL_KEY "Class\\" NETDEV_GUID
 #define CONNECTIONS_KEY CONTROL_KEY "Network\\" NETDEV_GUID
 
-typedef intptr_t (tap_callback)(struct openconnect_info *vpninfo, char *idx, wchar_t *name);
+#define ADAPTER_TUNTAP 0
+#define ADAPTER_WINTUN 1
 
-static intptr_t search_taps(struct openconnect_info *vpninfo, tap_callback *cb, int all)
+typedef intptr_t (tap_callback)(struct openconnect_info *vpninfo, int type, char *idx, wchar_t *name);
+
+#define SEARCH_CONTINUE	0
+#define SEARCH_DONE	1
+
+static intptr_t search_taps(struct openconnect_info *vpninfo, tap_callback *cb)
 {
 	LONG status;
 	HKEY adapters_key, hkey;
 	DWORD len, type;
+	int adapter_type;
 	char buf[40];
 	wchar_t name[40];
 	char keyname[strlen(CONNECTIONS_KEY) + sizeof(buf) + 1 + strlen("\\Connection")];
-	int i = 0, found = 0;
-	intptr_t ret = -1;
+	int i = 0;
+	intptr_t ret = OPEN_TUN_SOFTFAIL;
 
 	status = RegOpenKeyExA(HKEY_LOCAL_MACHINE, ADAPTERS_KEY, 0,
 			       KEY_READ, &adapters_key);
@@ -75,13 +86,13 @@ static intptr_t search_taps(struct openconnect_info *vpninfo, tap_callback *cb,
 			     _("Error accessing registry key for network adapters\n"));
 		return -EIO;
 	}
-	while (1) {
+	while (ret == OPEN_TUN_SOFTFAIL) {
 		len = sizeof(buf);
 		status = RegEnumKeyExA(adapters_key, i++, buf, &len,
 				       NULL, NULL, NULL, NULL);
 		if (status) {
 			if (status != ERROR_NO_MORE_ITEMS)
-				ret = -1;
+				ret = OPEN_TUN_HARDFAIL;
 			break;
 		}
 
@@ -103,16 +114,19 @@ static intptr_t search_taps(struct openconnect_info *vpninfo, tap_callback *cb,
 			RegCloseKey(hkey);
 			continue;
 		}
-		if (strcmp(buf, TAP_COMPONENT_ID)) {
-			vpn_progress(vpninfo, PRG_TRACE,
-				     _("%s\\ComponentId is '%s' not '%s'\n"),
-				     keyname, buf, TAP_COMPONENT_ID);
+		if (!strcmp(buf, TAP_COMPONENT_ID) || !strcmp(buf, "root\\" TAP_COMPONENT_ID))
+			adapter_type = ADAPTER_TUNTAP;
+		else if (!strcmp(buf, "wintun"))
+			adapter_type = ADAPTER_WINTUN;
+		else {
+			vpn_progress(vpninfo, PRG_TRACE, _("%s\\ComponentId is unknown '%s'\n"),
+				     keyname, buf);
 			RegCloseKey(hkey);
 			continue;
 		}
 
 		vpn_progress(vpninfo, PRG_TRACE, _("Found %s at %s\n"),
-			     TAP_COMPONENT_ID, keyname);
+			     buf, keyname);
 
 		len = sizeof(buf);
 		status = RegQueryValueExA(hkey, "NetCfgInstanceId", NULL,
@@ -144,19 +158,11 @@ static intptr_t search_taps(struct openconnect_info *vpninfo, tap_callback *cb,
 			continue;
 		}
 
-		found++;
-
-		ret = cb(vpninfo, buf, name);
-		if (!ret && !all)
-			break;
+		ret = cb(vpninfo, adapter_type, buf, name);
 	}
 
 	RegCloseKey(adapters_key);
 
-	if (!found)
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("No Windows-TAP adapters found. Is the driver installed?\n"));
-
 	return ret;
 }
 
@@ -180,7 +186,6 @@ static int get_adapter_index(struct openconnect_info *vpninfo, char *guid)
 		 * going to manage for GetAdaptersInfo(). Give up. */
 		return buf_free(buf);
 	}
-
 	status = GetAdapterIndex((void *)buf->data, &idx);
 	buf_free(buf);
 	if (status == NO_ERROR) {
@@ -216,33 +221,141 @@ static int get_adapter_index(struct openconnect_info *vpninfo, char *guid)
 			break;
 		}
 	}
-
 	free(adapters_buf);
 	return ret;
 }
 
-static intptr_t open_tun(struct openconnect_info *vpninfo, char *guid, wchar_t *wname)
+static int check_address_conflicts(struct openconnect_info *vpninfo)
+{
+	ULONG bufSize = 15000;
+	PIP_ADAPTER_ADDRESSES addresses = NULL;
+	DWORD LastError;
+	int ret = 0;
+
+	struct in_addr our_ipv4_addr;
+	struct in6_addr our_ipv6_addr;
+	if (vpninfo->ip_info.addr && !inet_aton(vpninfo->ip_info.addr, &our_ipv4_addr))
+		return -EINVAL;
+	if (vpninfo->ip_info.addr6 && !inet_pton(AF_INET6, vpninfo->ip_info.addr6, &our_ipv6_addr))
+		return -EINVAL;
+
+	/* XX: repeat twice, because it may tell us we need a bigger buffer */
+	for (int tries=0; tries<2; tries++) {
+		free(addresses);
+		addresses = malloc(bufSize);
+		if (!addresses)
+			return -ENOMEM;
+
+		/* AF_UNSPEC means both Legacy IP and IPv6 */
+		LastError = GetAdaptersAddresses(AF_UNSPEC,
+						 GAA_FLAG_SKIP_ANYCAST | GAA_FLAG_SKIP_MULTICAST,
+						 NULL, addresses, &bufSize);
+		if (LastError != ERROR_BUFFER_OVERFLOW)
+			break;
+	}
+	if (LastError != ERROR_SUCCESS) {
+		char *errstr = openconnect__win32_strerror(LastError);
+		vpn_progress(vpninfo, PRG_ERR, _("GetAdaptersAddresses() failed: %s\n"), errstr);
+		ret = -EIO;
+		goto out;
+	}
+
+	for (; addresses; addresses=addresses->Next) {
+		/* XX: skip "our own" adapter */
+		if (addresses->IfIndex == vpninfo->tun_idx)
+			continue;
+
+		for (PIP_ADAPTER_UNICAST_ADDRESS ua = addresses->FirstUnicastAddress;
+		     ua; ua=ua->Next) {
+			struct sockaddr *a = ua->Address.lpSockaddr;
+			union {
+				struct sockaddr_in s4;
+				struct sockaddr_in6 s6;
+			} *sa = (void *)a;
+			int needs_reclaim = 0;
+
+			if (a->sa_family == AF_INET) {
+				if (vpninfo->ip_info.addr && our_ipv4_addr.s_addr == sa->s4.sin_addr.s_addr)
+					needs_reclaim = 1;
+			} else if (a->sa_family == AF_INET6) {
+				if (vpninfo->ip_info.addr6 && !memcmp(&our_ipv6_addr, &sa->s6.sin6_addr, sizeof(sa->s6.sin6_addr)))
+					needs_reclaim = 1;
+			}
+
+			if (needs_reclaim) {
+				if (addresses->OperStatus == IfOperStatusUp) {
+					/* XX: Interface is up. We shouldn't steal its address */
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Adapter \"%S\" / %ld is UP and using our IPv%d address. Cannot resolve.\n"),
+						     addresses->FriendlyName, addresses->IfIndex, a->sa_family == AF_INET ? 4 : 6);
+					ret = -ENOENT;
+					goto out;
+				} else {
+					/* XX: Interface is down */
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Adapter \"%S\" / %ld is DOWN and using our IPv%d address. We will reclaim the address from it.\n"),
+						     addresses->FriendlyName, addresses->IfIndex, a->sa_family == AF_INET ? 4 : 6);
+
+					/* XX: In order to use DeleteUnicastIpAddressEntry(), we bizarrely have to iterate through a
+					 * whole different table of IP addresses to find the right row. I previously tried
+					 * faking/synthesizing the requisite MIB_UNICASTIPADDRESS_TABLE rows, and it did not
+					 * work.
+					 */
+					int found = 0;
+					PMIB_UNICASTIPADDRESS_TABLE pipTable = NULL;
+					LastError = GetUnicastIpAddressTable(AF_UNSPEC, &pipTable);
+					if (LastError != ERROR_SUCCESS) {
+						char *errstr = openconnect__win32_strerror(LastError);
+						vpn_progress(vpninfo, PRG_ERR, _("GetUnicastIpAddressTable() failed: %s\n"), errstr);
+						ret = -EIO;
+						goto out;
+					}
+					for (int i = 0; i < pipTable->NumEntries; i++) {
+						if (pipTable->Table[i].Address.si_family == AF_INET && a->sa_family == AF_INET) {
+							if (sa->s4.sin_addr.s_addr == pipTable->Table[i].Address.Ipv4.sin_addr.s_addr)
+								found = 1;
+						} else if (pipTable->Table[i].Address.si_family == AF_INET6 && a->sa_family == AF_INET6) {
+							if (!memcmp(&sa->s6.sin6_addr, &pipTable->Table[i].Address.Ipv6, sizeof(sa->s6.sin6_addr)))
+								found = 1;
+						}
+						if (found) {
+							LastError = DeleteUnicastIpAddressEntry(&pipTable->Table[i]);
+							break;
+						}
+					}
+					FreeMibTable(pipTable);
+
+					if (LastError != NO_ERROR) {
+						char *errstr = openconnect__win32_strerror(LastError);
+						vpn_progress(vpninfo, PRG_ERR, _("DeleteUnicastIpAddressEntry() failed: %s\n"), errstr);
+						ret = -EIO;
+						goto out;
+					}
+
+					if (!found) {
+						vpn_progress(vpninfo, PRG_ERR,
+							     _("GetUnicastIpAddressTable() did not find matching address to reclaim\n"));
+						/* ret = -EIO;
+						   goto out; */
+					}
+
+				}
+			}
+		}
+	}
+
+ out:
+	free(addresses);
+	return ret;
+}
+
+static intptr_t open_tuntap(struct openconnect_info *vpninfo, char *guid, wchar_t *wname)
 {
-	struct oc_text_buf *namebuf = buf_alloc();
 	char devname[80];
 	HANDLE tun_fh;
 	ULONG data[3];
 	DWORD len;
 
-	buf_append_from_utf16le(namebuf, wname);
-	if (buf_error(namebuf)) {
-	err_out:
-		buf_free(namebuf);
-		return -1;
-	}
-
-	if (vpninfo->ifname && strcmp(namebuf->data, vpninfo->ifname)) {
-		vpn_progress(vpninfo, PRG_DEBUG,
-			     _("Ignoring non-matching TAP interface \"%s\"\n"),
-			     namebuf->data);
-		goto err_out;
-	}
-
 	snprintf(devname, sizeof(devname), DEVTEMPLATE, guid);
 	tun_fh = CreateFileA(devname, GENERIC_WRITE|GENERIC_READ, 0, 0,
 			     OPEN_EXISTING,
@@ -251,26 +364,26 @@ static intptr_t open_tun(struct openconnect_info *vpninfo, char *guid, wchar_t *
 	if (tun_fh == INVALID_HANDLE_VALUE) {
 		vpn_progress(vpninfo, PRG_ERR, _("Failed to open %s\n"),
 			     devname);
-		goto err_out;
+		return OPEN_TUN_SOFTFAIL;
 
 	}
-	vpn_progress(vpninfo, PRG_DEBUG, _("Opened tun device %s\n"), namebuf->data);
+	vpn_progress(vpninfo, PRG_DEBUG, _("Opened tun device %S\n"), wname);
 
 	if (!DeviceIoControl(tun_fh, TAP_IOCTL_GET_VERSION,
-			     data, sizeof(&data), data, sizeof(data),
+			     data, sizeof(data), data, sizeof(data),
 			     &len, NULL)) {
 		char *errstr = openconnect__win32_strerror(GetLastError());
 
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to obtain TAP driver version: %s\n"), errstr);
 		free(errstr);
-		goto err_out;
+		return OPEN_TUN_HARDFAIL;
 	}
 	if (data[0] < 9 || (data[0] == 9 && data[1] < 9)) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Error: TAP-Windows driver v9.9 or greater is required (found %ld.%ld)\n"),
 			     data[0], data[1]);
-		goto err_out;
+		return -1;
 	}
 	vpn_progress(vpninfo, PRG_DEBUG, "TAP-Windows driver v%ld.%ld (%ld)\n",
 		     data[0], data[1], data[2]);
@@ -291,7 +404,7 @@ static intptr_t open_tun(struct openconnect_info *vpninfo, char *guid, wchar_t *
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to set TAP IP addresses: %s\n"), errstr);
 		free(errstr);
-		goto err_out;
+		return OPEN_TUN_HARDFAIL;
 	}
 
 	data[0] = 1;
@@ -303,28 +416,135 @@ static intptr_t open_tun(struct openconnect_info *vpninfo, char *guid, wchar_t *
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to set TAP media status: %s\n"), errstr);
 		free(errstr);
-		goto err_out;
+		return OPEN_TUN_HARDFAIL;
 	}
 
-	if (!vpninfo->ifname) {
-		vpninfo->ifname = namebuf->data;
-		namebuf->data = NULL;
-	}
-	buf_free(namebuf);
-
-	get_adapter_index(vpninfo, guid);
-
 	return (intptr_t)tun_fh;
 }
 
+static intptr_t open_tun(struct openconnect_info *vpninfo, int adapter_type, char *guid, wchar_t *wname)
+{
+	intptr_t ret = -1;
+
+	if (vpninfo->ifname_w && wcscmp(wname, vpninfo->ifname_w)) {
+		vpn_progress(vpninfo, PRG_DEBUG,
+			     _("Ignoring non-matching interface \"%S\"\n"),
+			     wname);
+		return 0;
+	}
+
+	if (adapter_type == ADAPTER_TUNTAP)
+		ret = open_tuntap(vpninfo, guid, wname);
+	else
+		ret = open_wintun(vpninfo, guid, wname);
+
+	if (ret == OPEN_TUN_SOFTFAIL || ret == OPEN_TUN_HARDFAIL)
+		return ret;
+
+	/*
+	 * We have found the adapter and opened it successfully. Now set
+	 * vpninfo->ifname accordingly, if necessary, and find $TUNIDX
+	 * for the script to use to configure it.
+	 */
+	if (!vpninfo->ifname) {
+		struct oc_text_buf *namebuf = buf_alloc();
+
+		buf_append_from_utf16le(namebuf, wname);
+		if (buf_error(namebuf)) {
+			vpn_progress(vpninfo, PRG_ERR, _("Could not convert interface name to UTF-8\n"));
+			os_shutdown_tun(vpninfo);
+
+			buf_free(namebuf);
+			return OPEN_TUN_HARDFAIL;
+		}
+		vpninfo->ifname = namebuf->data;
+		namebuf->data = NULL;
+		buf_free(namebuf);
+	}
+
+	get_adapter_index(vpninfo, guid);
+
+	vpn_progress(vpninfo, adapter_type ? PRG_ERR : PRG_INFO,
+		     _("Using %s device '%s', index %d\n"),
+		     adapter_type ? "Wintun" : "TAP-Windows",
+		     vpninfo->ifname, vpninfo->tun_idx);
+	if (adapter_type == ADAPTER_WINTUN)
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("WARNING: Support for Wintun is experimental and may be unstable. If you\n"
+			       "  encounter problems, install the TAP-Windows driver instead. See\n"
+			       "  https://www.infradead.org/openconnect/building.html\n"));
+
+	return ret;
+}
+
+static intptr_t create_ifname_w(struct openconnect_info *vpninfo, const char *ifname) {
+	struct oc_text_buf *ifname_buf = buf_alloc();
+	buf_append_utf16le(ifname_buf, vpninfo->ifname);
+
+	if (buf_error(ifname_buf)) {
+		vpn_progress(vpninfo, PRG_ERR, _("Could not construct interface name\n"));
+		return buf_free(ifname_buf);
+	}
+
+	free(vpninfo->ifname_w);
+	vpninfo->ifname_w = (wchar_t *)ifname_buf->data;
+	ifname_buf->data = NULL;
+	buf_free(ifname_buf);
+	return 0;
+}
+
 intptr_t os_setup_tun(struct openconnect_info *vpninfo)
 {
-	return search_taps(vpninfo, open_tun, 0);
+	intptr_t ret;
+
+	if (vpninfo->ifname) {
+		ret = create_ifname_w(vpninfo, vpninfo->ifname);
+		if (ret)
+			return ret;
+	}
+
+	ret = search_taps(vpninfo, open_tun);
+
+	if (ret == OPEN_TUN_SOFTFAIL) {
+		if (!vpninfo->ifname_w) {
+			ret = create_ifname_w(vpninfo, vpninfo->hostname);
+			if (ret)
+				return ret;
+		}
+
+		/* Try creating a Wintun instead of TAP */
+		int retw = create_wintun(vpninfo);
+		if (!retw) {
+			ret = search_taps(vpninfo, open_tun);
+
+			if (ret == OPEN_TUN_SOFTFAIL)
+				ret = OPEN_TUN_HARDFAIL;
+			if (ret == OPEN_TUN_HARDFAIL)
+				os_shutdown_wintun(vpninfo);
+		} else if (retw == -EPERM) {
+			ret = OPEN_TUN_HARDFAIL;
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Access denied creating Wintun adapter. Are you running with Administrator privileges?\n"));
+		}
+	}
+
+	if (ret == OPEN_TUN_SOFTFAIL) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Neither Windows-TAP nor Wintun adapters were found. Is the driver installed?\n"));
+		ret = OPEN_TUN_HARDFAIL;
+	}
+
+	if (check_address_conflicts(vpninfo) < 0)
+		ret = OPEN_TUN_HARDFAIL; /* already complained about it */
+
+	return ret;
 }
 
 int os_read_tun(struct openconnect_info *vpninfo, struct pkt *pkt)
 {
 	DWORD pkt_size;
+	if (vpninfo->wintun)
+		return os_read_wintun(vpninfo, pkt);
 
  reread:
 	if (!vpninfo->tun_rd_pending &&
@@ -377,6 +597,9 @@ int os_write_tun(struct openconnect_info *vpninfo, struct pkt *pkt)
 	DWORD err;
 	char *errstr;
 
+	if (vpninfo->wintun)
+		return os_write_wintun(vpninfo, pkt);
+
 	if (WriteFile(vpninfo->tun_fh, pkt->data, pkt->len, &pkt_size, &vpninfo->tun_wr_overlap)) {
 		vpn_progress(vpninfo, PRG_TRACE,
 			     _("Wrote %ld bytes to tun\n"), pkt_size);
@@ -407,6 +630,12 @@ int os_write_tun(struct openconnect_info *vpninfo, struct pkt *pkt)
 void os_shutdown_tun(struct openconnect_info *vpninfo)
 {
 	script_config_tun(vpninfo, "disconnect");
+
+	if (vpninfo->wintun) {
+		os_shutdown_wintun(vpninfo);
+		return;
+	}
+
 	CloseHandle(vpninfo->tun_fh);
 	vpninfo->tun_fh = NULL;
 	CloseHandle(vpninfo->tun_rd_overlap.hEvent);
@@ -418,6 +647,9 @@ int openconnect_setup_tun_fd(struct openconnect_info *vpninfo, intptr_t tun_fd)
 	ULONG data;
 	DWORD len;
 
+	if (vpninfo->wintun)
+		return setup_wintun_fd(vpninfo, tun_fd);
+
 	/* Toggle media status so that network location awareness picks up all the configuration
 	   that occurred and properly assigns the network so the user can adjust firewall
 	   settings. */
